with Strings, IPowers, QPowers, Roots, Std_Floats.Rounding;
use Strings;

pragma Elaborate_All (Strings,IPowers,Roots,Std_Floats.Rounding);
pragma Optimize (Off);

package body Balls is

  package Txt_IO renames Ada.Text_IO;
  
--- auxiliary
  
  function ReturnTrue(Dummy: Ball) return Boolean is
    pragma Unreferenced(Dummy);
  begin
    return True;
  end ReturnTrue;

  procedure NOP(Dummy: in out Ball) is
    pragma Unreferenced(Dummy);
  begin
    null;
  end NOP;
  
--- basic

  function Info(Dummy: Ball) return Scalar_Info is
    pragma Unreferenced(Dummy);
  begin
    return BInfo;
  end Info;

  function IsSharp(S: Ball) return Boolean is
  begin
    return S.R=RZero;
  end IsSharp;

  function IsZero(S: Ball) return Boolean is
  begin
    return IsZero(S.C) and then S.R=RZero;
  end IsZero;

  function "="(S1,S2: Ball) return Boolean is
  begin
    return (S1.C=S2.C) and then (S1.R=S2.R);
  end "=";
  
  procedure SetZero(S: in out Ball) is
  begin
    SetZero(S.C);
    S.R := RZero;
  end SetZero;

  function ReturnZero(Dummy: in Ball) return Ball is
    pragma Unreferenced(Dummy);
  begin
    return BallZero;
  end  ReturnZero;

  function Copy(S: Ball) return Ball is
  begin
    return S;
  end Copy;

  procedure Copy(S1: in Ball; S2: in out Ball) is
  begin
    Copy(S1.C,S2.C);
    S2.R := S1.R;
  end Copy;

  procedure Swap(S1,S2: in out Ball) is
    R: constant Rad := S1.R;
  begin
    Swap(S1.C,S2.C);
    S1.R := S2.R; S2.R := R;
  end Swap;

--- sets

  function Center0(S: Ball) return Boolean is
  begin
    return IsZero(S.C);
  end Center0;

  function Contains0(S: Ball) return Logical is
    SR: constant Rep := Scal(Flt(S.R));
    SC: constant Rep := Norm(S.C);
  begin
    if Compare(SR,SC)<0 then
      return False;
    else
      return True;
    end if;
  exception
    when Not_Certain => return Uncertain;
  end Contains0;

  function Contains(S1,S2: Ball) return Logical is
    Y: Rep;
  begin
    if S1.R<S2.R then return False; end if;
    if S2.C=S1.C then return True; end if;
    Y := abs(S2.C-S1.C);
    if (Val(S1.R-S2.R)-Y)<CZero then return False; end if;
    if (Val(S2.R-S1.R)+Y)>CZero then
      return Uncertain;
    end if;
    return True;
  end Contains;
  
  procedure BallAt0(F: in Flt; S: in out Ball) is
  begin
    S.R := abs(Rad(F));
    SetZero(S.C);
  end BallAt0;
  
  function BallAt0(F: Flt) return Ball is
  begin
    return (CZero,abs(Rad(F)));
  end BallAt0;
  
  procedure ToErr(S: in out Ball) is
  begin
    S.R := Rad(MaxNorm(S.C))+S.R;
    SetZero(S.C);
  end ToErr;
  
  procedure ToErr(S1: in Ball; S2: in out Ball) is
  begin
    S2.R := Rad(MaxNorm(S1.C))+S1.R;
    SetZero(S2.C);
  end ToErr;
    
  function ToErr(S: Ball) return Ball is
  begin
    return (CZero,Rad(MaxNorm(S.C))+S.R);
  end ToErr;
  
  procedure RBall(R1,R2: in Rep; S: in out Ball) is
    Eps: Rad;
    Tmp: Rep;
  begin
    if R2>R1 then
      Diff(R2,R1,Tmp);
    else
      Diff(R1,R2,Tmp);
    end if;
    SumErr(R1,R2,S.C,Eps);
    S.R := Half*(Upper(Tmp)+Eps);
    Mult(Half,S.C,Tmp);
  end RBall;
  
  procedure Center(S: in out Ball) is
  begin
    S.R := RZero;
  end Center;
  
  procedure Center(S1: in Ball; S2: in out Ball) is
  begin
    Copy(S1.C,S2.C);
    S2.R := RZero;
  end Center;
    
  function Center(S: Ball) return Ball is
  begin
    return (S.C,RZero);
  end Center;
  
  procedure ModCenter(S: in out Ball) is
  begin
    SetZero(S.C);
  end ModCenter;
  
  procedure ModCenter(S1: in Ball; S2: in out Ball) is
  begin
    SetZero(S2.C);
    S2.R := S1.R;
  end ModCenter;
  
  function ModCenter(S: Ball) return Ball is
  begin
    return (CZero,S.R);
  end ModCenter;
  
  procedure ErrMult(B: in Radius; S: in out Ball) is
  begin
    S.R := abs(Rad(B)*S.R);
  end ErrMult;
  
  procedure Union(S1: in Ball; S2: in out Ball) is
    Eps: Rad;
    RU1,RU2,YL,YU,Y: Rep;
  begin
    if S1.C=S2.C then
      S2.R := Max(S1.R,S2.R);
    else
      Val(S1.R,RU1);
      Val(S2.R,RU2);
      YL := Min(-(RU1-S1.C),-(RU2-S2.C));
      YU :=  Max(S1.C+RU1,S2.C+RU2);
      SumErr(YL,YU,Y,Eps);
      S2.C := Half*Y;
      S2.R := Half*(Upper(YU-YL)+Eps);
    end if;
  end Union;
  
  function Union(S1,S2: Ball) return Ball is
    S: Ball := S2;
  begin
    Union(S1,S);
    return S;
  end Union;
  
  procedure Intersection(S1: in Ball; S2: in out Ball; Empty: out Logical) is
    Eps: Rad;
    RU1,RU2,YL,YU,Y: Rep;
  begin
    Empty := Uncertain;
    if S1.C=S2.C then
      Empty := False;
      S2.R := Min(S1.R,S2.R);
    else
      Val(S1.R,RU1);
      Val(S2.R,RU2);
      YL := Max(S1.C-RU1,S2.C-RU2);
      YU := -Max(-S1.C-RU1,-S2.C-RU2);
      if YL <= YU then
        Empty := False;
      end if;
      YL := Max(-(RU1-S1.C),-(RU2-S2.C));
      YU := Min(S1.C+RU1,S2.C+RU2);
      if YU < YL then
        Empty := True;
      end if;
      SumErr(YL,YU,Y,Eps);
      S2.C := Half*Y;
      S2.R := Half*(Upper(YU-YL)+Eps);
    end if;
  end Intersection;
  
--- order
  
  procedure RSup(S: in Ball; R: in out Rep) is
  begin
    Val(S.R,R);
    Add(S.C,R);
  end RSup;

  function RSup(S: Ball) return Rep is
    RU1: Rep;
  begin
    RSup(S,RU1);
    return RU1;
  end RSup;

  procedure RInf(S: in Ball; R: in out Rep) is
  begin
    Val(S.R,R);
    Sub(S.C,R);
    Neg(R);
  end RInf;

  function RInf(S: Ball) return Rep is
    RL1: Rep;
  begin
    RInf(S,RL1);
    return RL1;
  end RInf;

  function Sign(S: Ball) return Integer is
  begin
    if Sign(RInf(S))>0 then return 1; end if;
    if Sign(RSup(S))<0 then return -1; end if;
    if not IsZero(S) then
      raise Not_Certain;
    end if;
    return 0;
  end Sign;
  
  function Compare(S1,S2: Ball) return Integer is
  begin
    if Compare(RInf(S1),RSup(S2))>0 then return 1; end if;
    if Compare(RSup(S1),RInf(S2))<0 then return -1; end if;
    if S1 /= S2 then
      raise Not_Certain;
    end if;
    return 0;
  end Compare;
    
  function "<"(S1,S2: Ball) return Boolean is
  begin
    if RSup(S1) < RInf(S2) then return True; end if;
    if RInf(S1) < RSup(S2) then
      raise Not_Certain;
    end if;
    return False;
  end "<";

  function "<="(S1,S2: Ball) return Boolean is
  begin
    if RSup(S1) <= RInf(S2) then return True; end if;
    if RInf(S1) <= RSup(S2) then
      raise Not_Certain;
    end if;
    return False;
  end "<=";

  function ">"(S1,S2: Ball) return Boolean is
  begin
    if RInf(S1) > RSup(S2) then return True; end if;
    if RSup(S1) > RInf(S2) then
      raise Not_Certain;
    end if;
    return False;
  end ">";

  function ">="(S1,S2: Ball) return Boolean is
  begin
    if RInf(S1) >= RSup(S2) then return True; end if;
    if RSup(S1) >= RInf(S2) then
      raise Not_Certain;
    end if;
    return False;
  end ">=";

  procedure Min(S1: in Ball; S2: in out Ball) is
    RU1,RU2,RL1,RL2: Rep;
  begin
    RInf(S1,RL1);
    RSup(S2,RU2);
    if RL1<RU2 then
      RSup(S1,RU1);
      RInf(S2,RL2);
      if RL2<RU1 then
        RBall(Min(RL1,RL2),Min(RU1,RU2),S2);
      else
        Copy(S1,S2);
      end if;
    end if;
  end Min;
  
  procedure Min(S1,S2: in Ball; S3: in out Ball) is
  begin
    Copy(S2,S3);
    Min(S1,S3);
  end Min;    
  
  function Min(S1,S2: Ball) return Ball is
    S3: Ball;
  begin
    Copy(S2,S3);
    Min(S1,S3);
    return S3;
  end Min;

  procedure Max(S1: in Ball; S2: in out Ball) is
    RU1,RU2,RL1,RL2: Rep;
  begin
    RSup(S1,RU1);
    RInf(S2,RL2);
    if RL2<RU1 then
      RInf(S1,RL1);
      RSup(S2,RU2);
      if RL1<RU2 then
        RBall(Max(RL1,RL2),Max(RU1,RU2),S2);
      else
        Copy(S1,S2);
      end if;
    end if;
  end Max;
  
  procedure Max(S1,S2: in Ball; S3: in out Ball) is
  begin
    Copy(S2,S3);
    Max(S1,S3);
  end Max;    
  
  function Max(S1,S2: Ball) return Ball is
    S3: Ball;
  begin
    Copy(S2,S3);
    Max(S1,S3);
    return S3;
  end Max;

  function Inf(S: Ball) return Flt is
  begin
    return Flt(Lower(RInf(S)));
  end Inf;

  function Sup(S: Ball) return Flt is
  begin
    return Flt(Upper(RSup(S)));
  end Sup;

--- addition and multiplication

  procedure Neg(S: in out Ball) is
  begin
    Neg(S.C);
  end Neg;

  procedure Neg(S1: in Ball; S2: in out Ball) is
  begin
    Neg(S1.C,S2.C);
    S2.R := S1.R;
  end Neg;

  function "-"(S: Ball) return Ball is
    T: Ball;
  begin
    Neg(S.C,T.C);
    T.R := S.R;
    return T;
  end "-";
  
  procedure Add(I: in Integer; S: in out Ball) is
    Eps: Rad;
  begin
    AddErr(I,S.C,Eps);
    S.R := S.R+Eps;
  end Add;

  procedure Add(S1: in Ball; S2: in out Ball) is
    Eps: Rad;
  begin
    AddErr(S1.C,S2.C,Eps);
    S2.R := S1.R+S2.R+Eps;
  end Add;

  procedure Sum(S1,S2: in Ball; S3: in out Ball) is
    Eps: Rad;
  begin
    SumErr(S1.C,S2.C,S3.C,Eps);
    S3.R := S1.R+S2.R+Eps;
  end Sum;
  
  function "+"(S1,S2: Ball) return Ball is
    Eps: Rad;
    S3: Ball;
  begin
    SumErr(S1.C,S2.C,S3.C,Eps);
    S3.R := S1.R+S2.R+Eps;
    return S3;
  end "+";
  
  procedure Sub(S1: in Ball; S2: in out Ball) is
    Eps: Rad;
  begin
    SubErr(S1.C,S2.C,Eps);
    S2.R := S1.R+S2.R+Eps;
  end Sub;

  procedure Diff(S1,S2: in Ball; S3: in out Ball) is
    Eps: Rad;
  begin
    DiffErr(S1.C,S2.C,S3.C,Eps);
    S3.R := S1.R+S2.R+Eps;
  end Diff;
  
  function "-"(S1,S2: Ball) return Ball is
    Eps: Rad;
    S3: Ball;
  begin
    DiffErr(S1.C,S2.C,S3.C,Eps);
    S3.R := S1.R+S2.R+Eps;
    return S3;
  end "-";
  
  procedure Mult(R: in Flt; S,Tmp: in out Ball) is
    Eps: Rad renames Tmp.R;
  begin
    MultErr(R,S.C,Tmp.C,Eps);
    S.R := abs(Rad(R))*S.R+Eps;
  end Mult;

  procedure Mult(R: in Flt; S: in out Ball) is
    Tmp: Ball;
  begin
    Mult(R,S,Tmp);
  end Mult;
  
  procedure Prod(R: in Flt; S1: in Ball; S2: in out Ball) is
    Eps: Rad;
  begin
    ProdErr(R,S1.C,S2.C,Eps);
    S2.R := abs(Rad(R))*S1.R+Eps;
  end Prod;
  
  function "*"(R: Flt; S: Ball) return Ball is
    T: Ball;
  begin
    Prod(R,S,T);
    return T;
  end "*";
  
  procedure AddProd(R: in Flt; S1: in Ball; S2,Tmp: in out Ball) is
    Eps: Rad renames Tmp.R;
  begin
    AddProdErr(R,S1.C,S2.C,Tmp.C,Eps);
    S2.R := abs(Rad(R))*S1.R+S2.R+Eps;
  end AddProd;

  procedure AddProd(R: in Flt; S1: in Ball; S2: in out Ball) is
    Tmp: Ball;
  begin
    AddProd(R,S1,S2,Tmp);
  end AddProd;
  
  procedure Mult(Q: in Rational; S: in out Ball) is
    N: constant Integer := Num(Q);
    D: constant Integer := Den(Q);
    R: Rad renames S.R;
    Tmp: Rep;
    Eps: Rad;
  begin
    MultErr(Flt(N),S.C,Tmp,Eps);
    R := abs(Rad(N))*R+Eps;
    DivErr(Flt(D),S.C,Tmp,Eps);
    R := R/abs(Rad(D))+Eps;
  end Mult;
  
  procedure Mult(S1: in Ball; S2: in out Ball) is
    R1: constant Rad := Rad(MaxNorm(S1.C));
    R2: constant Rad := Rad(MaxNorm(S2.C))+S2.R;
    Eps: Rad;
  begin
    MultErr(S1.C,S2.C,Eps);
    S2.R := R1*S2.R+S1.R*R2+Eps;
  end Mult;

  procedure Prod(S1,S2: in Ball; S3: in out Ball) is
    R1: constant Rad := Rad(MaxNorm(S1.C));
    R2: constant Rad := Rad(MaxNorm(S2.C))+S2.R;
    Eps: Rad;
  begin
    ProdErr(S1.C,S2.C,S3.C,Eps);
    S3.R := R1*S2.R+S1.R*R2+Eps;
  end Prod;

  function "*"(S1,S2: Ball) return Ball is
    S3: Ball;
  begin
    Prod(S1,S2,S3);
    return S3;
  end "*";

  procedure AddProd(S1,S2: in Ball; S3,Tmp: in out Ball) is
    R1: constant Rad := Rad(MaxNorm(S1.C));
    R2: constant Rad := Rad(MaxNorm(S2.C))+S2.R;
    Eps: Rad renames Tmp.R;
  begin
    AddProdErr(S1.C,S2.C,S3.C,Tmp.C,Eps);
    S3.R := R1*S2.R+S1.R*R2+S3.R+Eps;
  end AddProd;
  
  procedure AddProd(S1,S2: in Ball; S3: in out Ball) is
    R1: constant Rad := Rad(MaxNorm(S1.C));
    R2: constant Rad := Rad(MaxNorm(S2.C))+S2.R;
    Eps: Rad;
  begin
    AddProdErr(S1.C,S2.C,S3.C,Eps);
    S3.R := R1*S2.R+S1.R*R2+S3.R+Eps;
  end AddProd;

  procedure SumProd(S1,S2,S3: in Ball; S4: in out Ball) is
    R1: constant Rad := Rad(MaxNorm(S1.C));
    R2: constant Rad := Rad(MaxNorm(S2.C))+S2.R;
    Eps: Rad;
  begin
    SumProdErr(S1.C,S2.C,S3.C,S4.C,Eps);
    S4.R := R1*S2.R+S1.R*R2+S3.R+Eps;
  end SumProd;
  
  procedure SubProd(S1,S2: in Ball; S3,Tmp: in out Ball) is
    R1: constant Rad := Rad(MaxNorm(S1.C));
    R2: constant Rad := Rad(MaxNorm(S2.C))+S2.R;
    Eps: Rad renames Tmp.R;
  begin
    SubProdErr(S1.C,S2.C,S3.C,Tmp.C,Eps);
    S3.R := R1*S2.R+S1.R*R2+S3.R+Eps;
  end SubProd;
  
  procedure Div(R: in Flt; S,Tmp: in out Ball) is
    Eps: Rad renames Tmp.R;
  begin
    DivErr(R,S.C,Tmp.C,Eps);
    S.R := S.R/abs(Rad(R))+Eps;
  end Div;

  procedure Div(R: in Flt; S: in out Ball) is
  begin
    if IsZero(S.C) then
      S.R := S.R/abs(Rad(R));
    else
      declare
        Tmp: Ball;
        Eps: Rad renames Tmp.R;
      begin
        DivErr(R,S.C,Tmp.C,Eps);
        S.R := S.R/abs(Rad(R))+Eps;
      end;
    end if;
  end Div;
  
  procedure Quot(S1: in Ball; R: in Flt; S2: in out Ball) is
    Eps: Rad renames S2.R;
  begin
    QuotErr(S1.C,R,S2.C,Eps);
    S2.R := S1.R/abs(Rad(R))+Eps;
  end Quot;

  function "/"(S: Ball; R: Flt) return Ball is
    T: Ball;
  begin
    Quot(S,R,T);
    return T;
  end "/";

  procedure Div(S1: in Ball; S2: in out Ball) is
    S: Ball;
  begin
    Inv(S1,S);
    Mult(S,S2);
  end Div;

  procedure Quot(S1,S2: in Ball; S3: in out Ball) is
  begin
    Inv(S2,S3);
    Mult(S1,S3);
  end Quot;
  
  function "/"(S1,S2: Ball) return Ball is
    S3: Ball;
  begin
    Inv(S2,S3);
    Mult(S1,S3);
    return S3;
  end "/";
  
  procedure Inv(S1: in Ball; S2: in out Ball) is
    Eps: Rad renames S2.R;
    A,R: Rad;
  begin
    if Sign(S1.C)<0 then
      A := -Upper(S1.C);
    else
      A := Lower(S1.C);
    end if;
    R := S1.R-A;
    if R >= RZero then
      raise Undefined with "Balls.Inv error: possible division by zero";
    end if;
    InvErr(S1.C,S2.C,Eps);
    S2.R := (S1.R/A)/(-R)+Eps;
  end Inv;

  function Inv(S: Ball) return Ball is
    T: Ball;
  begin
    Inv(S,T);
    return T;
  end Inv;
  
--- functions

  procedure Norm(S1: in Ball; S2: in out Ball) is
    U: constant Rad := Rad(MaxNorm(S1.C));
  begin
    if S1.R>U then
      S2.R := Half*(U+S1.R);
      Val(S2.R,S2.C);
    elsif Sign(S1.C)<0 then
      Neg(S1.C,S2.C);
      S2.R := S1.R;
    else
      Copy(S1.C,S2.C);
      S2.R := S1.R;
    end if;
  end Norm;
  
  function Norm(S: Ball) return Ball is
    T: Ball;
  begin
    Norm(S,T);
    return T;
  end Norm;

  function MaxNorm(S: Ball) return Radius is
  begin
    return MaxNorm(S.C)+Flt(S.R);
  end MaxNorm;
  
  procedure Sqr(S: in out Ball) is
    R: constant Rad := Rad(MaxNorm(S.C));
    Eps: Rad;
  begin
    MultErr(S.C,S.C,Eps);
    S.R := (R+R+S.R)*S.R+Eps;
  end Sqr;
  
  function Sqr(S: Ball) return Ball is
    T: Ball := S;
  begin
    Mult(S,T);
    return T;
  end Sqr;
  
  procedure Sqrt(S1: in Ball; S2: in out Ball) is
--- assumig correct rounding for Sqrt
  begin
    if IsZero(S1) then
      SetZero(S2);
    else
      declare
        C1: Rep renames S1.C;
        C2: Rep renames S2.C;
      begin
        Val(S1.R,C2);
        if C1=C2 then
          S2.R := Half*Sqrt(Two*S1.R);
          C2 := Val(S2.R);
        else
          RBall(-((C2-C1)/Sqrt(C1-C2)),Sqrt(C1+C2),S2);
        end if;
      end;
    end if;
  end Sqrt;
  
  function Sqrt(S: Ball) return Ball is
--- assumig correct rounding for Sqrt
  begin
    if IsZero(S) then return S; end if;
    declare
      T: Ball;
    begin
      Val(S.R,T.C);
      RBall(-((T.C-S.C)/Sqrt(S.C-T.C)),Sqrt(S.C+T.C),T);
      return T;
    end;
  end Sqrt;
  
  procedure Root(K: Positive; S1: in Ball; S2: in out Ball) is
    DX,YL,YU: Rep;
    DY: Rep renames DX;
    Tmp: Rep renames S2.C;
    Eps: Rad;
  begin
    Val(S1.R,DX);
    Root(K,S1.C,DX,YL,YU);
    Mult(Half,YU,Tmp);
    Mult(NegHalf,YL,Tmp);
    Sum(YU,YL,DY);
    DiffErr(YU,YL,S2.C,Eps);
    S2.R := Upper(DY)+Eps;
  end Root;
  
  function Root(K: Positive; S: Ball) return Ball is
    T: Ball;
  begin
    Root(K,S,T);
    return T;
  end Root;

  procedure Exp(S1: in Ball; S2: in out Ball) is
    DX,YL,YU: Rep;
    DY: Rep renames DX;
    Tmp: Rep renames S2.C;
    Eps: Rad;
  begin
    Val(S1.R,DX);
    Exp(S1.C,DX,YL,YU);
    Mult(Half,YU,Tmp);
    Mult(NegHalf,YL,Tmp);
    Sum(YU,YL,DY);
    DiffErr(YU,YL,S2.C,Eps);
    S2.R := Upper(DY)+Eps;
  end Exp;

  function Exp(S: Ball) return Ball is
    T: Ball;
  begin
    Exp(S,T);
    return T;
  end Exp;

  procedure Log(S1: in Ball; S2: in out Ball) is
    DX,YL,YU: Rep;
    DY: Rep renames DX;
    Tmp: Rep renames S2.C;
    Eps: Rad;
  begin
    Val(S1.R,DX);
    Log(S1.C,DX,YL,YU);
    Mult(Half,YU,Tmp);
    Mult(NegHalf,YL,Tmp);
    Sum(YU,YL,DY);
    DiffErr(YU,YL,S2.C,Eps);
    S2.R := Upper(DY)+Eps;
  end Log;
  
  function Log(S: Ball) return Ball is
    T: Ball;
  begin
    Log(S,T);
    return T;
  end Log;

  procedure ArcCos(S1: in Ball; S2: in out Ball) is
    DX,YL,YU: Rep;
    DY: Rep renames DX;
    Tmp: Rep renames S2.C;
    Eps: Rad;
  begin
    Val(S1.R,DX);
    ArcCos(S1.C,DX,YL,YU);
    Mult(Half,YU,Tmp);
    Mult(NegHalf,YL,Tmp);
    Sum(YU,YL,DY);
    DiffErr(YU,YL,S2.C,Eps);
    S2.R := Upper(DY)+Eps;
  end ArcCos;
  
  function ArcCos(S: Ball) return Ball is
    T: Ball;
  begin
    ArcCos(S,T);
    return T;
  end ArcCos;
  
  procedure ArcSin(S1: in Ball; S2: in out Ball) is
    DX,YL,YU: Rep;
    DY: Rep renames DX;
    Tmp: Rep renames S2.C;
    Eps: Rad;
  begin
    Val(S1.R,DX);
    ArcSin(S1.C,DX,YL,YU);
    Mult(Half,YU,Tmp);
    Mult(NegHalf,YL,Tmp);
    Sum(YU,YL,DY);
    DiffErr(YU,YL,S2.C,Eps);
    S2.R := Upper(DY)+Eps;
  end ArcSin;
  
  function ArcSin(S: Ball) return Ball is
    T: Ball;
  begin
    ArcSin(S,T);
    return T;
  end ArcSin;
  
  procedure Cos(S1: in Ball; S2: in out Ball) is
    YL,YU,DY: Rep;
    Tmp: Rep renames S2.C;
    Eps: Rad;
  begin
    Cos(S1.C,YL,YU);
    Mult(Half,YU,Tmp);
    Mult(NegHalf,YL,Tmp);
    Sum(YU,YL,DY);
    DiffErr(YU,YL,S2.C,Eps);
    S2.R := Upper(DY)+Eps+S1.R;
  end Cos;
  
  function Cos(S: Ball) return Ball is
    T: Ball;
  begin
    Cos(S,T);
    return T;
  end Cos;
  
  procedure Sin(S1: in Ball; S2: in out Ball) is
    YL,YU,DY: Rep;
    Tmp: Rep renames S2.C;
    Eps: Rad;
  begin
    Sin(S1.C,YL,YU);
    Mult(Half,YU,Tmp);
    Mult(NegHalf,YL,Tmp);
    Sum(YU,YL,DY);
    DiffErr(YU,YL,S2.C,Eps);
    S2.R := Upper(DY)+Eps+S1.R;
  end Sin;
  
  function Sin(S: Ball) return Ball is
    T: Ball;
  begin
    Sin(S,T);
    return T;
  end Sin;
  
  procedure Simple_Random(S: in out Ball) is
  begin
    Simple_Random(S.C);
    S.R := RZero;
  end Simple_Random;
  
  function Pi return Ball is
--- 2*ArcCos(0)
    YL,YU,DY: Rep;
    S: Ball;
    Eps: Rad;
  begin
    ArcCos(CZero,CZero,YL,YU);
    Diff(YU,YL,DY);
    SumErr(YU,YL,S.C,Eps);
    S.R := Upper(DY)+Eps;
    return S;
  end Pi;
  
--- conversion and i/o

  procedure Assign(I: in Integer; S: in out Ball) is
  begin
    Assign(I,S.C);
    S.R := RZero;
  end Assign;
  
  procedure Assign(Q: in Rational; S: in out Ball) is
  begin
    Assign(Num(Q),S.C);
    S.R := RZero;
    Div(Flt(Den(Q)),S);
  end Assign;
  
  procedure Assign(R: in Flt; S: in out Ball) is
  begin
    Assign(R,S.C);
    S.R := RZero;
  end Assign;

  function Scal(I: Integer) return Ball is
  begin
    return (Scal(I),RZero);
  end Scal;
  
  function Scal(Q: Rational) return Ball is
    S: Ball;
  begin
    Assign(Num(Q),S.C);
    S.R := RZero;
    Div(Flt(Den(Q)),S);
    return S;
  end Scal;

  function Scal(R: Flt) return Ball is
  begin
    return (Scal(R),RZero);
  end Scal;
  
  procedure Enclose(R1,R2: in Flt; S: in out Ball) is
    H1:  constant Rad := Rad(Half*R1);
    H2:  constant Rad := Rad(Half*R2);
    H3:  constant Rad := H1+H2;
    Eps: constant Rad := H3+(-H1-H2);
  begin
    Val(H3,S.C);
    S.R := abs(H2-H1)+Eps;
  end Enclose;
    
  function Enclose(R1,R2: Flt) return Ball is
    S: Ball;
  begin
    Enclose(R1,R2,S);
    return S;
  end Enclose;

  function Approx(S: Ball) return Flt is
  begin
    return Approx(S.C);
  end Approx;

  procedure Show1(N: in String; S: in Ball; NewLine: in Boolean := True) is
    Digs: constant Natural := 0; -- use default
  begin
    Show0(N & DecStr(S.C,Digs),False);
    if S.R /= RZero then
      Show0(" [" & DecStr(S.R,Digs) & "]",False);
    end if;
    Show0(NewLine);
  end Show1;

  procedure Show2(N: in String; S1,S2: in Ball; NewLine: in Boolean := True) is
  begin
    Show1(N,S1,False);
    Show1("   ",S2,NewLine);
  end Show2;

  procedure Put(F: in File_Type; S: in Ball; Decimal: in Boolean := False) is
  begin
    if Decimal then
      Txt_IO.Put_Line(F,DecStr(S.C));
    elsif IsZero(S.R) then
      Txt_IO.Put_Line(F,HexStr(S.C));
    else
      Txt_IO.Put_Line(F,HexStr(S.C) & "," & HexStr(S.R));
    end if;
  end Put;

  procedure Get(F: in File_Type; S: in out Ball; Decimal: in Boolean := False) is
    use Strings;
    V: constant String := Strings.Get_Next_Line(F);
    Rounded: Boolean;
    N: Integer := V'First;
    E: Rad := RZero;
  begin
    SetZero(S);
    while N <= V'Last loop
      exit when V(N)=',';
      N := N+1;
    end loop;
    if Decimal then
      S.C := ValDec(V(V'First .. N-1));
    else
      ValHex(V(V'First .. N-1),S.C,Rounded);
      if Rounded then
        if IsZero(S.C) then
          raise Sorry with "Balls.Get: should not happen";
        end if;
        E := Upper((CTiny-S.C)+S.C); --- cheap
      end if;
      if N<V'Last then
        ValHex(V(N+1 .. V'Last),S.R,Rounded);
        if Rounded then
          if S.R=RZero then
            raise Sorry with "Balls.Get: should not happen";
          end if;
          S.R := S.R+Tiny;
        end if;
      end if;
      S.R := S.R+E;
    end if;
  end Get;

  procedure Write(FileName: in String; S: in Ball; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,S,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; S: in out Ball; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,S,Decimal);
    Close(F);
  end Read;
  
  procedure PutStd(S: in Ball; C,R: out LLFloat) is
  begin
    Enclose(S.C,C,R);
    R := S.R+R;
  end PutStd;
  
  procedure GetStd(C,R: in LLFloat; S: in out Ball) is
  begin
    Val(C,S.C);
    S.R := R;
  end GetStd;
  
  procedure PutNum(S: in Ball; R: in out Rep) is
  begin
    Copy(S.C,R);
  end PutNum;
  
  procedure GetNum(R: in Rep; S: in out Ball) is
  begin
    Copy(R,S.C);
    S.R := RZero;
  end GetNum;
  
--- misc  
    
  function Get_Precision(S: Ball) return Positive is
  begin
    return Get_Precision(S.C);
  end Get_Precision;

  procedure Set_Active_Mode(S: in Ball) is
    use Std_Floats.Rounding;
  begin
    if Verbosity>1 then
      Show0("Balls active: Strict Rounding Up");
    end if;
    Set_Rounding_Mode(S.C,(Up,Strict));
    Set_Rounding_Mode(S.R,(Up,Strict));
  end Set_Active_Mode;

  procedure Set_Wait_Mode(S: in Ball) is
    use Std_Floats.Rounding;
  begin
    if Verbosity>1 then
      Show0("Balls waiting: Relaxed Rounding Nearest");
    end if;
    Set_Rounding_Mode(S.C,(Nearest,Relaxed));
    Set_Rounding_Mode(S.R,(Nearest,Relaxed));
  end Set_Wait_Mode;
  
  procedure Free_Cache(Dummy: in Ball) is
    pragma Unreferenced(Dummy);
  begin
    null;
  end Free_Cache;
  
--------------------------------------------------------------------
--------------------------------------------------------------------
  package BIP is new IPowers (Scalar => Ball);
  package BQP is new QPowers (Scalar => Ball, IPower => BIP.IPower);
--------------------------------------------------------------------
--------------------------------------------------------------------
  
  function "**"(S: Ball; I: Integer) return Ball is
    T: Ball;
  begin
    BIP.IPower(I,S,T);
    return T;
  end "**";
  
  procedure QPower(Q: in Rational; S1: in Ball; S2: in out Ball) renames BQP.QPower;
  
-------------------------------------------
-------------------------------------------
  package SR is new Roots (Scalar => Ball);
-------------------------------------------
-------------------------------------------
 
  procedure Roots2(B,C: in Ball; U1,U2,V: in out Ball) is
  begin
    SR.NewtonRoots2(B,C,U1,U2,V,64,False);
  end Roots2;
  
  procedure Roots3(B,C,D: in Ball; U0,U1,U2,V: in out Ball) is
  begin
    SR.NewtonRoots3(B,C,D,U0,U1,U2,V,64,False);
  end Roots3;
  
begin
  
  if CInfo.IsComplex or (not CInfo.IsNumeric) then
    raise Sorry with "expecting real and numeric scalar";
  end if;

end Balls;
