with Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts,Std_Floats;
use Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts,Std_Floats;

pragma Elaborate_All (Ada.Text_IO,Logicals,Globals,Ints,Rationals,Flts,Std_Floats);

generic
  
-------------------- begin_include scalars Rep

  type Rep is private;

--- basic
  with function Info(Dummy: Rep) return Scalar_Info is <>;
  with function IsSharp(S: Rep) return Boolean is <>;
  with function IsZero(S: Rep) return Boolean is <>;
  with function "="(S1,S2: Rep) return Boolean is <>;
  with procedure SetZero(S: in out Rep) is <>;
  with procedure Copy(S1: in Rep; S2: in out Rep) is <>;
  with procedure Swap(S1,S2: in out Rep) is <>;

--- sets
  with function Center0(S: Rep) return Boolean is <>;
  with function Contains0(S: Rep) return Logical is <>;
  with function Contains(S1,S2: Rep) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Rep) is <>;
  with function BallAt0(R: Flt) return Rep is <>;
  with procedure ToErr(S: in out Rep) is <>;
  with procedure ToErr(S1: in Rep; S2: in out Rep) is <>;
  with function ToErr(S: Rep) return Rep is <>;
  with procedure Center(S: in out Rep) is <>;
  with procedure Center(S1: in Rep; S2: in out Rep) is <>;
  with function Center(S: Rep) return Rep is <>;
  with procedure ModCenter(S: in out Rep) is <>;
  with procedure ModCenter(S1: in Rep; S2: in out Rep) is <>;
  with function ModCenter(S: Rep) return Rep is <>;
  with procedure ErrMult(R: in Radius; S: in out Rep) is <>;
  with procedure Union(S1: in Rep; S2: in out Rep) is <>;
  with function Union(S1,S2: Rep) return Rep is <>;
  with procedure Intersection(S1: in Rep; S2: in out Rep; Empty: out Logical) is <>;

--- order
  with function Sign(S: Rep) return Integer is <>;
  with function Compare(S1,S2: Rep) return Integer is <>;
  with function "<"(S1,S2: Rep) return Boolean is <>;
  with function "<="(S1,S2: Rep) return Boolean is <>;
  with function ">"(S1,S2: Rep) return Boolean is <>;
  with function ">="(S1,S2: Rep) return Boolean is <>;
  with procedure Min(S1: in Rep; S2: in out Rep) is <>;
  with procedure Min(S1,S2: in Rep; S3: in out Rep) is <>;
  with function Min(S1,S2: Rep) return Rep is <>;
  with procedure Max(S1: in Rep; S2: in out Rep) is <>;
  with procedure Max(S1,S2: in Rep; S3: in out Rep) is <>;
  with function Max(S1,S2: Rep) return Rep is <>;
  with function Inf(S: Rep) return Flt is <>; -- inf of real part
  with function Sup(S: Rep) return Flt is <>; -- sup of real part

--- arithmetic
  with procedure Neg(S: in out Rep) is <>;
  with procedure Neg(S1: in Rep; S2: in out Rep) is <>;
  with function "-"(S: Rep) return Rep is <>;
  with procedure Add(I: in Integer; S: in out Rep) is <>;
  with procedure Add(S1: in Rep; S2: in out Rep) is <>;
  with procedure Sum(S1,S2: in Rep; S3: in out Rep) is <>;
  with function "+"(S1,S2: Rep) return Rep is <>;
  with procedure Sub(S1: in Rep; S2: in out Rep) is <>;
  with procedure Diff(S1,S2: in Rep; S3: in out Rep) is <>;
  with function "-"(S1,S2: Rep) return Rep is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Rep) is <>;
  with procedure Mult(R: in Flt; S: in out Rep) is <>;
  with procedure Prod(R: in Flt; S1: in Rep; S2: in out Rep) is <>;
  with function "*"(R: Flt; S: Rep) return Rep is <>;
  with procedure AddProd(R: in Flt; S1: in Rep; S2,Tmp: in out Rep) is <>;
  with procedure AddProd(R: in Flt; S1: in Rep; S2: in out Rep) is <>;
  with procedure Mult(Q: in Rational; S: in out Rep) is <>;
  with procedure Mult(S1: in Rep; S2: in out Rep) is <>;
  with procedure Prod(S1,S2: in Rep; S3: in out Rep) is <>;
  with function "*"(S1,S2: Rep) return Rep is <>;
  with procedure AddProd(S1,S2: in Rep; S3,Tmp: in out Rep) is <>;
  with procedure AddProd(S1,S2: in Rep; S3: in out Rep) is <>;
  with procedure SubProd(S1,S2: in Rep; S3,Tmp: in out Rep) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Rep) is <>;
  with procedure Div(R: in Flt; S: in out Rep) is <>;
  with procedure Quot(S1: in Rep; R: in Flt; S2: in out Rep) is <>;
  with function "/"(S: Rep; R: Flt) return Rep is <>;
  with procedure Div(S1: in Rep; S2: in out Rep) is <>;
  with procedure Quot(S1,S2: in Rep; S3: in out Rep) is <>;
  with function "/"(S1,S2: Rep) return Rep is <>;
  with procedure Inv(S1: in Rep; S2: in out Rep) is <>;
  with function Inv(S: Rep) return Rep is <>;
  with function "**"(S: Rep; I: Integer) return Rep is <>;

--- fun
  with function IsReal(S: Rep) return Boolean is <>;
  with procedure Adjoint(S: in out Rep) is <>;
  with procedure Adjoint(S1: in Rep; S2: in out Rep) is <>;
  with function Adjoint(S: Rep) return Rep is <>;
  with procedure Real_Part(S: in out Rep) is <>;
  with procedure Imag_Part(S: in out Rep) is <>;
  with procedure Eval0(S: in out Rep) is <>;
  with function "abs"(S: Rep) return Rep is <>;
  with procedure Norm(S1: in Rep; S2: in out Rep) is <>;
  with function Norm(S: Rep) return Rep is <>;
  with function MaxNorm(S: Rep) return Radius is <>;
  with procedure Sqr(S: in out Rep) is <>;
  with function Sqr(S: Rep) return Rep is <>;
  with procedure Sqrt(S1: in Rep; S2: in out Rep) is <>;
  with function Sqrt(S: Rep) return Rep is <>;
  with procedure Root(K: in Positive; S1: in Rep; S2: in out Rep) is <>;
  with function Root(K: Positive; S: Rep) return Rep is <>;
  with procedure Exp(S1: in Rep; S2: in out Rep) is <>;
  with function Exp(S: Rep) return Rep is <>;
  with procedure ArcCos(S1: in Rep; S2: in out Rep) is <>;
  with function ArcCos(S: Rep) return Rep is <>;
  with procedure ArcSin(S1: in Rep; S2: in out Rep) is <>;
  with function ArcSin(S: Rep) return Rep is <>;
  with procedure Cos(S1: in Rep; S2: in out Rep) is <>;
  with function Cos(S: Rep) return Rep is <>;
  with procedure Sin(S1: in Rep; S2: in out Rep) is <>;
  with function Sin(S: Rep) return Rep is <>;
  with procedure Log(S1: in Rep; S2: in out Rep) is <>;
  with function Log(S: Rep) return Rep is <>;
  with procedure Simple_Random(S: in out Rep) is <>;
  with procedure QPower(Q: in Rational; S1: in Rep; S2: in out Rep) is <>;
  with procedure Roots2(B,C: in Rep; U1,U2,V: in out Rep) is <>;
  with procedure Roots3(B,C,D: in Rep; U0,U1,U2,V: in out Rep) is <>;

--- conversion and i/o
  with function Approx(S: Rep) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Rep) is <>;
  with procedure Assign(Q: in Rational; S: in out Rep) is <>;
  with procedure Assign(R: in Flt; S: in out Rep) is <>;
  with function Scal(I: Integer) return Rep is <>;
  with function Scal(Q: Rational) return Rep is <>;
  with function Scal(R: Flt) return Rep is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Rep) is <>;
  with function Enclose(R1,R2: Flt) return Rep is <>;
  with procedure Show1(N: in String; S: in Rep; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Rep; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Rep; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Rep; Decimal: in Boolean := False) is <>;

--- misc
  with function Get_Precision(R: Rep) return Positive is <>;
  with procedure Set_Active_Mode(Dummy: in Rep) is <>;
  with procedure Set_Wait_Mode(Dummy: in Rep) is <>;
  with procedure Free_Cache(S: in Rep) is <>;
-------------------- end_include scalars
  
  with procedure Set_Rounding_Mode(Dummy: in Rep; M: in Rounding_Mode) is <>;
  
  with procedure Val(R: in LLFloat; S: in out Rep) is <>;
  with function Val(R: LLFloat) return Rep is <>;
  with function Upper(S: Rep) return LLFloat is <>;
  with function Lower(S: Rep) return LLFloat is <>;
  with function HexStr(S: Rep; Digs: Natural := 0) return String is <>;
  with function DecStr(S: Rep; Digs: Natural := 0) return String is <>;
  with procedure ValHex(N: in String; R: in out Rep; Rounded: out Boolean) is <>;
  with function ValDec(N: String) return Rep is <>;
  
  with procedure Enclose(S: in Rep; C,R: out LLFloat) is <>;
  with procedure AddErr(I: in Integer; S: in out Rep; Err: out LLFloat) is <>;
  with procedure AddErr(S1: in Rep; S2: in out Rep; Err: out LLFloat) is <>;
  with procedure SumErr(S1,S2: in Rep; S3: in out Rep; Err: out LLFloat) is <>;
  with procedure SubErr(S1: in Rep; S2: in out Rep; Err: out LLFloat) is <>;
  with procedure DiffErr(S1,S2: in Rep; S3: in out Rep; Err: out LLFloat) is <>;
  with procedure MultErr(R: in Flt; S,Tmp: in out Rep; Err: out LLFloat) is <>;
  with procedure ProdErr(R: in Flt; S1: in Rep; S2: in out Rep; Err: out LLFloat) is <>;
  with procedure MultErr(S1: in Rep; S2: in out Rep; Err: out LLFloat) is <>;
  with procedure ProdErr(S1,S2: in Rep; S3: in out Rep; Err: out LLFloat) is <>;
  with procedure AddProdErr(R: in Flt; S1: in Rep; S2,Tmp: in out Rep; Err: out LLFloat) is <>;
  with procedure AddProdErr(S1,S2: in Rep; S3,Tmp: in out Rep; Err: out LLFloat) is <>;
  with procedure AddProdErr(S1,S2: in Rep; S3: in out Rep; Err: out LLFloat) is <>;
  with procedure SumProdErr(S1,S2,S3: in Rep; S4: in out Rep; Err: out LLFloat) is <>;
  with procedure SubProdErr(S1,S2: in Rep; S3,Tmp: in out Rep; Err: out LLFloat) is <>;
  with procedure DivErr(R: in Flt; S,Tmp: in out Rep; Err: out LLFloat) is <>;
  with procedure QuotErr(S1: in Rep; R: in Flt; S2: in out Rep; Err: out LLFloat) is <>;
  with procedure InvErr(S1: in Rep; S2: in out Rep; Err: out LLFloat) is <>;
  
  with procedure Root(K: in Positive; X,DX: in Rep; YL,YU: in out Rep) is <>;
  with procedure Exp(X,DX: in Rep; YL,YU: in out Rep) is <>;
  with procedure Log(X,DX: in Rep; YL,YU: in out Rep) is <>;
  with procedure ArcCos(X,DX: in Rep; YL,YU: in out Rep) is <>;
  with procedure ArcSin(X,DX: in Rep; YL,YU: in out Rep) is <>;
  with procedure Cos(X: in Rep; YL,YU: in out Rep) is <>;
  with procedure Sin(X: in Rep; YL,YU: in out Rep) is <>;

package Balls is -- balls in a commutative Banach algebra with unit

  pragma Elaborate_Body;
  
  subtype Rad is LLFloat;

  type Ball is      -- ball around constant
    record
      C: Rep;       -- center
      R: Rad;       -- radius
    end record;
  
--- auxiliary
  procedure NOP(Dummy: in out Ball);
  function ReturnTrue(Dummy: Ball) return Boolean;
  
--- basic
  function Info(Dummy: Ball) return Scalar_Info;
  function IsSharp(S: Ball) return Boolean;
  function IsZero(S: Ball) return Boolean;
  function "="(S1,S2: Ball) return Boolean;
  procedure SetZero(S: in out Ball);
  function ReturnZero(Dummy: in Ball) return Ball;
  procedure Copy(S1: in Ball; S2: in out Ball);
  function Copy(S: Ball) return Ball;
  procedure Swap(S1,S2: in out Ball);

--- sets
  function Center0(S: Ball) return Boolean;
  function Contains0(S: Ball) return Logical;
  function Contains(S1,S2: Ball) return Logical;
  procedure BallAt0(F: in Flt; S: in out Ball);
  function BallAt0(F: Flt) return Ball;
  procedure ToErr(S: in out Ball);
  procedure ToErr(S1: in Ball; S2: in out Ball);
  function ToErr(S: Ball) return Ball;
  procedure Center(S: in out Ball);
  procedure Center(S1: in Ball; S2: in out Ball);
  function Center(S: Ball) return Ball;
  procedure ModCenter(S: in out Ball);
  procedure ModCenter(S1: in Ball; S2: in out Ball);
  function ModCenter(S: Ball) return Ball;
  procedure ErrMult(B: in Radius; S: in out Ball);
  procedure Union(S1: in Ball; S2: in out Ball);
  function Union(S1,S2: Ball) return Ball;
  procedure Intersection(S1: in Ball; S2: in out Ball; Empty: out Logical); -- an enclosure

--- order
  function Sign(S: Ball) return Integer;
  function Compare(S1,S2: Ball) return Integer;
  function "<"(S1,S2: Ball) return Boolean;
  function "<="(S1,S2: Ball) return Boolean;
  function ">"(S1,S2: Ball) return Boolean;
  function ">="(S1,S2: Ball) return Boolean;
  procedure Min(S1: in Ball; S2: in out Ball);
  procedure Min(S1,S2: in Ball; S3: in out Ball);
  function Min(S1,S2: Ball) return Ball;
  procedure Max(S1: in Ball; S2: in out Ball);
  procedure Max(S1,S2: in Ball; S3: in out Ball);
  function Max(S1,S2: Ball) return Ball;
  function Inf(S: Ball) return Flt;
  function Sup(S: Ball) return Flt;

--- addition and multiplication etc.
  procedure Neg(S: in out Ball);
  procedure Neg(S1: in Ball; S2: in out Ball);
  function "-"(S: Ball) return Ball;
  procedure Add(I: in Integer; S: in out Ball);
  procedure Add(S1: in Ball; S2: in out Ball);
  procedure Sum(S1,S2: in Ball; S3: in out Ball);
  function "+"(S1,S2: Ball) return Ball;
  procedure Sub(S1: in Ball; S2: in out Ball);
  procedure Diff(S1,S2: in Ball; S3: in out Ball);
  function "-"(S1,S2: Ball) return Ball;
  procedure Mult(R: in Flt; S,Tmp: in out Ball);
  procedure Mult(R: in Flt; S: in out Ball);
  procedure Prod(R: in Flt; S1: in Ball; S2: in out Ball);
  function "*"(R: Flt; S: Ball) return Ball;
  procedure AddProd(R: in Flt; S1: in Ball; S2,Tmp: in out Ball);
  procedure AddProd(R: in Flt; S1: in Ball; S2: in out Ball);
  procedure Mult(Q: in Rational; S: in out Ball);
  procedure Mult(S1: in Ball; S2: in out Ball);
  procedure Prod(S1,S2: in Ball; S3: in out Ball);
  function "*"(S1,S2: Ball) return Ball;
  procedure AddProd(S1,S2: in Ball; S3,Tmp: in out Ball);
  procedure AddProd(S1,S2: in Ball; S3: in out Ball);
  procedure SumProd(S1,S2,S3: in Ball; S4: in out Ball);
  procedure SubProd(S1,S2: in Ball; S3,Tmp: in out Ball);
  procedure Div(R: in Flt; S,Tmp: in out Ball);
  procedure Div(R: in Flt; S: in out Ball);
  procedure Quot(S1: in Ball; R: in Flt; S2: in out Ball);
  function "/"(S: Ball; R: Flt) return Ball;
  procedure Div(S1: in Ball; S2: in out Ball);
  procedure Quot(S1,S2: in Ball; S3: in out Ball);
  function "/"(S1,S2: Ball) return Ball;
  procedure Inv(S1: in Ball; S2: in out Ball);
  function Inv(S: Ball) return Ball;
  function "**"(S: Ball; I: Integer) return Ball;

--- fun
--- Sqrt, Root, Exp, Log, ArcCos, Cos, Sin are only for real numbers
  function IsReal(S: Ball) return Boolean renames ReturnTrue;
  procedure Adjoint(S: in out Ball) renames NOP;
  procedure Adjoint(S1: in Ball; S2: in out Ball) renames Copy;
  function Adjoint(S: Ball) return Ball renames Copy;
  procedure Norm(S1: in Ball; S2: in out Ball);
  function Norm(S: Ball) return Ball;
  function "abs"(S: Ball) return Ball renames Norm;
  function MaxNorm(S: Ball) return Radius;
  procedure Sqr(S: in out Ball);
  function Sqr(S: Ball) return Ball;
  procedure Sqrt(S1: in Ball; S2: in out Ball);
  function Sqrt(S: Ball) return Ball;
  procedure Root(K: Positive; S1: in Ball; S2: in out Ball);
  function Root(K: Positive; S: Ball) return Ball; --- uses MPFR
  procedure Exp(S1: in Ball; S2: in out Ball);
  function Exp(S: Ball) return Ball;               --- uses MPFR
  procedure Log(S1: in Ball; S2: in out Ball);
  function Log(S: Ball) return Ball;               --- uses MPFR
  procedure ArcCos(S1: in Ball; S2: in out Ball);
  function ArcCos(S: Ball) return Ball;            --- uses MPFR
  procedure ArcSin(S1: in Ball; S2: in out Ball);
  function ArcSin(S: Ball) return Ball;            --- uses MPFR
  procedure Cos(S1: in Ball; S2: in out Ball);
  function Cos(S: Ball) return Ball;               --- uses MPFR
  procedure Sin(S1: in Ball; S2: in out Ball);
  function Sin(S: Ball) return Ball;               --- uses MPFR
  procedure Simple_Random(S: in out Ball);
  function Pi return Ball;
  procedure QPower(Q: in Rational; S1: in Ball; S2: in out Ball);
  procedure Roots2(B,C: in Ball; U1,U2,V: in out Ball);
  procedure Roots3(B,C,D: in Ball; U0,U1,U2,V: in out Ball);

--- conversion and i/o
  procedure Assign(I: in Integer; S: in out Ball);
  procedure Assign(Q: in Rational; S: in out Ball);
  procedure Assign(R: in Flt; S: in out Ball);
  function Scal(I: Integer) return Ball;
  function Scal(Q: Rational) return Ball;
  function Scal(R: Flt) return Ball;
  procedure Enclose(R1,R2: in Flt; S: in out Ball);
  function Enclose(R1,R2: Flt) return Ball;
  function Approx(S: Ball) return Flt;
  procedure Real_Part(Dummy: in out Ball) renames NOP;
  procedure Imag_Part(S: in out Ball) renames SetZero;
  procedure Eval0(S: in out Ball) renames NOP;
  procedure Show1(N: in String; S: in Ball; NewLine: in Boolean := True);
  procedure Show2(N: in String; S1,S2: in Ball; NewLine: in Boolean := True);
  procedure Put(F: in File_Type; S: in Ball; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; S: in out Ball; Decimal: in Boolean := False);
  procedure Write(FileName: in String; S: in Ball; Decimal: in Boolean := False);
  procedure Read(FileName: in String; S: in out Ball; Decimal: in Boolean := False);
  
  procedure PutStd(S: in Ball; C,R: out LLFloat);
  procedure GetStd(C,R: in LLFloat; S: in out Ball);
  procedure PutNum(S: in Ball; R: in out Rep);
  procedure GetNum(R: in Rep; S: in out Ball);
  
--- misc
  function Get_Precision(S: Ball) return Positive;
  procedure Set_Active_Mode(S: in Ball);
  procedure Set_Wait_Mode(S: in Ball);
  procedure Free_Cache(Dummy: in Ball);

private
  
  procedure RBall(R1,R2: in Rep; S: in out Ball);
  procedure RSup(S: in Ball; R: in out Rep);
  function RSup(S: Ball) return Rep;
  procedure RInf(S: in Ball; R: in out Rep);
  function RInf(S: Ball) return Rep;
  
  RZero:     constant Rad := Rad(0);
  ROne:      constant Rad := Rad(1);
  Tiny:      constant Rad := Rad'Scaling(ROne,Flt'Machine_Emin);
  CZero:     constant Rep := Scal(0);
  COne:      constant Rep := Scal(1);
  CTiny:     constant Rep := Val(Tiny);
  CInfo:     constant Scalar_Info := Info(CZero);
  
  BallZero:  constant Ball := (CZero,RZero);
  BallOne:   constant Ball := (COne,RZero);
  PlusMinus: constant Ball := (CZero,ROne);

  BInfo: constant Scalar_Info := (RepType    => CInfo.RepType,
                                  IsStandard => CInfo.IsStandard,
                                  IsNumeric  => False,
                                  IsComplex  => False,
                                  NReps      => 2);
end Balls;
