with Strings, IPowers, Roots;
use Strings;

pragma Elaborate_All (Strings,IPowers,Roots);

package body Complexs is

  pragma Inline (Compare);
  
  procedure NOP(Dummy: in out Complex) is
    pragma Unreferenced(Dummy);
  begin
    null;
  end NOP;
  
--- basic

  function Info(Dummy: Complex) return Scalar_Info is
    pragma Unreferenced(Dummy);
  begin
    return CInfo;
  end Info;

  function IsSharp(Z: Complex) return Boolean is
  begin
    return IsSharp(Z.Re) and then IsSharp(Z.Im);
  end IsSharp;

  function IsZero(Z: Complex) return Boolean is
  begin
    return IsZero(Z.Re) and then IsZero(Z.Im);
  end IsZero;

  function "="(Z1,Z2: Complex) return Boolean is
  begin
    return (Z1.Re=Z2.Re) and then (Z1.Im=Z1.Im);
  end "=";

  procedure SetZero(Z: in out Complex) is
  begin
    SetZero(Z.Re);
    SetZero(Z.Im);
  end SetZero;

  procedure Copy(Z1: in Complex; Z2: in out Complex) is
  begin
    Copy(Z1.Re,Z2.Re);
    Copy(Z1.Im,Z2.Im);
  end Copy;

  procedure Swap(Z1,Z2: in out Complex) is
  begin
    Swap(Z1.Re,Z2.Re);
    Swap(Z1.Im,Z2.Im);
  end Swap;

--- sets

  function Center0(Z: Complex) return Boolean is
  begin
    return Center0(Z.Re) and then Center0(Z.Im);
  end Center0;

  function Contains0(Z: Complex) return Logical is
  begin
    return Contains0(Z.Re) and Contains0(Z.Im);
  end Contains0;

  function Contains(Z1,Z2: Complex) return Logical is
  begin
    return Contains(Z1.Re,Z2.Re) and Contains(Z1.Im,Z2.Im);
  end Contains;
  
  procedure BallAt0(R: in Flt; Z: in out Complex) is
  begin
    BallAt0(R,Z.Re);
    Copy(Z.Re,Z.Im);
  end BallAt0;
  
  function BallAt0(R: Flt) return Complex is
    Z: Complex;
  begin
    BallAt0(R,Z.Re);
    Copy(Z.Re,Z.Im);
    return Z;
  end BallAt0;
  
  procedure ToErr(Z: in out Complex) is
  begin
    ToErr(Z.Re);
    ToErr(Z.Im);
  end ToErr;
  
  procedure ToErr(Z1: in Complex; Z2: in out Complex) is
  begin
    ToErr(Z1.Re,Z2.Re);
    ToErr(Z1.Im,Z2.Im);
  end ToErr;
  
  function ToErr(Z: Complex) return Complex is
    W: Complex;
  begin
    ToErr(Z.Re,W.Re);
    ToErr(Z.Im,W.Im);
    return W;
  end ToErr;
  
  procedure Center(Z: in out Complex) is
  begin
    Center(Z.Re);
    Center(Z.Im);
  end Center;
  
  procedure Center(Z1: in Complex; Z2: in out Complex) is
  begin
    Center(Z1.Re,Z2.Re);
    Center(Z1.Im,Z2.Im);
  end Center;
  
  function Center(Z: Complex) return Complex is
    W: Complex;
  begin
    Center(Z.Re,W.Re);
    Center(Z.Im,W.Im);
    return W;
  end Center;
  
  procedure ModCenter(Z: in out Complex) is
  begin
    ModCenter(Z.Re);
    ModCenter(Z.Im);
  end ModCenter;
  
  procedure ModCenter(Z1: in Complex; Z2: in out Complex) is
  begin
    ModCenter(Z1.Re,Z2.Re);
    ModCenter(Z1.Im,Z2.Im);
  end ModCenter;
  
  function ModCenter(Z: Complex) return Complex is
    W: Complex;
  begin
    ModCenter(Z.Re,W.Re);
    ModCenter(Z.Im,W.Im);
    return W;
  end ModCenter;
  
  procedure ErrMult(R: in Radius; Z: in out Complex) is
  begin
    ErrMult(R,Z.Re);
    ErrMult(R,Z.Im);
  end ErrMult;
  
  procedure Union(Z1: in Complex; Z2: in out Complex) is
  begin
    Union(Z1.Re,Z2.Re);
    Union(Z1.Im,Z2.Im);
  end Union;
    
  function Union(Z1,Z2: Complex) return Complex is
    Z3: Complex := Z2;
  begin
    Union(Z1.Re,Z3.Re);
    Union(Z1.Im,Z3.Im);
    return Z3;
  end Union;
  
  procedure Intersection(Z1: in Complex; Z2: in out Complex; Empty: out Logical) is
    ER,EI: Logical;
  begin
    Intersection(Z1.Re,Z2.Re,ER);
    Intersection(Z1.Im,Z2.Im,EI);
    Empty := ER or EI;
  end Intersection;
    
--- order
  
  function Sign(Z: Complex) return Integer is
  begin
    if not IsZero(Z.Im) then
      raise Undefined with "Complex.Sign error: argument non-real";
    end if;
    return Sign(Z.Re);
  end Sign;
  
  function Compare(Z1,Z2: Complex) return Integer is
  begin
    if not (IsZero(Z1.Im) and IsZero(Z2.Im)) then
      raise Undefined with "Complex.Compare error: argument non-real";
    end if;
    return Compare(Z1.Re,Z2.Re);
  end Compare;
  
  function "<"(Z1,Z2: Complex) return Boolean is
  begin
    return (Compare(Z1,Z2) < 0);
  end "<";

  function "<="(Z1,Z2: Complex) return Boolean is
  begin
    return (Compare(Z1,Z2) <= 0);
  end "<=";

  function ">="(Z1,Z2: Complex) return Boolean is
  begin
    return (Compare(Z1,Z2) >= 0);
  end ">=";

  function ">"(Z1,Z2: Complex) return Boolean is
  begin
    return (Compare(Z1,Z2) > 0);
  end ">";

  procedure Min(Z1: in Complex; Z2: in out Complex) is
  begin
    if not (IsZero(Z1.Im) and IsZero(Z2.Im)) then
      raise Undefined with "Complex.Min error: argument non-real";
    end if;
    Min(Z1.Re,Z2.Re);
  end Min;
  
  procedure Min(Z1,Z2: in Complex; Z3: in out Complex) is
  begin
    if not (IsZero(Z1.Im) and IsZero(Z2.Im)) then
      raise Undefined with "Complex.Min error: argument non-real";
    end if;
    Min(Z1.Re,Z2.Re,Z3.Re);
    SetZero(Z3.Im);
  end Min;
  
  function Min(Z1,Z2: Complex) return Complex is
    Z3: Complex;
  begin
    Min(Z1,Z2,Z3);
    return Z3;
  end Min;

  procedure Max(Z1: in Complex; Z2: in out Complex) is
  begin
    if not (IsZero(Z1.Im) and IsZero(Z2.Im)) then
      raise Undefined with "Complex.Max error: argument non-real";
    end if;
    Max(Z1.Re,Z2.Re);
  end Max;

  procedure Max(Z1,Z2: in Complex; Z3: in out Complex) is
  begin
    if not (IsZero(Z1.Im) and IsZero(Z2.Im)) then
      raise Undefined with "Complex.Max error: argument non-real";
    end if;
    Max(Z1.Re,Z2.Re,Z3.Re);
    SetZero(Z3.Im);
  end Max;
  
  function Max(Z1,Z2: Complex) return Complex is
    Z3: Complex;
  begin
    Max(Z1,Z2,Z3);
    return Z3;
  end Max;

  function Inf(Z: Complex) return Flt is
  begin
    return Inf(Z.Re);
  end Inf;

  function Sup(Z: Complex) return Flt is
  begin
    return Sup(Z.Re);
  end Sup;

--- addition and multiplication etc.

  procedure Neg(Z: in out Complex) is
  begin
    Neg(Z.Re);
    Neg(Z.Im);
  end Neg;

  procedure Neg(Z1: in Complex; Z2: in out Complex) is
  begin
    Neg(Z1.Re,Z2.Re);
    Neg(Z1.Im,Z2.Im);
  end Neg;

  function "-"(Z: Complex) return Complex is
    W: Complex;
  begin
    Neg(Z.Re,W.Re);
    Neg(Z.Im,W.Im);
    return W;
  end "-";

  procedure Add(I: in Integer; Z: in out Complex) is
  begin
    Add(I,Z.Re);
  end Add;

  procedure Add(Z1: in Complex; Z2: in out Complex) is
  begin
    Add(Z1.Re,Z2.Re);
    Add(Z1.Im,Z2.Im);
  end Add;

  procedure Sum(Z1,Z2: in Complex; Z3: in out Complex) is
  begin
    Sum(Z1.Re,Z2.Re,Z3.Re);
    Sum(Z1.Im,Z2.Im,Z3.Im);
  end Sum;

  function "+"(Z1,Z2: Complex) return Complex is
    Z: Complex;
  begin
    Sum(Z1.Re,Z2.Re,Z.Re);
    Sum(Z1.Im,Z2.Im,Z.Im);
    return Z;
  end "+";

  procedure Sub(Z1: in Complex; Z2: in out Complex) is
  begin
    Sub(Z1.Re,Z2.Re);
    Sub(Z1.Im,Z2.Im);
  end Sub;

  procedure Diff(Z1,Z2: in Complex; Z3: in out Complex) is
  begin
    Diff(Z1.Re,Z2.Re,Z3.Re);
    Diff(Z1.Im,Z2.Im,Z3.Im);
  end Diff;

  function "-"(Z1,Z2: Complex) return Complex is
    W: Complex;
  begin
    Diff(Z1.Re,Z2.Re,W.Re);
    Diff(Z1.Im,Z2.Im,W.Im);
    return W;
  end "-";

  procedure Mult(R: in Flt; Z,Tmp: in out Complex) is
  begin
    Mult(R,Z.Re,Tmp.Re);
    Mult(R,Z.Im,Tmp.Im);
  end Mult;

  procedure Mult(R: in Flt; Z: in out Complex) is
    Tmp: Real;
  begin
    Mult(R,Z.Re,Tmp);
    Mult(R,Z.Im,Tmp);
  end Mult;

  procedure Prod(R: in Flt; Z1: in Complex; Z2: in out Complex) is
  begin
    Prod(R,Z1.Re,Z2.Re);
    Prod(R,Z1.Im,Z2.Im);
  end Prod;

  function "*"(R: Flt; Z: Complex) return Complex is
    W: Complex;
  begin
    Prod(R,Z.Re,W.Re);
    Prod(R,Z.Im,W.Im);
    return W;
  end "*";
  
  procedure AddProd(R: in Flt; Z1: in Complex; Z2,Tmp: in out Complex) is
  begin
    AddProd(R,Z1.Re,Z2.Re,Tmp.Re);
    AddProd(R,Z1.Im,Z2.Im,Tmp.Im);
  end AddProd;

  procedure AddProd(R: in Flt; Z1: in Complex; Z2: in out Complex) is
    Tmp: Real;
  begin
    AddProd(R,Z1.Re,Z2.Re,Tmp);
    AddProd(R,Z1.Im,Z2.Im,Tmp);
  end AddProd;
  
  procedure Mult(Q: in Rational; Z: in out Complex) is
    RN: constant Flt := Flt(Num(Q));
    RD: constant Flt := Flt(Den(Q));
    Tmp: Real;
  begin
    Mult(RN,Z.Re,Tmp);
    Mult(RN,Z.Im,Tmp);
    Div(RD,Z.Re,Tmp);
    Div(RD,Z.Im,Tmp);
  end Mult;
  
  procedure Mult(Z1: in Complex; Z2: in out Complex) is
    W: constant Complex := Z2;
  begin
    Prod(Z1,W,Z2);
  end Mult;

  procedure Prod(Z1,Z2: in Complex; Z3: in out Complex) is
    Tmp: Real;
  begin
    Prod(Z1.Re,Z2.Re,Z3.Re);
    SubProd(Z1.Im,Z2.Im,Z3.Re,Tmp);
    Prod(Z1.Re,Z2.Im,Z3.Im);
    AddProd(Z1.Im,Z2.Re,Z3.Im,Tmp);
  end Prod;

  function "*"(Z1,Z2: Complex) return Complex is
    W: Complex;
  begin
    Prod(Z1,Z2,W);
    return W;
  end "*";

  procedure AddProd(Z1,Z2: in Complex; Z3,Tmp: in out Complex) is
    RTmp: Real renames Tmp.Re;
  begin
    AddProd(Z1.Re,Z2.Re,Z3.Re,RTmp);
    SubProd(Z1.Im,Z2.Im,Z3.Re,RTmp);
    AddProd(Z1.Re,Z2.Im,Z3.Im,RTmp);
    AddProd(Z1.Im,Z2.Re,Z3.Im,RTmp);
  end AddProd;

  procedure AddProd(Z1,Z2: in Complex; Z3: in out Complex) is
    RTmp: Real;
  begin
    AddProd(Z1.Re,Z2.Re,Z3.Re,RTmp);
    SubProd(Z1.Im,Z2.Im,Z3.Re,RTmp);
    AddProd(Z1.Re,Z2.Im,Z3.Im,RTmp);
    AddProd(Z1.Im,Z2.Re,Z3.Im,RTmp);
  end AddProd;
  
  procedure SubProd(Z1,Z2: in Complex; Z3,Tmp: in out Complex) is
    RTmp: Real renames Tmp.Re;
  begin
    SubProd(Z1.Re,Z2.Re,Z3.Re,RTmp);
    AddProd(Z1.Im,Z2.Im,Z3.Re,RTmp);
    SubProd(Z1.Re,Z2.Im,Z3.Im,RTmp);
    SubProd(Z1.Im,Z2.Re,Z3.Im,RTmp);
  end SubProd;

  procedure Div(R: in Flt; Z,Tmp: in out Complex) is
    RTmp: Real renames Tmp.Re;
  begin
    Div(R,Z.Re,RTmp);
    Div(R,Z.Im,RTmp);
  end Div;

  procedure Div(R: in Flt; Z: in out Complex) is
    RTmp: Real;
  begin
    Div(R,Z.Re,RTmp);
    Div(R,Z.Im,RTmp);
  end Div;

  procedure Quot(Z1: in Complex; R: in Flt; Z2: in out Complex) is
  begin
    Quot(Z1.Re,R,Z2.Re);
    Quot(Z1.Im,R,Z2.Im);
  end Quot;

  function "/"(Z: Complex; R: Flt) return Complex is
    W: Complex;
  begin
    Quot(Z.Re,R,W.Re);
    Quot(Z.Im,R,W.Im);
    return W;
  end "/";

  procedure Div(Z1: in Complex; Z2: in out Complex) is
    Z: Complex;
  begin
    Inv(Z1,Z);
    Mult(Z,Z2);
  end Div;

  procedure Quot(Z1,Z2: in Complex; Z3: in out Complex) is
  begin
    Inv(Z2,Z3);
    Mult(Z1,Z3);
  end Quot;

  function "/"(Z1,Z2: Complex) return Complex is
    Z: Complex;
  begin
    Inv(Z2,Z);
    Mult(Z1,Z);
    return Z;
  end "/";

  procedure Inv(Z1: in Complex; Z2: in out Complex) is
    R,Ri: Real;
  begin
    Prod(Z1.Re,Z1.Re,R);
    AddProd(Z1.Im,Z1.Im,R,Ri); -- Ri is Tmp
    Inv(R,Ri);
    Prod(Ri,Z1.Re,Z2.Re);
    Neg(Ri);
    Prod(Ri,Z1.Im,Z2.Im);
  end Inv;

  function Inv(Z: Complex) return Complex is
    W: Complex;
  begin
    Inv(Z,W);
    return W;
  end Inv;
  
-------------------------------------------------
-------------------------------------------------
  package CIP is new IPowers (Scalar => Complex);
-------------------------------------------------
-------------------------------------------------
  
  procedure IPower(I: in Integer; S1: in Complex; S2: in out Complex) renames CIP.IPower;

  function "**"(Z: Complex; I: Integer) return Complex is
    W: Complex;
  begin
    IPower(I,Z,W);
    return W;
  end "**";


--- functions
  
  function IsReal(Z: Complex) return Boolean is
  begin
    return IsZero(Z.Im);
  end IsReal;
  
  procedure Adjoint(Z1: in Complex; Z2: in out Complex) is
  begin
    Copy(Z1.Re,Z2.Re);
    Neg(Z1.Im,Z2.Im);
  end Adjoint;
  
  procedure Adjoint(Z: in out Complex) is
  begin
    Neg(Z.Im);
  end Adjoint;
    
  function Adjoint(Z: Complex) return Complex is
    W: Complex;
  begin
    Copy(Z.Re,W.Re);
    Neg(Z.Im,W.Im);
    return W;
  end Adjoint;
  
  procedure Real_Part(Z: in out Complex) is
  begin
    SetZero(Z.Im);
  end Real_Part;

  procedure Imag_Part(Z: in out Complex) is
  begin
    SetZero(Z.Re);
  end Imag_Part;
  
  procedure Eval0(Z: in out Complex) is
  begin
    Eval0(Z.Re);
    Eval0(Z.Im);
  end Eval0;
  
  procedure Norm(Z1: in Complex; Z2: in out Complex) is
  begin
    Norm(Z1,Z2.Re,Z2.Im);
    SetZero(Z2.Im);
  end Norm;

  function Norm(Z: Complex) return Complex is
    W: Complex;
  begin
    Norm(Z,W.Re,W.Im);
    SetZero(W.Im);
    return W;
  end Norm;

  function MaxNorm(Z: Complex) return Radius is
--- assuming Flt ops are rounded up
    use Flt_EF;
    NX: constant Radius := MaxNorm(Z.Re);
    NY: constant Radius := MaxNorm(Z.Im);
  begin
    if NY=Zero then return NX; end if;
    if NX=Zero then return NY; end if;
    return Sqrt(NX*NX+NY*NY);
  end MaxNorm;
  
  procedure Sqr(Z: in out Complex) is
    Tmp: Real;
    W: Complex;
  begin
    Prod(Z.Re,Z.Im,Tmp);
    Sum(Tmp,Tmp,W.Im);
    Prod(Z.Re,Z.Re,W.Re);
    SubProd(Z.Im,Z.Im,W.Re,Tmp);
    Copy(W,Z);
  end Sqr;
  
  function Sqr(Z: Complex) return Complex is
    Tmp: Real;
    W: Complex;
  begin
    Prod(Z.Re,Z.Im,Tmp);
    Sum(Tmp,Tmp,W.Im);
    Prod(Z.Re,Z.Re,W.Re);
    SubProd(Z.Im,Z.Im,W.Re,Tmp);
    return W;
  end Sqr;
  
  procedure Sqrt(Z1: in Complex; Z2: in out Complex) is
--- complex square root, eispack style
--- branch chosen so that Z2.Re>=0 and sign(Z2.Im)=sign(Z1.Im)
    Re_Sign: constant Integer := Sign(Z1.Re);
    Im_Sign: constant Integer := Sign(Z1.Im);
  begin
    if Im_Sign = 0 then    --- Z1 is real
      if Re_Sign >= 0 then
        Sqrt(Z1.Re,Z2.Re);
        SetZero(Z2.Im);
      else
        Neg(Z1.Re,Z2.Re);
        Sqrt(Z2.Re,Z2.Im);
        SetZero(Z2.Re);
      end if;
    else
      declare
        RR,Tmp: Real;
      begin
        Norm(Z1,Tmp,RR);
        Norm(Z1.Re,RR);
        Add(RR,Tmp);
        Mult(Half,Tmp,RR);
        Sqrt(Tmp,RR);     -- RR := Sqrt(Half*(Norm(Z1)+abs(Z1.Re)));
        if Re_Sign >= 0 then
          Copy(RR,Z2.Re);
        end if;
        if Im_Sign < 0 then
          Neg(RR);
        end if;
        if Re_Sign <= 0 then
          Copy(RR,Z2.Im);
        end if;
        if Re_Sign < 0 then
          Quot(Z1.Im,Z2.Im,Z2.Re);
          Mult(Half,Z2.Re,Tmp);
        end if;
        if Re_Sign > 0 then
          Quot(Z1.Im,Z2.Re,Z2.Im);
          Mult(Half,Z2.Im,Tmp);
        end if;
      end;
    end if;
  end Sqrt;
  
  function Sqrt(Z: Complex) return Complex is
    W: Complex;
  begin
    Sqrt(Z,W);
    return W;
  end Sqrt;
  
  procedure Root(K: in Positive; Z1: in Complex; Z2: in out Complex) is
  begin
    if IsZero(Z1.Im) and then Sign(Z1.Re) >= 0 then
      Root(K,Z1.Re,Z2.Re);
      SetZero(Z2.Im);
    elsif K>2 then
      declare
        Rad,Phi: Real;
        Tmp: Real renames Z2.Re;
      begin
        Polar(Z1,Tmp,Phi);
        Root(K,Tmp,Rad);
        Div(Flt(K),Phi,Tmp);
        UnPolar(Rad,Phi,Z2);
      end;
    elsif K=2 then
      Sqrt(Z1,Z2);
    else
      Copy(Z1,Z2);
    end if;
  end Root;
  
  function Root(K: Positive; Z: Complex) return Complex is
    W: Complex;
  begin
    Root(K,Z,W);
    return W;
  end Root;
  
  procedure Exp(Z1: in Complex; Z2: in out Complex) is
  begin
    if IsZero(Z1.Im) then
      Exp(Z1.Re,Z2.Re);
      SetZero(Z2.Im);
    else
      declare
        R,Tmp: Real;
      begin
        Exp(Z1.Re,R);
        Cos(Z1.Im,Tmp);
        Prod(R,Tmp,Z2.Re);
        Sin(Z1.Im,Tmp);
        Prod(R,Tmp,Z2.Im);
      end;
    end if;
  end Exp;
  
  function Exp(Z: Complex) return Complex is
    W: Complex;
  begin
    Exp(Z,W);
    return W;
  end Exp;
  
  procedure ArcCos(Z1: in Complex; Z2: in out Complex) is
  begin
    if IsZero(Z1.Im) then
      ArcCos(Z1.Re,Z2.Re);
      SetZero(Z2.Im);
    else
      declare
        Tmp: Complex;
      begin
        Prod(Z1,Z1,Tmp);
        Neg(Tmp);
        Add(1,Tmp.Re);
        Sqrt(Tmp,Z2);    -- Z2 := Sqrt(1-Z1*Z1)
        IMult(Z2);
        Sum(Z1,Z2,Tmp);
        Log(Tmp,Z2);     -- Z2 := Log(Z1+i*Z2)
        NegIMult(Z2);
      end;
    end if;
  end ArcCos;
  
  function ArcCos(Z: Complex) return Complex is
    W: Complex;
  begin
    ArcCos(Z,W);
    return W;
  end ArcCos;
  
  procedure ArcSin(Z1: in Complex; Z2: in out Complex) is
  begin
    if IsZero(Z1.Im) then
      ArcSin(Z1.Re,Z2.Re);
      SetZero(Z2.Im);
    else
      declare
        Tmp: Complex;
      begin
        Prod(Z1,Z1,Tmp);
        Neg(Tmp);
        Add(1,Tmp.Re);
        Sqrt(Tmp,Z2);    -- Z2 := Sqrt(1-Z1*Z1)
        Copy(Z1,Tmp);
        IMult(Tmp);
        Add(Z2,Tmp);
        Log(Tmp,Z2);     -- Z2 := Log(i*Z1+Z2)
        NegIMult(Z2);
      end;
    end if;
  end ArcSin;
  
  function ArcSin(Z: Complex) return Complex is
    W: Complex;
  begin
    ArcSin(Z,W);
    return W;
  end ArcSin;
  
  procedure Cos(Z1: in Complex; Z2: in out Complex) is
  begin
    if IsZero(Z1.Im) then
      Cos(Z1.Re,Z2.Re);
      SetZero(Z2.Im);
    else
      declare
        EP,EM,Tmp: Real;
      begin
        Exp(Z1.Im,EP);
        Mult(Half,EP,Tmp);    -- EP0 := Half*Exp(Z.Im);
        Quot(RQuarter,EP,EM); -- EM0 := RQuarter/EP;
        Sum(EM,EP,Tmp);
        Sub(EP,EM);           -- EM := EM0-EP0
        Copy(Tmp,EP);         -- EP := EM0+EP0
        Cos(Z1.Re,Tmp);
        Prod(Tmp,EP,Z2.Re);   -- Cos(Z1.Re)*(EM0+EP0)
        Sin(Z1.Re,Tmp);
        Prod(Tmp,EM,Z2.Im);   -- Sin(Z1.Re)*(EM0-EP0)
      end;
    end if;
  end Cos;
  
  function Cos(Z: Complex) return Complex is
    W: Complex;
  begin
    Cos(Z,W);
    return W;
  end Cos;
  
  procedure Sin(Z1: in Complex; Z2: in out Complex) is
  begin
    if IsZero(Z1.Im) then
      Sin(Z1.Re,Z2.Re);
      SetZero(Z2.Im);
    else
      declare
        EP,EM,Tmp: Real;
      begin
        Exp(Z1.Im,EP);
        Mult(Half,EP,Tmp);    -- EP0 := Half*Exp(Z.Im);
        Quot(RQuarter,EP,EM); -- EM0 := RQuarter/EP;
        Sum(EM,EP,Tmp);
        Sub(EM,EP);           -- EP := EP0-EM0
        Copy(Tmp,EM);         -- EM := EP0+EM0
        Sin(Z1.Re,Tmp);
        Prod(Tmp,EM,Z2.Re);   -- Sin(Z1.Re)*(EP0+EM0)
        Cos(Z1.Re,Tmp);
        Prod(Tmp,EP,Z2.Im);   -- Cos(Z1.Re)*(EP0-EM0)
      end;
    end if;
  end Sin;

  function Sin(Z: Complex) return Complex is
    W: Complex;
  begin
    Sin(Z,W);
    return W;
  end Sin;
  
  procedure Log(Z1: in Complex; Z2: in out Complex) is
    Rad: Real;
  begin
    Polar(Z1.Re,Z1.Im,Rad,Z2.Im);
    Log(Rad,Z2.Re);
  end Log;
    
  function Log(Z: Complex) return Complex is
    Rad: Real;
    W: Complex;
  begin
    Polar(Z.Re,Z.Im,Rad,W.Im);
    Log(Rad,W.Re);
    return W;
  end Log;
  
  procedure Simple_Random(Z: in out Complex) is
  begin
    Simple_Random(Z.Re);
    Simple_Random(Z.Im);
  end Simple_Random;
  
  procedure QPower(Q: in Rational; Z1: in Complex; Z2: in out Complex) is
    N: constant Integer := Num(Q);
    D: constant Integer := Den(Q);
  begin
    if D=1 then
      IPower(N,Z1,Z2);
    elsif N=1 then
      Root(D,Z1,Z2);
    elsif IsZero(Z1.Im) and then Sign(Z1.Re) >= 0 then
      QPower(Q,Z1.Re,Z2.Re);
      SetZero(Z2.Im);
    elsif abs(N)>D then
      raise Not_Implemented with "Complexs.QPower: |Q|>1";
    else
      declare
        Rad,Phi: Real;
        Tmp: Real renames Z2.Re;
      begin
        Polar(Z1,Tmp,Phi);
        QPower(Q,Tmp,Rad);
        Mult(Flt(N),Phi,Tmp);
        Div(Flt(D),Phi,Tmp);
        UnPolar(Rad,Phi,Z2);
      end;
    end if;
  end QPower;
  
--- conversion and i/o
  
  function Approx(Z: Complex) return Flt is
  begin
    return Approx(Z.Re);
  end Approx;
  
  procedure Assign(I: in Integer; Z: in out Complex) is
  begin
    Assign(I,Z.Re);
    SetZero(Z.Im);
  end Assign;
  
  procedure Assign(Q: in Rational; Z: in out Complex) is
  begin
    Assign(Q,Z.Re);
    SetZero(Z.Im);
  end Assign;
  
  procedure Assign(R: in Flt; Z: in out Complex) is
  begin
    Assign(R,Z.Re);
    SetZero(Z.Im);
  end Assign;
  
  function Scal(I: Integer) return Complex is
    Z: Complex;
  begin
    Assign(I,Z.Re);
    SetZero(Z.Im);
    return Z;
  end Scal;

  function Scal(Q: Rational) return Complex is
    Z: Complex;
  begin
    Assign(Q,Z.Re);
    SetZero(Z.Im);
    return Z;
  end Scal;

  function Scal(R: Flt) return Complex is
    Z: Complex;
  begin
    Assign(R,Z.Re);
    SetZero(Z.Im);
    return Z;
  end Scal;
  
  procedure Enclose(R1,R2: in Flt; Z: in out Complex) is
  begin
    Enclose(R1,R2,Z.Re);
    SetZero(Z.Im);
  end Enclose;
    
  function Enclose(R1,R2: Flt) return Complex is
    Z: Complex;
  begin
    Enclose(R1,R2,Z.Re);
    SetZero(Z.Im);
    return Z;
  end Enclose;

  procedure Show1(N: in String; Z: in Complex; Hide0: in Boolean := True) is
  begin
    if IsZero(Z.Im) then
      Show1(N,Z.Re,Hide0);
    else
      Show2(N & "Re Im ",Z.Re,Z.Im,True);
    end if;
  end Show1;

  procedure Show2(N: in String; Z1,Z2: in Complex; Hide0: in Boolean := True) is
  begin
    if IsZero(Z1.Im) and then IsZero(Z2.Im) then
      Show2(N,Z1.Re,Z2.Re,Hide0);
    else
      Show2(N & "Re ",Z1.Re,Z2.Re,True);
      Show2(N & "Im ",Z1.Im,Z2.Im,True);
    end if;
  end Show2;

  procedure Put(F: in File_Type; Z: in Complex; Decimal: in Boolean := False) is
  begin
    Put(F,Z.Re,Decimal);
    Put(F,Z.Im,Decimal);
  end Put;

  procedure Get(F: in File_Type; Z: in out Complex; Decimal: in Boolean := False) is
  begin
    Get(F,Z.Re,Decimal);
    Get(F,Z.Im,Decimal);
  end Get;

  procedure Write(FileName: in String; Z: in Complex; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,Z,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; Z: in out Complex; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,Z,Decimal);
    Close(F);
  end Read;

--- misc

  function Get_Precision(Z: Complex) return Positive is
  begin
    return Get_Precision(Z.Re);
  end Get_Precision;

  procedure Set_Active_Mode(Dummy: in Complex) is
  begin
    Set_Active_Mode(Dummy.Re);
  end Set_Active_Mode;

  procedure Set_Wait_Mode(Dummy: in Complex) is
  begin
    Set_Wait_Mode(Dummy.Re);
  end Set_Wait_Mode;
  
  procedure Free_Cache(Dummy: in Complex) is
    pragma Unreferenced(Dummy);
  begin
    null;
  end Free_Cache;
  
----------------------------------------------
----------------------------------------------
  package SR is new Roots (Scalar => Complex);
----------------------------------------------
----------------------------------------------
  
  procedure Roots2(B,C: in Complex; U1,U2,V: in out Complex) is
    Steps: constant Integer := 64;
  begin
    if IsZero(B.Im) and then IsZero(C.Im) then
      Roots2(B.Re,C.Re,U1.Re,U2.Re,U1.Im);
      Neg(U1.Im,U2.Im);
      SetZero(V);
    else
      SR.NewtonRoots2(B,C,U1,U2,V,Steps,False);
    end if;
  end Roots2;    

  procedure Roots3(B,C,D: in Complex; U0,U1,U2,V: in out Complex) is
    Steps: constant Integer := 64;
  begin
    if IsZero(B.Im) and then IsZero(C.Im) and then IsZero(D.Im) then
      Roots3(B.Re,C.Re,D.Re,U0.Re,U1.Re,U2.Re,U1.Im);
      SetZero(U0.Im);
      Neg(U1.Im,U2.Im);
      SetZero(V);
    else
      SR.NewtonRoots3(B,C,D,U0,U1,U2,V,Steps,False);
    end if;
  end Roots3;
  
--- additional functions
  
  procedure Norm(X,Y: in Real; S,Tmp: in out Real) is
  begin
    if IsZero(Y) then
      Norm(X,S);
    elsif IsZero(X) then
      Norm(Y,S);
    else
      Prod(X,X,S);
      AddProd(Y,Y,S,Tmp);
      Norm(S,Tmp);
      Sqrt(Tmp,S);
    end if;
  end Norm;
  
  procedure Norm(Z: in Complex; S,Tmp: in out Real) is
  begin
    if IsZero(Z.Im) then
      Norm(Z.Re,S);
    elsif IsZero(Z.Re) then
      Norm(Z.Im,S);
    else
      Prod(Z.Re,Z.Re,S);
      AddProd(Z.Im,Z.Im,S,Tmp);
      Norm(S,Tmp);
      Sqrt(Tmp,S);
    end if;
  end Norm;
  
  procedure Norm(Z: in Complex; S: in out Real) is
    Tmp: Real;
  begin
    Norm(Z,S,Tmp);
  end Norm;
  
  function Norm(Z: Complex) return Real is
    S,Tmp: Real;
  begin
    Norm(Z,S,Tmp);
    return S;
  end Norm;
  
  procedure Mult(S: in Real; Z: in out Complex) is
  begin
    Mult(S,Z.Re);
    Mult(S,Z.Im);
  end Mult;

  function "*"(S: Real; Z: Complex) return Complex is
    W: Complex;
  begin
    Prod(S,Z.Re,W.Re);
    Prod(S,Z.Im,W.Im);
    return W;
  end "*";
  
  procedure IMult(Z: in out Complex) is
  begin
    Swap(Z.Re,Z.Im);
    Neg(Z.Re);
  end IMult;
  
  procedure NegIMult(Z: in out Complex) is
  begin
    Swap(Z.Re,Z.Im);
    Neg(Z.Im);
  end NegIMult;
  
  procedure UPolar(X,Y: in Real; Rad,Phi,Tmp: in out Real) is
--- assuming |X| < Y
  begin
    Norm(X,Y,Rad,Tmp);
    Quot(X,Rad,Tmp);
    ArcCos(Tmp,Phi);  -- ArcCos(X/Rad)
  end UPolar;
  
  procedure RPolar(X,Y: in Real; Rad,Phi,Tmp: in out Real) is
--- assuming |Y| < X
  begin
    Norm(X,Y,Rad,Tmp);
    Quot(Y,Rad,Tmp);
    ArcSin(Tmp,Phi);  -- ArcSin(Y/Rad)
  end RPolar;
    
  procedure Polar(X,Y: in Real; Rad,Phi: in out Real) is
    Sig: Integer := Sign(Y);
    Tmp: Real;
  begin
    if Sig<0 then               -- lower half-plane
      Neg(Y,Tmp);
      Polar(X,Tmp,Rad,Phi);
      Neg(Phi);
    elsif MaxNorm(X)>MaxNorm(Y) then
      Sig := Sign(X);
      if Sig >= 0 then          -- right half-plane
        RPolar(X,Y,Rad,Phi,Tmp);
      else                      -- left half-plane
        declare
          B: Real;
        begin
          Neg(X,B);
          RPolar(B,Y,Rad,Phi,Tmp);
          Neg(Phi);
          ArcCos(RZero,B);      -- Pi/2
          AddProd(Two,B,Phi,Tmp);
        end;
      end if;
    elsif Sig>0 then
      UPolar(X,Y,Rad,Phi,Tmp);  -- upper half plane
    else
      SetZero(Rad);
      SetZero(Phi);
    end if;
  end Polar;
  
  procedure Polar(Z: in Complex; Rad,Phi: in out Real) is
  begin
    Polar(Z.Re,Z.Im,Rad,Phi);
  end Polar;
  
  procedure UnPolar(Rad,Phi: in Real; Z: in out Complex) is
    Tmp: Real;
  begin
    Cos(Phi,Tmp);
    Prod(Rad,Tmp,Z.Re);
    Sin(Phi,Tmp);
    Prod(Rad,Tmp,Z.Im);
  end UnPolar;
  
  function UnPolar(Rad,Phi: Real) return Complex is
    Tmp: Real;
    Z: Complex;
  begin
    Cos(Phi,Tmp);
    Prod(Rad,Tmp,Z.Re);
    Sin(Phi,Tmp);
    Prod(Rad,Tmp,Z.Im);
    return Z;
  end UnPolar;
  
begin

  if RInfo.IsComplex then
    raise Undefined with "Complexs error: expect real Scalar";
  end if;

end Complexs;
