with Std_Floats, Std_Balls, Std_Ball_Scalars;

pragma Elaborate_All (Std_Floats,Std_Balls,Std_Ball_Scalars);

generic
  
  with procedure PutStd(S: in R_Scalar; C,R: out Std_Floats.LLFloat) is <>;
  with procedure GetStd(C,R: in Std_Floats.LLFloat; S: in out R_Scalar) is <>;
  
package FHN_Homoclinic.Faster is
  
  pragma Elaborate_Body;
  
  package SBC is new Std_Ball_Scalars (SPDeg1 => SPDeg1);
  use SBC;
  
--------- Real Numbers
  package R_AHome renames Std_Balls;
  subtype R_AScalar is Std_Scalars.Real_Number;
  package R_AMatrices renames Std_Scalars.Real_Matrices;
  subtype R_AMatrix is Std_Scalars.Real_Matrix;
  package R_ATriangulars renames Std_Scalars.Real_Triangulars;
  subtype R_ATriangular is Std_Scalars.Real_Triangular;
  package R_AVectors renames Std_Scalars.Real_Vectors;
  subtype R_AVector is Std_Scalars.Real_Vector;

--------- Real Taylor1
  package RT1_AHome renames Std_Scalars.Real_Taylors1;
  subtype RT1_AScalar is Std_Scalars.Real_Taylor1;
  package RT1_AMatrices renames Std_Scalars.Real_Taylor1_Matrices;
  subtype RT1_AMatrix is Std_Scalars.Real_Taylor1_Matrix;
  package RT1_ATriangulars renames Std_Scalars.Real_Taylor1_Triangulars;
  subtype RT1_ATriangular is Std_Scalars.Real_Taylor1_Triangular;
  package RT1_AVectors renames Std_Scalars.Real_Taylor1_Vectors;
  subtype RT1_AVector is Std_Scalars.Real_Taylor1_Vector;

--------- Complex Numbers
  package C_AHome renames Std_Scalars.Complex_Numbers;
  subtype C_AScalar is Std_Scalars.Complex_Number;
  package C_AMatrices renames Std_Scalars.Complex_Matrices;
  subtype C_AMatrix is Std_Scalars.Complex_Matrix;
  package C_ATriangulars renames Std_Scalars.Complex_Triangulars;
  subtype C_ATriangular is Std_Scalars.Complex_Triangular;
  package C_AVectors renames Std_Scalars.Complex_Vectors;
  subtype C_AVector is Std_Scalars.Complex_Vector;

--------- Complex Taylor1
  package CT1_AHome renames Std_Scalars.Complex_Taylors1;
  subtype CT1_AScalar is Std_Scalars.Complex_Taylor1;
  package CT1_AMatrices renames Std_Scalars.Complex_Taylor1_Matrices;
  subtype CT1_AMatrix is Std_Scalars.Complex_Taylor1_Matrix;
  package CT1_ATriangulars renames Std_Scalars.Complex_Taylor1_Triangulars;
  subtype CT1_ATriangular is Std_Scalars.Complex_Taylor1_Triangular;
  package CT1_AVectors renames Std_Scalars.Complex_Taylor1_Vectors;
  subtype CT1_AVector is Std_Scalars.Complex_Taylor1_Vector;
  
------------------------------------------------------------------
--- other complex types ------------------------------------------
------------------------------------------------------------------
  
  use C_AHome;
  package C_APolynoms2 is new Polynoms2(PDeg => Deg_Taylor2,
                                        Scalar => C_AScalar,
                                        SM => C_AMatrices,
                                        SV => C_AVectors,
                                        ST => C_ATriangulars);
  
  subtype C_APoly2 is C_APolynoms2.Polynom2;
  subtype C_AEVector is C_APolynoms2.EVector;
  
---------------------------------------------------
  
  use CT1_AHome;
  package CT1_APolynoms2 is new Polynoms2(PDeg => Deg_Taylor2,
                                        Scalar => CT1_AScalar,
                                        SM => CT1_AMatrices,
                                        SV => CT1_AVectors,
                                        ST => CT1_ATriangulars);
  
  subtype CT1_APoly2 is CT1_APolynoms2.Polynom2;
  subtype CT1_AEVector is CT1_APolynoms2.EVector;
  
------------------------------------------------------------------
--- the real stuff -----------------------------------------------
------------------------------------------------------------------
  
--- these will be substituted for original versions via A_Prod_Err
  procedure Fast_Prod_Err(C1,C2: in C_Poly2; D1,D2: in Natural; R: in Radius; E: in out C_EVector);
  procedure Fast_Prod_Err(C1,C2: in CT1_Poly2; D1,D2: in Natural; R: in Radius; E: in out CT1_EVector);
  
private
  
------------------------------------------------------------------
--- conversion of scalar types -----------------------------------
------------------------------------------------------------------

  procedure S2A(S: in R_Scalar; A: in out R_AScalar);
  procedure A2S(A: in R_AScalar; S: in out R_Scalar);
  procedure S2A(S: in R_Vector; A: in out R_AVector);
  procedure A2S(A: in R_AVector; S: in out R_Vector);
  procedure S2A(S: in R_Triangular; A: in out R_ATriangular);
  procedure A2S(A: in R_ATriangular; S: in out R_Triangular);
  
  procedure S2A(S: in C_Scalar; A: in out C_AScalar);
  procedure A2S(A: in C_AScalar; S: in out C_Scalar);
  procedure S2A(S: in C_Vector; A: in out C_AVector);
  procedure A2S(A: in C_AVector; S: in out C_Vector);
  procedure S2A(S: in C_Triangular; A: in out C_ATriangular);
  procedure A2S(A: in C_ATriangular; S: in out C_Triangular);
  
  procedure S2A(S: in RT1_Scalar; A: in out RT1_AScalar);
  procedure A2S(A: in RT1_AScalar; S: in out RT1_Scalar);
  procedure S2A(S: in RT1_Vector; A: in out RT1_AVector);
  procedure A2S(A: in RT1_AVector; S: in out RT1_Vector);
  procedure S2A(S: in RT1_Triangular; A: in out RT1_ATriangular);
  procedure A2S(A: in RT1_ATriangular; S: in out RT1_Triangular);
  
  procedure S2A(S: in CT1_Scalar; A: in out CT1_AScalar);
  procedure A2S(A: in CT1_AScalar; S: in out CT1_Scalar);
  procedure S2A(S: in CT1_Vector; A: in out CT1_AVector);
  procedure A2S(A: in CT1_AVector; S: in out CT1_Vector);
  procedure S2A(S: in CT1_Triangular; A: in out CT1_ATriangular);
  procedure A2S(A: in CT1_ATriangular; S: in out CT1_Triangular);
  
end FHN_Homoclinic.Faster;
