with Strings, Messages;
use Strings, Messages;

pragma Elaborate_All (Strings,Messages);

package body FHN_Homoclinic is

------------------------------------------------------------------
--- conversion of taylor types -----------------------------------
------------------------------------------------------------------

  procedure Set_Real(R: in R_Taylor1; C: in out C_Taylor1) is
    use Gen_Scalars;
  begin
    Set_Real(R.C,C.C);
    C.F := R.F;
    C.R := R.R;
  end Set_Real;

  procedure Set_Imag(R: in R_Taylor1; C: in out C_Taylor1) is
    use Gen_Scalars;
  begin
    Set_Imag(R.C,C.C);
    C.F := R.F;
    C.R := R.R;
  end Set_Imag;

  procedure Set_Complex(Rx,Ry: in R_Taylor1; C: in out C_Taylor1) is
    use Gen_Scalars;
  begin
    Set_Real(Rx.C,C.C);
    Set_Imag(Ry.C,C.C);
    C.F := IMin(Rx.F,Ry.F);
    C.R := RMin(Rx.R,Ry.R);
  end Set_Complex;

  procedure Real_Part(C: in C_Taylor1; R: in out R_Taylor1) is
    use Gen_Scalars;
  begin
    Real_Part(C.C,R.C);
    R.F := C.F;
    R.R := C.R;
  end Real_Part;

  procedure Imag_Part(C: in C_Taylor1; R: in out R_Taylor1) is
    use Gen_Scalars;
  begin
    Imag_Part(C.C,R.C);
    R.F := C.F;
    R.R := C.R;
  end Imag_Part;
  
  function ScalRho(P: in RT1_Taylor1) return Radius is
    use Gen_Scalars;
  begin
    return Rho(P.C);
  end ScalRho;
  
  function ScalRho(P: in CT1_Taylor1) return Radius is
    use Gen_Scalars;
  begin
    return Rho(P.C);
  end ScalRho;
  
  function ScalRho(P: in CT1_Taylor2) return Radius is
    use Gen_Scalars;
  begin
    return Rho(P.C);
  end ScalRho;
  
  function ScalRho(P: in RT1_Taylor1_Vector) return Radius is
    R: Radius := ScalRho(P(P'First));
    RN: Radius;
  begin
    for N in P'First+1 .. P'Last loop
      RN := ScalRho(P(N));
      if RN /= R then
        if Verbosity>0 then
          Show2("FHN_Homoclinic.ScalRho warning: different radii in RT1_Taylor1_Vector",RN,R);
        end if;
        R := RMin(RN,R);
      end if;
    end loop;
    return R;
  end ScalRho;
  
  function ScalRho(P: in CT1_Taylor1_Vector) return Radius is
    R: Radius := ScalRho(P(P'First));
    RN: Radius;
  begin
    for N in P'First+1 .. P'Last loop
      RN := ScalRho(P(N));
      if RN /= R then
        if Verbosity>0 then
          Show2("FHN_Homoclinic.ScalRho warning: different radii in CT1_Taylor1_Vector",RN,R);
        end if;
        R := RMin(RN,R);
      end if;
    end loop;
    return R;
  end ScalRho;
  
  function ScalRho(P: in CT1_Taylor2_Vector) return Radius is
    R: Radius := ScalRho(P(P'First));
    RN: Radius;
  begin
    for N in P'First+1 .. P'Last loop
      RN := ScalRho(P(N));
      if RN /= R then
        if Verbosity>0 then
          Show2("FHN_Homoclinic.ScalRho warning: different radii in CT1_Taylor2_Vector",RN,R);
        end if;
        R := RMin(RN,R);
      end if;
    end loop;
    return R;
  end ScalRho;
  
  procedure SetScalRho(R: in Radius; P: in out RT1_Taylor1; Check: in Boolean := True) is
    use Gen_Scalars;
  begin
    SetRho(R,P.C,Check);
  end SetScalRho;
  
  procedure SetScalRho(R: in Radius; P: in out CT1_Taylor1; Check: in Boolean := True) is
    use Gen_Scalars;
  begin
    SetRho(R,P.C,Check);
  end SetScalRho;
  
  procedure SetScalRho(R: in Radius; P: in out CT1_Taylor2; Check: in Boolean := True) is
    use Gen_Scalars;
  begin
    SetRho(R,P.C,Check);
  end SetScalRho;
  
  procedure SetScalRho(R: in Radius; P: in out RT1_Taylor1_Vector; Check: in Boolean := True) is
  begin
    for N in P'Range loop SetScalRho(R,P(N),Check); end loop;
  end SetScalRho;
  
  procedure SetScalRho(R: in Radius; P: in out CT1_Taylor1_Vector; Check: in Boolean := True) is
  begin
    for N in P'Range loop SetScalRho(R,P(N),Check); end loop;
  end SetScalRho;
  
  procedure SetScalRho(R: in Radius; P: in out CT1_Taylor2_Vector; Check: in Boolean := True) is
  begin
    for N in P'Range loop SetScalRho(R,P(N),Check); end loop;
  end SetScalRho;
  
  procedure Assign(R1: in R_Taylor1; R2: in out RT1_Taylor1) is
    use Gen_Scalars;
  begin
    Assign(R1.C,R2.C);
    R2.F := R1.F;
    R2.R := R1.R;
  end Assign;

  procedure Assign(R: in R_Taylor1; C: in out CT1_Taylor1) is
    use Gen_Scalars;
  begin
    Assign(R.C,C.C);
    C.F := R.F;
    C.R := R.R;
  end Assign;

  procedure Assign(C1: in C_Taylor2; C2: in out CT1_Taylor2) is
    use Gen_Scalars;
  begin
    Assign(C1.C,C2.C);
    C2.F := C1.F;
    C2.R := C1.R;
  end Assign;

  procedure Eval(P: in RT1_Taylor1; R: in R_Scalar; Q: out R_Taylor1) is
    use Gen_Scalars;
  begin
    Val(P.C,R,Q.C);
    Q.F := P.F;
    Q.R := P.R;
  end Eval;

  procedure Eval(P: in CT1_Taylor1; C: in C_Scalar; Q: out C_Taylor1) is
    use Gen_Scalars;
  begin
    Val(P.C,C,Q.C);
    Q.F := P.F;
    Q.R := P.R;
  end Eval;

  procedure Eval(P: in CT1_Taylor2; C: in C_Scalar; Q: out C_Taylor2) is
    use Gen_Scalars;
  begin
    Val(P.C,C,Q.C);
    Q.F := P.F;
    Q.R := P.R;
  end Eval;

  procedure Eval(P: in RT1_Taylor1_Vector; R: in R_Scalar; Q: out R_Taylor1_Vector) is
  begin
    for N in Q'Range loop Eval(P(N),R,Q(N)); end loop;
  end Eval;

  procedure Eval(P: in CT1_Taylor1_Vector; C: in C_Scalar; Q: out C_Taylor1_Vector) is
  begin
    for N in Q'Range loop Eval(P(N),C,Q(N)); end loop;
  end Eval;

  procedure Eval(P: in CT1_Taylor2_Vector; C: in C_Scalar; Q: out C_Taylor2_Vector) is
  begin
    for N in Q'Range loop Eval(P(N),C,Q(N)); end loop;
  end Eval;

  procedure EvalDer(P: in RT1_Taylor1; R: in R_Scalar; Q: out R_Taylor1) is
    use Gen_Scalars;
  begin
    ValDer(P.C,R,Q.C);
    Q.F := P.F;
    Q.R := P.R;
  end EvalDer;

  procedure EvalDer(P: in CT1_Taylor1; C: in C_Scalar; Q: out C_Taylor1) is
    use Gen_Scalars;
  begin
    ValDer(P.C,C,Q.C);
    Q.F := P.F;
    Q.R := P.R;
  end EvalDer;

  procedure EvalDer(P: in CT1_Taylor2; C: in C_Scalar; Q: out C_Taylor2) is
    use Gen_Scalars;
  begin
    ValDer(P.C,C,Q.C);
    Q.F := P.F;
    Q.R := P.R;
  end EvalDer;

  procedure EvalDer(P: in RT1_Taylor1_Vector; R: in R_Scalar; Q: out R_Taylor1_Vector) is
  begin
    for N in Q'Range loop EvalDer(P(N),R,Q(N)); end loop;
  end EvalDer;

  procedure EvalDer(P: in CT1_Taylor1_Vector; C: in C_Scalar; Q: out C_Taylor1_Vector) is
  begin
    for N in Q'Range loop EvalDer(P(N),C,Q(N)); end loop;
  end EvalDer;

  procedure EvalDer(P: in CT1_Taylor2_Vector; C: in C_Scalar; Q: out C_Taylor2_Vector) is
  begin
    for N in Q'Range loop EvalDer(P(N),C,Q(N)); end loop;
  end EvalDer;

------------------------------------------------------------------
--- precision etc ------------------------------------------------
------------------------------------------------------------------

  procedure Set_Scalar_Precision(Binary_Digs: in Natural := 0) is
  begin
    if Binary_Digs>0 then
      Scalar_Precision := Binary_Digs;
    end if;
    Reps.Set_Precision(Scalar_Precision);
  end Set_Scalar_Precision;

  procedure Set_Precisions(Binary_Digs: in Positive) is
  begin
    Set_Scalar_Precision(Binary_Digs);
  end Set_Precisions;

  function Effective_Precision(Binary_Digs: Positive) return Positive is
    use Reps;
    SP: constant Positive := Scalar_Precision;
    EP: Positive;
  begin
    Reps.Set_Precision(Binary_Digs);
    EP := Reps.Get_Precision(Reps.Pi); -- involves a computation
    Reps.Set_Precision(SP);
    return EP;
  end Effective_Precision;

  procedure Set_SRad1(R: in Radius) is
  begin
    RT1_Home.Rad  := R;
    CT1_Home.Rad  := R;
  end Set_SRad1;

  procedure Set_Newton_Parallel(P: in Boolean := True) is
  begin
    Newton_Parallel := P;
    R_Taylors1.Newton_Parallel := P;
    C_Taylors1.Newton_Parallel := P;
    RT1_Taylors1.Newton_Parallel := P;
    CT1_Taylors1.Newton_Parallel := P;
    C_Taylors2.Newton_Parallel := P;
    CT1_Taylors2.Newton_Parallel := P;
  end Set_Newton_Parallel;

  procedure Set_Taylor2_Tasks(N: in Positive := 1) is
  begin
    Taylor2_Tasks := N;
    C_Taylors2.Prod_Tasks := N;
    CT1_Taylors2.Prod_Tasks := N;
    if N>1 then
      R_Taylors1.Prod_Parallel := True;
      C_Taylors1.Prod_Parallel := True;
      RT1_Taylors1.Prod_Parallel := True;
      CT1_Taylors1.Prod_Parallel := True;
    end if;
  end Set_Taylor2_Tasks;

  procedure Show_Param is
  begin
    if Trunc then
      Show0("Scalar type: Numeric");
    else
      Show0("Scalar type: Ball");
    end if;
    Show0("Rep type: " & Info(R_Zero).RepType);
    Show1("S-Bits =",Effective_Precision(Scalar_Precision));
    Show1("Taylor2_Tasks =",Taylor2_Tasks);
    Show1("Newton_Parallel: ",Newton_Parallel);
    Show1("SPDeg1 =",SPDeg1);
    Show1("Deg_Taylor1  =",Deg_Taylor1);
    Show1("Deg_Taylor2  =",Deg_Taylor2);
    Show1("Verbosity =",Verbosity);
  end Show_Param;

------------------------------------------------------------------
--- the real stuff -----------------------------------------------
------------------------------------------------------------------

  function Input_Prefix return String is
  begin
    if Trunc then
      return Input_Dir & "approx-";
    else
      return Input_Dir;
    end if;
  end Input_Prefix;

  function Output_Prefix return String is
  begin
    if Trunc then
      return Output_Dir & "approx-";
    else
      return Output_Dir;
    end if;
  end Output_Prefix;

  function CApprox return R_Scalar is
    C: R_Scalar;
  begin
    Read(Input_Dir & "approx-c",C);
    Center(C);
    return C;
  end CApprox;
  
  function SApprox return C_Scalar is
    S: C_Scalar;
  begin
    Read(Input_Dir & "approx-s",S);
    Center(S);
    return S;
  end SApprox;
  
  function MinDistRPlus(Z: C_Scalar) return Flt is
--- lower bound on dist(Z,R_plus)      
    use Gen_Scalars;
    SupX,SupY,InfY: Flt; 
    X,Y: R_Scalar;  
  begin
    Real_Part(Z,X);
    Imag_Part(Z,Y);
    SupX := Sup(X);
    SupY := Sup(Y);
    if SupY >= Zero then
      InfY := Inf(Y);
    else
      InfY := -SupY;
      SupY := -Inf(Y);
    end if;
    if InfY > Zero then
      if SupX >= Zero then return InfY; end if;
      return RMax(InfY,-SupX);  --- not optimal
    end if;
    if SupX >= Zero then return Zero; end if;
    return -SupX;
  end MinDistRPlus;
  
  procedure Val_HalfCircle(R: in Radius; N: in Positive; K: in Integer; S: in CT1_Scalar; Z,W: in out C_Scalar) is
--- value of S on a disk centered at R*Exp(iA), A=Pi*K/(2N+1)
    use Gen_Scalars;
    A,X,Y: R_Scalar;
    E: C_Scalar;
    T: CT1_Scalar;
  begin
    if abs(K)>N then
      raise Sorry with "FHN_Homoclinic.Val_HalfCircle: expect |K| <= N";
    end if;
    Quot(R_Pi,Flt(4*N+2),A);
    BallAt0(MaxNorm(A),E);    -- size of disk
    Mult(Flt(2*K),A);         -- Pi*K/(2N+1)
    Cos(A,X);
    Sin(A,Y);
    Set_Complex(X,Y,Z);
    Add(E,Z);
    Mult(R,Z);
    Center(S,T);
    Val(T,Z,W);
    ModCenter(S,T);
    BallAt0(MaxNorm(R,T),E);
    Add(E,W);
  end Val_HalfCircle;
  
  procedure Write_RT1_Ws0(FileName: in String; R_C: in R_Scalar; R: in Radius) is
    use RT1_Homoclinic1, Gen_Scalars;
    Tr: constant Tracer := Starting("Write_RT1_Ws0");
    RT1_C: RT1_Scalar;
  begin
    Set_SRad1(R);
    Affine(R_C,Scal(1),RT1_C);
    Write_Ws(FileName,RT1_C);
  end Write_RT1_Ws0;
  
  procedure Write_CT1_Unstable0(FileName: in String; R_C: in R_Scalar; R0,R: in Radius) is
    use CT1_Taylor2_Vectors, CT1_Homoclinic2, Gen_Scalars;
    Tr: constant Tracer := Starting("Write_CT1_Unstable0");
    C_C: C_Scalar;
    CT1_C: CT1_Scalar;
    Phi: CT1_Taylor2_Vector3;
  begin
    Set_SRad1(R);
    Set_Real(R_C,C_C);
    Affine(C_C,Scal(1),CT1_C);
    Local_Unstable(CT1_C,R0,Phi);
    Write(FileName,Phi);
  end Write_CT1_Unstable0;

  procedure Write_R_Stable(FileName: in String; Y: in FltVec; R_C: in R_Scalar) is
--- uses RT1_Scalars for accuracy
    use Gen_Scalars;
    Tr: constant Tracer := Starting("Write_R_Stable");
  begin
    if Trunc then
      R_Homoclinic1.Write_Stable(FileName,Y,R_C);
    else
      declare
        use RT1_Homoclinic1;
        R_CC: constant R_Scalar := Center(R_C);
        R_CB: constant R_Scalar := ModCenter(R_C);
        Rad: constant Radius := MaxNorm(R_CB);
        RT1_C: RT1_Scalar;
        RT1_W: RT1_Vector3;
        R_Phi: R_Taylor1_Vector3;
        F: File_Type;
        procedure Put_Next(T: in Flt; RT1_Phi: in RT1_Taylor1_Vector3) is
          use R_Taylor1_Vectors;
        begin
          Put(F,T);
          Eval(RT1_Phi,R_CB,R_Phi);
          Put(F,R_Phi);
        end Put_Next;
        procedure Put_R_Stable is new Follow_Stable (Inspect => Put_Next);
      begin
        Set_SRad1(Rad);
        Affine(R_CC,Scal(1),RT1_C);
        Create(F,Out_File,Output_Prefix & "r_stable");
        Put(F,Y'Last+1);
        Put_R_Stable(RT1_C,Y,RT1_W);
        Close(F);
        if Verbosity>0 then Show0("Wrote " & Output_Prefix & "r_stable"); end if;
      end;
    end if;
  end Write_R_Stable;
  
  procedure Write_C_Unstable(FileName: in String; R_C: in R_Scalar; R0: in Radius) is
--- uses CT1_Scalars for accuracy
    use Gen_Scalars;
    Tr: constant Tracer := Starting("Write_C_Unstable");
    C_C: C_Scalar;
    C_Phi: C_Taylor2_Vector3;
  begin
    Set_Real(R_C,C_C);
    if Trunc then
      C_Homoclinic2.Local_Unstable(C_C,R0,C_Phi);
    else
      declare
        use CT1_Taylor2_Vectors, CT1_Homoclinic2;
        CT1_C: CT1_Scalar;
        CT1_Phi: CT1_Taylor2_Vector3;
        C_CC: constant C_Scalar := Center(C_C);
        C_CB: constant C_Scalar := ModCenter(C_C);
        Rad: constant Radius := MaxNorm(C_CB);
      begin
        Set_SRad1(Rad);
        Affine(C_CC,Scal(1),CT1_C);
        Local_Unstable(CT1_C,R0,CT1_Phi);
        Eval(CT1_Phi,C_CB,C_Phi);
      end;
    end if;
    C_Taylor2_Vectors.Write(FileName,C_Phi);
  end Write_C_Unstable;
  
  procedure Find_Homoclinic(RT1_Ws: in RT1_Vector3; CT1_Pu: in CT1_Taylor2_Vector3;
                            C: in out R_Scalar; S: in out C_Scalar; Steps: in Natural := 64) is
    use Gen_Scalars;
    Tr: constant Tracer := Starting("Find_Homoclinic");
    C_Pu: C_Taylor2_Vector3;
    U0,U1,U2: C_Vector3;
    W1,W2,W3: R_Vector3;

    procedure F(V: in R_Vector3; W: in out R_Vector3) is
      use R_Vectors, C_MultiTaylors1, C_MultiTaylors2;
      RDC: R_Scalar renames V(1);
      C_Wu: C_Vector3 renames U2;
      R_Ws: R_Vector3 renames W2;
      CDC,CS: C_Scalar;
    begin
      if Verbosity>4 then
        Show1(" C =",C+V(1));
      end if;
      Set_Real(RDC,CDC);
      Eval(CT1_Pu,CDC,C_Pu);
      Set_Complex(V(2),V(3),CS);
      Val(C_Pu,CS,Adjoint(CS),C_Wu);
      Real_Part(C_Wu,W);
      Val(RT1_Ws,RDC,R_Ws);
      Sub(R_Ws,W);
    end F;

    procedure D1F(RDC: in R_Scalar; CDC,CS: in C_Scalar; W: in out R_Vector3) is
      use R_Vectors, C_MultiTaylors1, C_MultiTaylors2;
      C_Wu: C_Vector3 renames U2;
      R_Ws: R_Vector3 renames W2;
    begin
      EvalDer(CT1_Pu,CDC,C_Pu);
      Val(C_Pu,CS,Adjoint(CS),C_Wu);
      Real_Part(C_Wu,W);
      ValDer(RT1_Ws,RDC,R_Ws);
      Sub(R_Ws,W);
    end D1F;

    procedure D23F(RDC: in R_Scalar; CDC,CS: in C_Scalar; Wr,Wi: in out R_Vector3) is
      use C_Vectors, C_MultiTaylors1, C_MultiTaylors2;
    begin
      Eval(CT1_Pu,CDC,C_Pu);
      ValDer(C_Pu,CS,Adjoint(CS),U1,U2);
      Sum(U2,U1,U0);
      Real_Part(U0,Wr);
      Diff(U2,U1,U0);
      Imag_Part(U0,Wi);
    end D23F;

    use R_ScalVectors;

    procedure DF(V: in R_Vector3; A: in out R_Matrix3) is
      RDC: R_Scalar renames V(1);
      CDC,CS: C_Scalar;
    begin
      Set_Real(RDC,CDC);
      Set_Complex(V(2),V(3),CS);
      D1F(RDC,CDC,CS,W1);
      D23F(RDC,CDC,CS,W2,W3);
      SetColumn(1,W1,A);
      SetColumn(2,W2,A);
      SetColumn(3,W3,A);
      if Verbosity>6 then
        R_Vectors.Show1("V: ",V);
        R_Matrices.Show1("A: ",A);
      end if;      
    end DF;

    procedure Find is new FindZero3(F => F, DF => DF);

    Sr,Si: R_Scalar;
    RS: R_Vector3;

  begin
    if Verbosity>2 then
      Show1("RT1_Ws c-radius:",Rho(RT1_Ws));
      Show1("CT1_Pu c-radius:",ScalRho(CT1_Pu));
    end if;
    Real_Part(S,Sr);
    Imag_Part(S,Si);
    SetZero(RS(1));
    Copy(Sr,RS(2));
    Copy(Si,RS(3));
    Norm_Weight := (Two**20,One,One); -- used in Find
    Find(RS,Steps,False,Verbosity);
    Add(RS(1),C);
    Copy(RS(2),Sr);
    Copy(RS(3),Si);
    Set_Complex(Sr,Si,S);
  end Find_Homoclinic;

  procedure Find_Homoclinic(R: in Radius; C: in out R_Scalar; S: in out C_Scalar) is
    use Gen_Scalars, RT1_Vectors, CT1_Taylor2_Vectors;
    Check: constant Boolean := not Testing;
    RT1_Ws: RT1_Vector3;
    CT1_Pu: CT1_Taylor2_Vector3;
  begin
    Read(Input_Prefix & "rt1_ws0",RT1_Ws);
    Read(Input_Prefix & "ct1_unstable0",CT1_Pu);
    Set_SRad1(R);
    SetRho(R,RT1_Ws,Check);
    SetScalRho(R,CT1_Pu,Check);
    Find_Homoclinic(RT1_Ws,CT1_Pu,C,S,0);
    Write(Output_Prefix & "c",C);
    Write(Output_Prefix & "s",S);
    if Verbosity>0 then
      Show1("C: ",C);
      Show1("S: ",S);
    end if;    
  end Find_Homoclinic;

  procedure Iter_Find_Homoclinic(R0,R: in Radius; Iter: in Positive) is
--- used numerically to improve data
    use Gen_Scalars;
    use RT1_Vectors, RT1_Homoclinic1, CT1_Taylor2_Vectors, CT1_Homoclinic2;
    Y: constant FltVec := Times(0);
    RC: R_Scalar;
    CS: C_Scalar;
    RT1_C: RT1_Scalar;
    CT1_C: CT1_Scalar;
    RT1_Ws: RT1_Vector3;
    CT1_Pu: CT1_Taylor2_Vector3;
  begin
    Set_SRad1(R);
    Read(Input_Prefix & "c",RC);
    Read(Input_Prefix & "s",CS);
    Read(Input_Prefix & "rt1_ws0",RT1_Ws);
    Read(Input_Prefix & "ct1_unstable0",CT1_Pu);
    for I in 1 .. Iter loop
      Find_Homoclinic(RT1_Ws,CT1_Pu,RC,CS);
      Write(Output_Prefix & "c",RC);
      Write(Output_Prefix & "s",CS);
      if Verbosity>4 then
        Show1(Strng(I,3) & " C =",RC);
        Put(Current_Output,RC);
        Put(Current_Output,CS);
      end if;
      Affine(RC,Scal(1),RT1_C);
      Stable(RT1_C,Y,RT1_Ws);
      Write(Output_Prefix & "rt1_ws0",RT1_Ws);
      Set_Real(RT1_C,CT1_C);
      Local_Unstable(CT1_C,R0,CT1_Pu);
      Write(Output_Prefix & "ct1_unstable0",CT1_Pu);
    end loop;
  end Iter_Find_Homoclinic;

---------------------------------
  
  procedure C_Spec_Stable(C,Z: in C_Scalar; U: in out C_Vector3) is
--- uses global SPhi
    use Gen_Scalars, C_Homoclinic1;
  begin
    R_Homoclinic1.Read(Input_Prefix & "r_stable",SPhi);
    declare
      procedure Phi_Segment1(L: in Natural; DY: out Flt; Phi2: in out C_Taylor1) is
      begin
        if L=0 then
          DY := SPhi.Y(0);
        else
          DY := SPhi.Y(L)-SPhi.Y(L-1);
        end if;
        Assign(SPhi.P(L)(2),Phi2);
      end Phi_Segment1;
      procedure Spec_Stable1 is new Spec_Stable (Phi_Segment => Phi_Segment1);
    begin
      Spec_Stable1(C,Z,SPhi.Y'Last,U);
    end;
  end C_Spec_Stable;
  
  procedure C_Spec_Local_Unstable(C,Z: in C_Scalar; V: in out C_Vector3) is
--- uses global UPhi
    use Gen_Scalars, C_Taylor2_Vectors, C_Homoclinic2;
    S: C_Scalar;
  begin
    Read(Input_Prefix & "s",S);
    if IsZero(UPhi) then
      Read(Input_Prefix & "c_unstable",UPhi);
    end if;
    Spec_Local_Unstable(C,Z,S,UPhi(2),V);
  end C_Spec_Local_Unstable;
  
  procedure C_Delta(Z: in C_Scalar; D: in out C_Scalar) is
    use Gen_Scalars, C_ScalVectors;
    Tr: constant Tracer := Starting("C_Delta");
    RC: R_Scalar;
    CC: C_Scalar;
    U,V: C_Vector3;
  begin
    Read(Input_Prefix & "c",RC);
    Set_Real(RC,CC);
    C_Spec_Local_Unstable(CC,Z,V);
    C_Spec_Stable(CC,Z,U);
    Dot(U,V,D);
    Mult(Delta_Fac,D);
  end C_Delta;

  procedure C_Delta(X,Y: in Flt; U,V: out Flt) is
--- for plotting etc
    use Gen_Scalars;
    Zr: constant R_Scalar := Scal(X);
    Zi: constant R_Scalar := Scal(Y);
    Wr,Wi: R_Scalar;
    Z,W: C_Scalar;
  begin
    Show2("X,Y: ",X,Y);
    Set_Complex(Zr,Zi,Z);
    C_Delta(Z,W);
    Real_Part(W,Wr);
    Imag_Part(W,Wi);
    U := Approx(Wr);
    V := Approx(Wi);
    Show2("U,V: ",U,V);
  end C_Delta;

---------------------------

  procedure CT1_Spec_Stable(C,Z: in CT1_Scalar; U: in out CT1_Vector3) is
-- uses global SPhi
    use Gen_Scalars, CT1_Homoclinic1;
  begin
    R_Homoclinic1.Read(Input_Prefix & "r_stable",SPhi);
    declare
      procedure Phi_Segment1(L: in Natural; DY: out Flt; Phi2: in out CT1_Taylor1) is
      begin
        if L=0 then
          DY := SPhi.Y(0);
        else
          DY := SPhi.Y(L)-SPhi.Y(L-1);
        end if;
        Assign(SPhi.P(L)(2),Phi2);
      end Phi_Segment1;
      procedure Spec_Stable1 is new Spec_Stable (Phi_Segment => Phi_Segment1);
    begin
      Set_SRad1(Rho(Z));
      Spec_Stable1(C,Z,SPhi.Y'Last,U);
    end;
  end CT1_Spec_Stable;

  procedure CT1_Spec_Local_Unstable(C,Z: in CT1_Scalar; V: in out CT1_Vector3) is
--- uses global UPhi
    use Gen_Scalars, C_Taylor2_Vectors, CT1_Homoclinic2;
    CS: C_Scalar;
    TS: CT1_Scalar;
    Phi2: CT1_Taylor2;
  begin
    Set_SRad1(Rho(Z));
    Read(Input_Prefix & "s",CS);
    Assign(CS,TS);
    if IsZero(UPhi) then
      Read(Input_Prefix & "c_unstable",UPhi);
    end if;
    Assign(UPhi(2),Phi2);
    Spec_Local_Unstable(C,Z,TS,Phi2,V);
  end CT1_Spec_Local_Unstable;
  
  procedure CT1_Delta(Z: in CT1_Scalar; D: in out CT1_Scalar) is
    use Gen_Scalars, CT1_ScalVectors;
    Tr: constant Tracer := Starting("CT1_Delta");
    RC: R_Scalar;
    CC: C_Scalar;
    TC: CT1_Scalar;
    U,V: CT1_Vector3;
  begin
    Set_SRad1(Rho(Z));
    Read(Input_Prefix & "c",RC);
    Set_Real(RC,CC);
    Assign(CC,TC);
    CT1_Spec_Local_Unstable(TC,Z,V);
    CT1_Spec_Stable(TC,Z,U);
    Dot(U,V,D);
    Mult(Delta_Fac,D);
  end CT1_Delta;
  
  procedure CT1_Delta(Z: in C_Scalar; R: in Radius; D: in out CT1_Scalar) is
    CT1_Z: CT1_Scalar;
  begin
    if Verbosity>4 then
      Show1("Z: ",Z);
      Show1("R: ",R);
    end if;
    Set_SRad1(R);
    Affine(Z,Scal(1),CT1_Z);
    SetRho(R,CT1_Z);
    CT1_Delta(CT1_Z,D);
  end CT1_Delta;
  
  procedure Delta_Check_Origin(R: in Radius; D: in out CT1_Scalar; OK: out Boolean) is
    use Paths;
    Tr: constant Tracer := Starting("Delta_Check_Origin");
    NPoints: constant Integer := 64;
    DistR: Radius;
    Z,V: C_Scalar;
    D1: CT1_Scalar;
  begin
    Show1("checking disk |z| <=",R);
    OK := False;
    SetZero(Z);
    CT1_Delta(Z,R,D);
    for K in 0 .. NPoints loop
      Val_HalfCircle(R,NPoints,K,D,Z,V);
      DistR := MinDistRPlus(V);
      if Verbosity>4 then Show1(Strng(K),DistR); end if;
      if DistR=Zero then return; end if;
    end loop;
    Show0("verified: argument of Delta");
    Copy(D,D1);
    DivArg(D1);
    Ran(D1,V);
    if Contains0(V)=False then
      Show0("verified: simple zero");
      OK := True;
    end if;
  end Delta_Check_Origin;
  
  function NZ(N: Natural) return C_Scalar is
    use Gen_Scalars, Paths;
    X,Y: Flt;
    R_X,R_Y: R_Scalar;
    C_Z: C_Scalar;
  begin
    RXY(N,X,Y);
    Assign(X,R_X);
    Assign(Y,R_Y);
    Set_Complex(R_X,R_Y,C_Z);
    return C_Z;
  end NZ;
  
  function Alt_Radius(S: Segment; D: CT1_Scalar) return Natural is
    use Paths;
    V: C_Scalar;
  begin
    for NR in reverse 1 .. S.NR loop
      ValBall(D,N2R(NR),V);
      if MinDistRPlus(V)>Zero then return NR; end if;
    end loop;
    return 0;
  end Alt_Radius;
  
  procedure Write_Germ(S: in Segment; D: in CT1_Scalar) is
    WC: constant Positive := 5;
    WR: constant Positive := 3;
  begin
    if S.Status=2 then
      Write(Output_Prefix & "done." & Strng0(S.NC,WC) & "." & Strng0(S.NR,WR),D);
    elsif S.Status<0 then
      Write(Output_Prefix &  "foo." & Strng0(S.NC,WC) & "." & Strng0(S.NR,WR),D);
    else
      Write(Output_Prefix & "germ." & Strng0(S.NC,WC) & "." & Strng0(S.NR,WR),D);
    end if;
  end Write_Germ;
  
  procedure Delta_Check_Segment(S: in out Segment) is
    use Gen_Scalars, Paths;
    Tr: constant Tracer := Starting("Delta_Check_Segment");
    NC: Natural renames S.NC;
    NR: Natural renames S.NR;
    R: constant Radius := N2R(NR);
    OK: Boolean;
    V: C_Scalar;
    D: CT1_Scalar;
  begin
    Show2("checking segment",NC,NR);
    if NC=0 then
      Delta_Check_Origin(R,D,OK);
    else
      CT1_Delta(NZ(NC),R,D);
      Ran(D,V);
      OK := (MinDistRPlus(V) /= Zero);
    end if;
    if OK then
      S.Status := 2;
--      Write_Germ(S,D);
      if Verbosity>0 then
        Show2("succeeded",NC,NR);
      end if;
    else
      S.Status := -1;
--      Write_Germ(S,D);
      if Verbosity>0 then
        Show2("failed",NC,NR);
        Show2("   try",NC,Alt_Radius(S,D));
      end if;      
    end if;      
  end Delta_Check_Segment;
  
end FHN_Homoclinic;
