with Ada.Text_IO, Logicals, Globals, Param, Ints, Rationals, Flts, Paths;
use Ada.Text_IO, Logicals, Globals, Param, Ints, Rationals, Flts;
with Reps, Scalars, ScalVectors, ScalVectors.Homoclinic;
with Taylors1, MultiTaylors1, MultiTaylors1.Homoclinic;
with Triangulars, Polynoms2, Taylors2, MultiTaylors2, MultiTaylors2.Homoclinic;

pragma Elaborate_All (Ada.Text_IO,Logicals,Globals,Param,Ints,Rationals,Flts);
pragma Elaborate_All (Reps,Scalars,ScalVectors,ScalVectors.Homoclinic);
pragma Elaborate_All (Taylors1,MultiTaylors1,MultiTaylors1.Homoclinic);
pragma Elaborate_All (Triangulars,Polynoms2,Taylors2,MultiTaylors2,MultiTaylors2.Homoclinic);

generic

  Deg_Taylor1: in Positive := 2;
  Deg_Taylor2: in Positive := 2;
  SPDeg1:      in Positive := 2;

-------------------- begin_include scalars Real

  type Real is private;

--- basic
  with function Info(Dummy: Real) return Scalar_Info is <>;
  with function IsSharp(S: Real) return Boolean is <>;
  with function IsZero(S: Real) return Boolean is <>;
  with function "="(S1,S2: Real) return Boolean is <>;
  with procedure SetZero(S: in out Real) is <>;
  with procedure Copy(S1: in Real; S2: in out Real) is <>;
  with procedure Swap(S1,S2: in out Real) is <>;

--- sets
  with function Center0(S: Real) return Boolean is <>;
  with function Contains0(S: Real) return Logical is <>;
  with function Contains(S1,S2: Real) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Real) is <>;
  with function BallAt0(R: Flt) return Real is <>;
  with procedure ToErr(S: in out Real) is <>;
  with procedure ToErr(S1: in Real; S2: in out Real) is <>;
  with function ToErr(S: Real) return Real is <>;
  with procedure Center(S: in out Real) is <>;
  with procedure Center(S1: in Real; S2: in out Real) is <>;
  with function Center(S: Real) return Real is <>;
  with procedure ModCenter(S: in out Real) is <>;
  with procedure ModCenter(S1: in Real; S2: in out Real) is <>;
  with function ModCenter(S: Real) return Real is <>;
  with procedure ErrMult(R: in Radius; S: in out Real) is <>;
  with procedure Union(S1: in Real; S2: in out Real) is <>;
  with function Union(S1,S2: Real) return Real is <>;
  with procedure Intersection(S1: in Real; S2: in out Real; Empty: out Logical) is <>;

--- order
  with function Sign(S: Real) return Integer is <>;
  with function Compare(S1,S2: Real) return Integer is <>;
  with function "<"(S1,S2: Real) return Boolean is <>;
  with function "<="(S1,S2: Real) return Boolean is <>;
  with function ">"(S1,S2: Real) return Boolean is <>;
  with function ">="(S1,S2: Real) return Boolean is <>;
  with procedure Min(S1: in Real; S2: in out Real) is <>;
  with procedure Min(S1,S2: in Real; S3: in out Real) is <>;
  with function Min(S1,S2: Real) return Real is <>;
  with procedure Max(S1: in Real; S2: in out Real) is <>;
  with procedure Max(S1,S2: in Real; S3: in out Real) is <>;
  with function Max(S1,S2: Real) return Real is <>;
  with function Inf(S: Real) return Flt is <>; -- inf of real part
  with function Sup(S: Real) return Flt is <>; -- sup of real part

--- arithmetic
  with procedure Neg(S: in out Real) is <>;
  with procedure Neg(S1: in Real; S2: in out Real) is <>;
  with function "-"(S: Real) return Real is <>;
  with procedure Add(I: in Integer; S: in out Real) is <>;
  with procedure Add(S1: in Real; S2: in out Real) is <>;
  with procedure Sum(S1,S2: in Real; S3: in out Real) is <>;
  with function "+"(S1,S2: Real) return Real is <>;
  with procedure Sub(S1: in Real; S2: in out Real) is <>;
  with procedure Diff(S1,S2: in Real; S3: in out Real) is <>;
  with function "-"(S1,S2: Real) return Real is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Real) is <>;
  with procedure Mult(R: in Flt; S: in out Real) is <>;
  with procedure Prod(R: in Flt; S1: in Real; S2: in out Real) is <>;
  with function "*"(R: Flt; S: Real) return Real is <>;
  with procedure AddProd(R: in Flt; S1: in Real; S2,Tmp: in out Real) is <>;
  with procedure AddProd(R: in Flt; S1: in Real; S2: in out Real) is <>;
  with procedure Mult(Q: in Rational; S: in out Real) is <>;
  with procedure Mult(S1: in Real; S2: in out Real) is <>;
  with procedure Prod(S1,S2: in Real; S3: in out Real) is <>;
  with function "*"(S1,S2: Real) return Real is <>;
  with procedure AddProd(S1,S2: in Real; S3,Tmp: in out Real) is <>;
  with procedure AddProd(S1,S2: in Real; S3: in out Real) is <>;
  with procedure SubProd(S1,S2: in Real; S3,Tmp: in out Real) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Real) is <>;
  with procedure Div(R: in Flt; S: in out Real) is <>;
  with procedure Quot(S1: in Real; R: in Flt; S2: in out Real) is <>;
  with function "/"(S: Real; R: Flt) return Real is <>;
  with procedure Div(S1: in Real; S2: in out Real) is <>;
  with procedure Quot(S1,S2: in Real; S3: in out Real) is <>;
  with function "/"(S1,S2: Real) return Real is <>;
  with procedure Inv(S1: in Real; S2: in out Real) is <>;
  with function Inv(S: Real) return Real is <>;
  with function "**"(S: Real; I: Integer) return Real is <>;

--- fun
  with function IsReal(S: Real) return Boolean is <>;
  with procedure Adjoint(S: in out Real) is <>;
  with procedure Adjoint(S1: in Real; S2: in out Real) is <>;
  with function Adjoint(S: Real) return Real is <>;
  with procedure Real_Part(S: in out Real) is <>;
  with procedure Imag_Part(S: in out Real) is <>;
  with procedure Eval0(S: in out Real) is <>;
  with function "abs"(S: Real) return Real is <>;
  with procedure Norm(S1: in Real; S2: in out Real) is <>;
  with function Norm(S: Real) return Real is <>;
  with function MaxNorm(S: Real) return Radius is <>;
  with procedure Sqr(S: in out Real) is <>;
  with function Sqr(S: Real) return Real is <>;
  with procedure Sqrt(S1: in Real; S2: in out Real) is <>;
  with function Sqrt(S: Real) return Real is <>;
  with procedure Root(K: in Positive; S1: in Real; S2: in out Real) is <>;
  with function Root(K: Positive; S: Real) return Real is <>;
  with procedure Exp(S1: in Real; S2: in out Real) is <>;
  with function Exp(S: Real) return Real is <>;
  with procedure ArcCos(S1: in Real; S2: in out Real) is <>;
  with function ArcCos(S: Real) return Real is <>;
  with procedure ArcSin(S1: in Real; S2: in out Real) is <>;
  with function ArcSin(S: Real) return Real is <>;
  with procedure Cos(S1: in Real; S2: in out Real) is <>;
  with function Cos(S: Real) return Real is <>;
  with procedure Sin(S1: in Real; S2: in out Real) is <>;
  with function Sin(S: Real) return Real is <>;
  with procedure Log(S1: in Real; S2: in out Real) is <>;
  with function Log(S: Real) return Real is <>;
  with procedure Simple_Random(S: in out Real) is <>;
  with procedure QPower(Q: in Rational; S1: in Real; S2: in out Real) is <>;
  with procedure Roots2(B,C: in Real; U1,U2,V: in out Real) is <>;
  with procedure Roots3(B,C,D: in Real; U0,U1,U2,V: in out Real) is <>;

--- conversion and i/o
  with function Approx(S: Real) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Real) is <>;
  with procedure Assign(Q: in Rational; S: in out Real) is <>;
  with procedure Assign(R: in Flt; S: in out Real) is <>;
  with function Scal(I: Integer) return Real is <>;
  with function Scal(Q: Rational) return Real is <>;
  with function Scal(R: Flt) return Real is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Real) is <>;
  with function Enclose(R1,R2: Flt) return Real is <>;
  with procedure Show1(N: in String; S: in Real; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Real; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Real; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Real; Decimal: in Boolean := False) is <>;

--- misc
  with function Get_Precision(R: Real) return Positive is <>;
  with procedure Set_Active_Mode(Dummy: in Real) is <>;
  with procedure Set_Wait_Mode(Dummy: in Real) is <>;
  with procedure Free_Cache(S: in Real) is <>;
-------------------- end_include scalars

  with procedure NOP(Dummy: in out Real) is <>;
  with procedure Write(FileName: in String; R: in Real; Decimal: in Boolean := False) is <>;
  with procedure Read(FileName: in String; R: in out Real; Decimal: in Boolean := False) is <>;
  with procedure PutNum(S: in Real; R: in out Reps.Rep) is <>; -- not used here
  with procedure GetNum(R: in Reps.Rep; S: in out Real) is <>; -- not used here

package FHN_Homoclinic is
  
------------------------------------------------------------------
--- scalar types -------------------------------------------------
------------------------------------------------------------------

  package Gen_Scalars is new Scalars (SPDeg1 => SPDeg1, Real => Real);

-------- Real Numbers
  subtype R_Scalar is Gen_Scalars.Real_Number;
  package R_Matrices renames Gen_Scalars.Real_Matrices;
  subtype R_Matrix is Gen_Scalars.Real_Matrix;
  subtype R_Matrix2 is Gen_Scalars.Real_Matrix2;
  subtype R_Matrix3 is Gen_Scalars.Real_Matrix3;
  package R_Vectors renames Gen_Scalars.Real_Vectors;
  subtype R_Vector is Gen_Scalars.Real_Vector;
  subtype R_Polynom1 is R_Vector;
  subtype R_Vector2 is Gen_Scalars.Real_Vector2;
  subtype R_Vector3 is Gen_Scalars.Real_Vector3;
  package R_Triangulars renames Gen_Scalars.Real_Triangulars;
  subtype R_Triangular is Gen_Scalars.Real_Triangular;
  package R_ScalVectors renames Gen_Scalars.Real_ScalVectors;

  procedure HiErrs(R: in out R_Scalar) renames NOP;

-------- Real Taylor1
  package RT1_Home renames Gen_Scalars.Real_Taylors1;
  subtype RT1_Scalar is Gen_Scalars.Real_Taylor1;
  package RT1_Matrices renames Gen_Scalars.Real_Taylor1_Matrices;
  subtype RT1_Matrix is Gen_Scalars.Real_Taylor1_Matrix;
  subtype RT1_Matrix2 is Gen_Scalars.Real_Taylor1_Matrix2;
  subtype RT1_Matrix3 is Gen_Scalars.Real_Taylor1_Matrix3;
  package RT1_Vectors renames Gen_Scalars.Real_Taylor1_Vectors;
  subtype RT1_Vector is Gen_Scalars.Real_Taylor1_Vector;
  subtype RT1_Polynom1 is RT1_Vector;
  subtype RT1_Vector2 is Gen_Scalars.Real_Taylor1_Vector2;
  subtype RT1_Vector3 is Gen_Scalars.Real_Taylor1_Vector3;
  package RT1_Triangulars renames Gen_Scalars.Real_Taylor1_Triangulars;
  subtype RT1_Triangular is Gen_Scalars.Real_Taylor1_Triangular;
  package RT1_ScalVectors renames Gen_Scalars.Real_Taylor1_ScalVectors;

-------- Complex Numbers
  package C_Home renames Gen_Scalars.Complex_Numbers;
  subtype C_Scalar is Gen_Scalars.Complex_Number;
  package C_Matrices renames Gen_Scalars.Complex_Matrices;
  subtype C_Matrix is Gen_Scalars.Complex_Matrix;
  subtype C_Matrix2 is Gen_Scalars.Complex_Matrix2;
  subtype C_Matrix3 is Gen_Scalars.Complex_Matrix3;
  package C_Vectors renames Gen_Scalars.Complex_Vectors;
  subtype C_Vector is Gen_Scalars.Complex_Vector;
  subtype C_Polynom1 is C_Vector;
  subtype C_Vector2 is Gen_Scalars.Complex_Vector2;
  subtype C_Vector3 is Gen_Scalars.Complex_Vector3;
  package C_Triangulars renames Gen_Scalars.Complex_Triangulars;
  subtype C_Triangular is Gen_Scalars.Complex_Triangular;
  package C_ScalVectors renames Gen_Scalars.Complex_ScalVectors;

  procedure HiErrs(C: in out C_Scalar) renames C_Home.NOP;

-------- Complex Taylor1
  package CT1_Home renames Gen_Scalars.Complex_Taylors1;
  subtype CT1_Scalar is Gen_Scalars.Complex_Taylor1;
  package CT1_Matrices renames Gen_Scalars.Complex_Taylor1_Matrices;
  subtype CT1_Matrix is Gen_Scalars.Complex_Taylor1_Matrix;
  subtype CT1_Matrix2 is Gen_Scalars.Complex_Taylor1_Matrix2;
  subtype CT1_Matrix3 is Gen_Scalars.Complex_Taylor1_Matrix3;
  package CT1_Vectors renames Gen_Scalars.Complex_Taylor1_Vectors;
  subtype CT1_Vector is Gen_Scalars.Complex_Taylor1_Vector;
  subtype CT1_Vector2 is Gen_Scalars.Complex_Taylor1_Vector2;
  subtype CT1_Vector3 is Gen_Scalars.Complex_Taylor1_Vector3;
  package CT1_Triangulars renames Gen_Scalars.Complex_Taylor1_Triangulars;
  subtype CT1_Triangular is Gen_Scalars.Complex_Taylor1_Triangular;
  package CT1_ScalVectors renames Gen_Scalars.Complex_Taylor1_ScalVectors;

------------------------------------------------------------------
--- other real types ---------------------------------------------
------------------------------------------------------------------

  package R_Taylors1 is new Taylors1 (PDeg => Deg_Taylor1,
                                      Scalar => R_Scalar,
                                      SM => R_Matrices,
                                      SV => R_Vectors);
  
  package R_MultiTaylors1 is new MultiTaylors1 (PDeg => Deg_Taylor1,
                                                Scalar => R_Scalar,
                                                SM => R_Matrices,
                                                SV => R_Vectors,
                                                SSV => R_ScalVectors,
                                                ST1 => R_Taylors1);
  
  package R_Homoclinic1 is new R_MultiTaylors1.Homoclinic;
  package R_Taylor1_Vectors renames R_MultiTaylors1.STV;

--- Taylor series in 1 variable whose coefficients are real numbers
  subtype R_Poly1 is R_Taylors1.Poly1;
  subtype R_Taylor1 is R_Taylors1.Taylor1;
  subtype R_Taylor1_Vector is R_Taylor1_Vectors.Vector; --- scalar is R_Scalar
  subtype R_Taylor1_Vector2 is R_MultiTaylors1.Taylor1_Vector2;
  subtype R_Taylor1_Vector3 is R_MultiTaylors1.Taylor1_Vector3;
  subtype R_Simple_Operator1 is R_Homoclinic1.Simple_Operator1;

  subtype R_Taylor1_Vector3_Array is R_Homoclinic1.Taylor1_Vector3_Array;
  subtype R_Curve3 is R_Homoclinic1.Curve3;

------------------------------------------------------------------

  use RT1_Home;
  package RT1_Taylors1 is new Taylors1 (PDeg => Deg_Taylor1,
                                        Scalar => RT1_Scalar,
                                        SM => RT1_Matrices,
                                        SV => RT1_Vectors);
  
  package RT1_MultiTaylors1 is new MultiTaylors1 (PDeg => Deg_Taylor1,
                                                  Scalar => RT1_Scalar,
                                                  SM => RT1_Matrices,
                                                  SV => RT1_Vectors,
                                                  SSV => RT1_ScalVectors,
                                                  ST1 => RT1_Taylors1);
  
  package RT1_Homoclinic1 is new RT1_MultiTaylors1.Homoclinic;
  package RT1_Taylor1_Vectors renames RT1_MultiTaylors1.STV;

--- Taylor series in 1 variable whose coefficients are real Taylor1
  subtype RT1_Poly1 is RT1_Taylors1.Poly1;
  subtype RT1_Taylor1 is RT1_Taylors1.Taylor1;
  subtype RT1_Taylor1_Vector is RT1_Taylor1_Vectors.Vector; --- scalar is RT1_Scalar
  subtype RT1_Taylor1_Vector2 is RT1_MultiTaylors1.Taylor1_Vector2;
  subtype RT1_Taylor1_Vector3 is RT1_MultiTaylors1.Taylor1_Vector3;
  subtype RT1_Simple_Operator1 is RT1_Homoclinic1.Simple_Operator1;

------------------------------------------------------------------
--- other complex types ------------------------------------------
------------------------------------------------------------------

  use C_Home;
  package C_Taylors1 is new Taylors1 (PDeg => Deg_Taylor1,
                                      Scalar => C_Scalar,
                                      SM => C_Matrices,
                                      SV => C_Vectors);
  
  package C_MultiTaylors1 is new MultiTaylors1 (PDeg => Deg_Taylor1,
                                                Scalar => C_Scalar,
                                                SM => C_Matrices,
                                                SV => C_Vectors,
                                                SSV => C_ScalVectors,
                                                ST1 => C_Taylors1);
  
  package C_Homoclinic1 is new C_MultiTaylors1.Homoclinic;
  package C_Taylor1_Vectors renames C_MultiTaylors1.STV;

--- Taylor series in 1 variable whose coefficients are complex numbers
  subtype C_Poly1 is C_Taylors1.Poly1;
  subtype C_Taylor1 is C_Taylors1.Taylor1;
  subtype C_Taylor1_Vector is C_Taylor1_Vectors.Vector; --- scalar is C_Scalar
  subtype C_Taylor1_Vector2 is C_MultiTaylors1.Taylor1_Vector2;
  subtype C_Taylor1_Vector3 is C_MultiTaylors1.Taylor1_Vector3;
  subtype C_Simple_Operator1 is C_Homoclinic1.Simple_Operator1;

  subtype C_Taylor1_Vector3_Array is C_Homoclinic1.Taylor1_Vector3_Array;
  subtype C_Curve3 is C_Homoclinic1.Curve3;

------------------------------------------------------------------
  
  package C_Polynoms2 is new Polynoms2(PDeg => Deg_Taylor2,
                                       Scalar => C_Scalar,
                                       SM => C_Matrices,
                                       SV => C_Vectors,
                                       ST => C_Triangulars);
  
  package C_Taylors2 is new Taylors2 (PDeg => Deg_Taylor2,
                                      Scalar => C_Scalar,
                                      SM => C_Matrices,
                                      SV => C_Vectors,
                                      ST => C_Triangulars,
                                      SP2 => C_Polynoms2);
  
  package C_MultiTaylors2 is new MultiTaylors2 (PDeg => Deg_Taylor2,
                                                Scalar => C_Scalar,
                                                SM => C_Matrices,
                                                SV => C_Vectors,
                                                SSV => C_ScalVectors,
                                                ST => C_Triangulars,
                                                SP2 => C_Polynoms2,
                                                ST2 => C_Taylors2);
  
  package C_Homoclinic2 is new C_MultiTaylors2.Homoclinic;
  package C_Taylor2_Vectors renames C_MultiTaylors2.STV;

--- Taylor series in 2 variable whose coefficients are complex numbers
  subtype C_Poly2 is C_Polynoms2.Polynom2;
  subtype C_EVector is C_Polynoms2.EVector;
  subtype C_Taylor2 is C_Taylors2.Taylor2;
  subtype C_Taylor2_Vector is C_Taylor2_Vectors.Vector; --- scalar is C_Scalar
  subtype C_Taylor2_Vector2 is C_MultiTaylors2.Taylor2_Vector2;
  subtype C_Taylor2_Vector3 is C_MultiTaylors2.Taylor2_Vector3;
  subtype C_Simple_Operator2 is C_Homoclinic2.Simple_Operator2;

------------------------------------------------------------------

  use CT1_Home;
  package CT1_Taylors1 is new Taylors1 (PDeg => Deg_Taylor1,
                                        Scalar => CT1_Scalar,
                                        SM => CT1_Matrices,
                                        SV => CT1_Vectors);
  
  package CT1_MultiTaylors1 is new MultiTaylors1 (PDeg => Deg_Taylor1,
                                                  Scalar => CT1_Scalar,
                                                  SM => CT1_Matrices,
                                                  SV => CT1_Vectors,
                                                  SSV => CT1_ScalVectors,
                                                  ST1 => CT1_Taylors1);
  
  package CT1_Homoclinic1 is new CT1_MultiTaylors1.Homoclinic;
  package CT1_Taylor1_Vectors renames CT1_MultiTaylors1.STV;

--- Taylor series in 1 variable whose coefficients are real Taylor1
  subtype CT1_Taylor1 is CT1_Taylors1.Taylor1;
  subtype CT1_Poly1 is CT1_Taylors1.Poly1;
  subtype CT1_Taylor1_Vector is CT1_Taylor1_Vectors.Vector; --- scalar is CT1_Scalar
  subtype CT1_Taylor1_Vector2 is CT1_MultiTaylors1.Taylor1_Vector2;
  subtype CT1_Taylor1_Vector3 is CT1_MultiTaylors1.Taylor1_Vector3;
  subtype CT1_Simple_Operator1 is CT1_Homoclinic1.Simple_Operator1;

------------------------------------------------------------------
  
  package CT1_Polynoms2 is new Polynoms2(PDeg => Deg_Taylor2,
                                         Scalar => CT1_Scalar,
                                         SM => CT1_Matrices,
                                         SV => CT1_Vectors,
                                         ST => CT1_Triangulars);
  
  package CT1_Taylors2 is new Taylors2 (PDeg => Deg_Taylor2,
                                        Scalar => CT1_Scalar,
                                        SM => CT1_Matrices,
                                        SV => CT1_Vectors,
                                        ST => CT1_Triangulars,
                                        SP2 => CT1_Polynoms2);
                                        
  package CT1_MultiTaylors2 is new MultiTaylors2 (PDeg => Deg_Taylor2,
                                                  Scalar => CT1_Scalar,
                                                  SM => CT1_Matrices,
                                                  SV => CT1_Vectors,
                                                  SSV => CT1_ScalVectors,
                                                  ST => CT1_Triangulars,
                                                  SP2 => CT1_Polynoms2,
                                                  ST2 => CT1_Taylors2);
  
  package CT1_Homoclinic2 is new CT1_MultiTaylors2.Homoclinic;
  package CT1_Taylor2_Vectors renames CT1_MultiTaylors2.STV; --- scalar is CT1_Scalar

--- Taylor series in 2 variable whose coefficients are complex Taylor1
  subtype CT1_Poly2 is CT1_Polynoms2.Polynom2;
  subtype CT1_EVector is CT1_Polynoms2.EVector;
  subtype CT1_Taylor2 is CT1_Taylors2.Taylor2;
  subtype CT1_Taylor2_Vector is CT1_Taylor2_Vectors.Vector;
  subtype CT1_Taylor2_Vector2 is CT1_MultiTaylors2.Taylor2_Vector2;
  subtype CT1_Taylor2_Vector3 is CT1_MultiTaylors2.Taylor2_Vector3;
  subtype CT1_Simple_Operator2 is CT1_Homoclinic2.Simple_Operator2;

------------------------------------------------------------------
--- FHN types ----------------------------------------------------
------------------------------------------------------------------

  package R_ScalVectors_Homoclinic is new R_ScalVectors.Homoclinic;
  package RT1_ScalVectors_Homoclinic is new RT1_ScalVectors.Homoclinic;
  package C_ScalVectors_Homoclinic is new C_ScalVectors.Homoclinic;
  package CT1_ScalVectors_Homoclinic is new CT1_ScalVectors.Homoclinic;

  subtype R_FHN_Data0 is R_ScalVectors_Homoclinic.FHN_Data0;
  subtype RT1_FHN_Data0 is RT1_ScalVectors_Homoclinic.FHN_Data0;
  subtype C_FHN_Data0 is C_ScalVectors_Homoclinic.FHN_Data0;
  subtype CT1_FHN_Data0 is CT1_ScalVectors_Homoclinic.FHN_Data0;

------------------------------------------------------------------
--- precision etc ------------------------------------------------
------------------------------------------------------------------

  procedure Set_Scalar_Precision(Binary_Digs: in Natural := 0);
  procedure Set_Precisions(Binary_Digs: in Positive);

  procedure Set_SRad1(R: in Radius)               ; --- set default radius for Taylor1-type scalars
  procedure Set_Newton_Parallel(P: in Boolean := True);
  procedure Set_Taylor2_Tasks(N: in Positive := 1); --- if N>1 then also set Tylors1.Prod_Parallel
  procedure Show_Param;

  function Input_Prefix return String;
  function Output_Prefix return String;

  R_Zero: constant R_Scalar := Scal(0);
  Trunc:  constant Boolean  := Info(R_Zero).IsNumeric;
  Stand:  constant Boolean  := Info(R_Zero).IsStandard;

------------------------------------------------------------------
--- the real stuff -----------------------------------------------
------------------------------------------------------------------
  
  function CApprox return R_Scalar;
  function SApprox return C_Scalar;
  
  procedure Write_RT1_Ws0(FileName: in String; R_C: in R_Scalar; R: in Radius);
  procedure Write_CT1_Unstable0(FileName: in String; R_C: in R_Scalar; R0,R: in Radius);
  procedure Write_R_Stable(FileName: in String; Y: in FltVec; R_C: in R_Scalar);
  procedure Write_C_Unstable(FileName: in String; R_C: in R_Scalar; R0: in Radius);

  procedure Find_Homoclinic(RT1_Ws: in RT1_Vector3; CT1_Pu: in CT1_Taylor2_Vector3;
                            C: in out R_Scalar; S: in out C_Scalar; Steps: in Natural := 64);
  procedure Find_Homoclinic(R: in Radius; C: in out R_Scalar; S: in out C_Scalar);
  procedure Iter_Find_Homoclinic(R0,R: in Radius; Iter: in Positive);

--- spectrum
  procedure C_Spec_Stable(C,Z: in C_Scalar; U: in out C_Vector3);
  procedure C_Spec_Local_Unstable(C,Z: in C_Scalar; V: in out C_Vector3);
  procedure C_Delta(Z: in C_Scalar; D: in out C_Scalar);
  procedure C_Delta(X,Y: in Flt; U,V: out Flt);

  procedure CT1_Spec_Stable(C,Z: in CT1_Scalar; U: in out CT1_Vector3);
  procedure CT1_Spec_Local_Unstable(C,Z: in CT1_Scalar; V: in out CT1_Vector3);
  procedure CT1_Delta(Z: in CT1_Scalar; D: in out CT1_Scalar);
  procedure CT1_Delta(Z: in C_Scalar; R: in Radius; D: in out CT1_Scalar);
  procedure Delta_Check_Origin(R: in Radius; D: in out CT1_Scalar; OK: out Boolean);
  
  subtype Segment is Paths.Segment;
  
  procedure Show_Corners renames Paths.ShowCorners;
  function NZ(N: Natural) return C_Scalar;
  function Alt_Radius(S: Segment; D: CT1_Scalar) return Natural;
  procedure Delta_Check_Segment(S: in out Segment);
  
  Testing: Boolean := False;
  UPhi: C_Taylor2_Vector3 := C_Taylor2_Vectors.ZeroVector(1,3); --- unstable manifold
  SPhi: R_Homoclinic1.Curve3_Access := null;                    --- stable manifold
  
private
  
------------------------------------------------------------------
--- conversion of taylor types -----------------------------------
------------------------------------------------------------------

  procedure Set_Real(R: in R_Taylor1; C: in out C_Taylor1);
  procedure Set_Imag(R: in R_Taylor1; C: in out C_Taylor1);
  procedure Set_Complex(Rx,Ry: in R_Taylor1; C: in out C_Taylor1);
  procedure Real_Part(C: in C_Taylor1; R: in out R_Taylor1);
  procedure Imag_Part(C: in C_Taylor1; R: in out R_Taylor1);
  
  function ScalRho(P: in RT1_Taylor1) return Radius;
  function ScalRho(P: in CT1_Taylor1) return Radius;
  function ScalRho(P: in CT1_Taylor2) return Radius;
  function ScalRho(P: in RT1_Taylor1_Vector) return Radius;
  function ScalRho(P: in CT1_Taylor1_Vector) return Radius;
  function ScalRho(P: in CT1_Taylor2_Vector) return Radius;
  
  procedure SetScalRho(R: in Radius; P: in out RT1_Taylor1; Check: in Boolean := True);
  procedure SetScalRho(R: in Radius; P: in out CT1_Taylor1; Check: in Boolean := True);
  procedure SetScalRho(R: in Radius; P: in out CT1_Taylor2; Check: in Boolean := True);
  procedure SetScalRho(R: in Radius; P: in out RT1_Taylor1_Vector; Check: in Boolean := True);
  procedure SetScalRho(R: in Radius; P: in out CT1_Taylor1_Vector; Check: in Boolean := True);
  procedure SetScalRho(R: in Radius; P: in out CT1_Taylor2_Vector; Check: in Boolean := True);

  procedure Assign(R: in R_Taylor1; C: in out C_Taylor1) renames Set_Real;
  procedure Assign(R1: in R_Taylor1; R2: in out RT1_Taylor1);
  procedure Assign(R: in R_Taylor1; C: in out CT1_Taylor1);
  procedure Assign(C1: in C_Taylor2; C2: in out C_Taylor2) renames C_Taylors2.Copy;
  procedure Assign(C1: in C_Taylor2; C2: in out CT1_Taylor2);

  procedure Eval(P: in RT1_Taylor1; R: in R_Scalar; Q: out R_Taylor1);
  procedure Eval(P: in CT1_Taylor1; C: in C_Scalar; Q: out C_Taylor1);
  procedure Eval(P: in CT1_Taylor2; C: in C_Scalar; Q: out C_Taylor2);
  procedure Eval(P: in RT1_Taylor1_Vector; R: in R_Scalar; Q: out R_Taylor1_Vector);
  procedure Eval(P: in CT1_Taylor1_Vector; C: in C_Scalar; Q: out C_Taylor1_Vector);
  procedure Eval(P: in CT1_Taylor2_Vector; C: in C_Scalar; Q: out C_Taylor2_Vector);
  procedure EvalDer(P: in RT1_Taylor1; R: in R_Scalar; Q: out R_Taylor1);
  procedure EvalDer(P: in CT1_Taylor1; C: in C_Scalar; Q: out C_Taylor1);
  procedure EvalDer(P: in CT1_Taylor2; C: in C_Scalar; Q: out C_Taylor2);
  procedure EvalDer(P: in RT1_Taylor1_Vector; R: in R_Scalar; Q: out R_Taylor1_Vector);
  procedure EvalDer(P: in CT1_Taylor1_Vector; C: in C_Scalar; Q: out C_Taylor1_Vector);
  procedure EvalDer(P: in CT1_Taylor2_Vector; C: in C_Scalar; Q: out C_Taylor2_Vector);

------------------------------------------------------------------
--- precision etc ------------------------------------------------
------------------------------------------------------------------

  function Effective_Precision(Binary_Digs: Positive) return Positive;

  Scalar_Precision: Positive := Initial_Precision;
  Newton_Parallel: Boolean := False;
  Taylor2_Tasks: Positive := 1;

------------------------------------------------------------------
--- FHN stuff ----------------------------------------------------
------------------------------------------------------------------
  
  function MinDistRPlus(Z: C_Scalar) return Flt;
  procedure Val_HalfCircle(R: in Radius; N: in Positive; K: in Integer; S: in CT1_Scalar; Z,W: in out C_Scalar);
  procedure Write_Germ(S: in Segment; D: in CT1_Scalar);
  
  R_Pi: constant R_Scalar := Two*ArcCos(Scal(0));
  
end FHN_Homoclinic;
