with Messages;
use Messages;

package body FHN_Periodic is


   ------------------------------------------------------------------
   --- precision etc ------------------------------------------------
   ------------------------------------------------------------------

   procedure Set_Scalar_Precision(Binary_Digs: in Natural := 0) is
   begin
      if Binary_Digs>0 then
         Scalar_Precision := Binary_Digs;
      end if;
      R_NHome.Set_Precision(Scalar_Precision);
   end Set_Scalar_Precision;

   procedure Set_Numeric_Precision(Binary_Digs: in Natural := 0) is
   begin
      if Binary_Digs>0 then
         Numeric_Precision := Binary_Digs;
      end if;
      R_NHome.Set_Precision(Numeric_Precision);
   end Set_Numeric_Precision;

   procedure Set_Precisions(Binary_Digs: in Positive) is
   begin
      Set_Scalar_Precision(Binary_Digs);
      Set_Numeric_Precision(Binary_Digs);
   end Set_Precisions;

   function Effective_Precision(Binary_Digs: Positive) return Positive is
      use R_NHome;
      NP: constant Positive := Numeric_Precision;
      EP: Positive;
   begin
      R_NHome.Set_Precision(Binary_Digs);
      EP := R_NHome.Get_Precision(R_NHome.Pi); -- involves a computation
      R_NHome.Set_Precision(NP);
      return EP;
   end Effective_Precision;

   procedure Relax_Checks is
   begin
      Set_Wait_Mode(R_Zero);
      R_NHome.Set_Active_Mode(R_NZero);
   end Relax_Checks;

   procedure Restore_Checks is
   begin
      R_NHome.Set_Wait_Mode(R_NZero);
      Set_Active_Mode(R_Zero);
   end Restore_Checks;

   procedure Set_SRad1(R: in Radius) is
   begin
      RT1_Home.Rad  := R;
      RT1_NHome.Rad := R;
      CT1_Home.Rad  := R;
      CT1_NHome.Rad := R;
   end Set_SRad1;

   procedure Set_Fourier_Parallel(P: in Boolean := True) is
   begin
     Fourier_Parallel := P;
     R_Fouriers1.Prod_Parallel := P;
     C_Fouriers1.Prod_Parallel := P;
     RT1_Fouriers1.Prod_Parallel := P;
   end Set_Fourier_Parallel;

   ------------------------------------------------------------------
   --- conversion of scalar types -----------------------------------
   ------------------------------------------------------------------

   -- Real

   procedure S2N(S: in R_Matrix; N: in out R_NMatrix) is
   begin
      for I in S'Range(1) loop
         for J in S'Range(2) loop S2N(S(I,J),N(I,J)); end loop;
      end loop;
   end S2N;

   procedure N2S(N: in R_NMatrix; S: in out R_Matrix) is
   begin
      for I in N'Range(1) loop
         for J in N'Range(2) loop N2S(N(I,J),S(I,J)); end loop;
      end loop;
   end N2S;

   procedure S2N(S: in R_Vector; N: in out R_NVector) is
   begin
      for I in S'Range loop S2N(S(I),N(I)); end loop;
   end S2N;

   procedure N2S(N: in R_NVector; S: in out R_Vector) is
   begin
      for I in N'Range loop N2S(N(I),S(I)); end loop;
   end N2S;

   -- Real Taylor1

   procedure S2N(S: in RT1_Scalar; N: in out RT1_Numeric) is
   begin
      S2N(S.C,N.C);
      if S.F=FConst then N.F := FConst; else N.F := RT1_NHome.PDeg1; end if;
      N.R := S.R;
   end S2N;

   procedure N2S(N: in RT1_Numeric; S: in out RT1_Scalar) is
   begin
      N2S(N.C,S.C);
      S.F := N.F;
      S.R := N.R;
   end N2S;

   procedure S2N(S: in RT1_Matrix; N: in out RT1_NMatrix) is
   begin
      for I in S'Range(1) loop
         for J in S'Range(2) loop S2N(S(I,J),N(I,J)); end loop;
      end loop;
   end S2N;

   procedure N2S(N: in RT1_NMatrix; S: in out RT1_Matrix) is
   begin
      for I in N'Range(1) loop
         for J in N'Range(2) loop N2S(N(I,J),S(I,J)); end loop;
      end loop;
   end N2S;

   procedure S2N(S: in RT1_Vector; N: in out RT1_NVector) is
   begin
      for I in S'Range loop S2N(S(I),N(I)); end loop;
   end S2N;

   procedure N2S(N: in RT1_NVector; S: in out RT1_Vector) is
   begin
      for I in N'Range loop N2S(N(I),S(I)); end loop;
   end N2S;

   -- Complex

   procedure S2N(S: in C_Scalar; N: in out C_Numeric) is
   begin
      S2N(S.Re,N.Re);
      S2N(S.Im,N.Im);
   end S2N;

   procedure N2S(N: in C_Numeric; S: in out C_Scalar) is
   begin
      N2S(N.Re,S.Re);
      N2S(N.Im,S.Im);
   end N2S;

   procedure S2N(S: in C_Matrix; N: in out C_NMatrix) is
   begin
      for I in S'Range(1) loop
         for J in S'Range(2) loop S2N(S(I,J),N(I,J)); end loop;
      end loop;
   end S2N;

   procedure N2S(N: in C_NMatrix; S: in out C_Matrix) is
   begin
      for I in N'Range(1) loop
         for J in N'Range(2) loop N2S(N(I,J),S(I,J)); end loop;
      end loop;
   end N2S;

   procedure S2N(S: in C_Vector; N: in out C_NVector) is
   begin
      for I in S'Range loop S2N(S(I),N(I)); end loop;
   end S2N;

   procedure N2S(N: in C_NVector; S: in out C_Vector) is
   begin
      for I in N'Range loop N2S(N(I),S(I)); end loop;
   end N2S;

   -- Complex Taylor1

   procedure S2N(S: in CT1_Scalar; N: in out CT1_Numeric) is
   begin
      S2N(S.C,N.C);
      if S.F=FConst then N.F := FConst; else N.F := CT1_NHome.PDeg1; end if;
      N.R := S.R;
   end S2N;

   procedure N2S(N: in CT1_Numeric; S: in out CT1_Scalar) is
   begin
      N2S(N.C,S.C);
      S.F := N.F;
      S.R := N.R;
   end N2S;

   procedure S2N(S: in CT1_Matrix; N: in out CT1_NMatrix) is
   begin
      for I in S'Range(1) loop
         for J in S'Range(2) loop S2N(S(I,J),N(I,J)); end loop;
      end loop;
   end S2N;

   procedure N2S(N: in CT1_NMatrix; S: in out CT1_Matrix) is
   begin
      for I in N'Range(1) loop
         for J in N'Range(2) loop N2S(N(I,J),S(I,J)); end loop;
      end loop;
   end N2S;

   procedure S2N(S: in CT1_Vector; N: in out CT1_NVector) is
   begin
      for I in S'Range loop S2N(S(I),N(I)); end loop;
   end S2N;

   procedure N2S(N: in CT1_NVector; S: in out CT1_Vector) is
   begin
      for I in N'Range loop N2S(N(I),S(I)); end loop;
   end N2S;

   ------------------------------------------------------------------
   --- conversion of other types ------------------------------------
   ------------------------------------------------------------------

   procedure S2N(S: in RT1_Fourier1; N: in out RT1_NFourier1) is
   begin
      S2N(S.C,N.C);
      N.R := S.R;
   end S2N;

   procedure N2S(N: in RT1_NFourier1; S: in out RT1_Fourier1) is
   begin
      N2S(N.C,S.C);
      if not Trunc then FV.SetZero(S.E); end if;
      S.R := N.R;
   end N2S;

   procedure T2R(T: in RT1_NFourier1; R: in out R_NFourier1) is
   begin
      for K in R_NFouriers1.Freq loop
         R.C(K) := Val(T.C(K),Scal(Zero));
      end loop;
      R.R := T.R;
   end T2R;

   procedure R2T(R: in R_NFourier1; T: in out RT1_NFourier1) is
   begin
      RT1_NFouriers1.SetZero(T);
      for K in R_NFouriers1.Freq loop
         RT1_NFouriers1.AddCoeff(K,Scal(R.C(K)),T);
      end loop;
      T.R := R.R;
   end R2T;

   procedure T2R(T: in RT1_Fourier1; R: in out R_Fourier1) is
   begin
      R_Fouriers1.SetZero(R);
      for K in R_Fouriers1.Freq loop
         R.C(K) := Val(T.C(K),Enclose(-One,One));
      end loop;
      R.R := T.R;
   end T2R;

   procedure R2T(R: in R_Fourier1; T: in out RT1_Fourier1) is
   begin
      RT1_Fouriers1.SetZero(T);
      for K in R_Fouriers1.Freq loop
         RT1_Fouriers1.AddCoeff(K,Scal(R.C(K)),T);
      end loop;
      T.R := R.R;
   end R2T;

   procedure S2N(S: in R_Fourier1; N: in out R_NFourier1) is
   begin
      S2N(S.C,N.C);
      N.R := S.R;
   end S2N;

   procedure N2S(N: in R_NFourier1; S: in out R_Fourier1) is
   begin
      N2S(N.C,S.C);
      if not Trunc then FV.SetZero(S.E); end if;
      S.R := N.R;
   end N2S;

   ------------------------------------------------------------------
   --- the real stuff -----------------------------------------------
   ------------------------------------------------------------------

   procedure Show_Param is
   begin
      if Trunc then
         Show0("Scalar type: Numeric");
      else
         Show0("Scalar type: Ball");
      end if;
      Show0("Rep type: " & Info(R_Zero).RepType);
      Show1("S-Bits =",Effective_Precision(Scalar_Precision));
      Show1("N-Bits =",Effective_Precision(Numeric_Precision));
      Show1("Fourier_Parallel: ",Fourier_Parallel);
      Show1("SPDeg1 =",SPDeg1);
      Show1("Deg_Taylor1  =",Deg_Taylor1);
      Show1("Deg_Fourier1 =",Deg_Fourier1);
   end Show_Param;

   function CApprox return R_Scalar is
     C: R_Scalar;
   begin
     Read(PData_Dir & "CApprox",C);
     Center(C);
     return C;
   end CApprox;

   function SetParam(C: in R_Scalar; DC,Length: in Radius) return R_Param is
     P: R_Param;
   begin
     Copy((C+BallAt0(DC))*Pi2/Length,P(1));
     Copy(Sqr(Pi2/Length),P(2));
     return P;
   end SetParam;
   
   function SetParam(C: in R_Scalar; DC,Length: in Radius) return RT1_Param is
     P: RT1_Param;
   begin
     Affine(C*Pi2/Length,DC*Pi2/Length,P(1));
     Assign(Sqr(Pi2/Length),P(2));
     return P;
   end SetParam;
   
   procedure Num_ContrMatrix(P: in RT1_Param; F,Ev: in RT1_Fourier1; A: in out RT1_Matrix) is
     use RT1F_NPeriodic;
      PN: RT1_NParam;
      AN: RT1_NMatrix(A'Range(1),A'Range(2));
      FN,EvN: RT1_NFourier1;
   begin
      Relax_Checks;
      S2N(P,PN);
      S2N(F,FN);
      S2N(Ev,EvN);
      ContrMatrix(PN,FN,EvN,AN);
      N2S(AN,A);
      Restore_Checks;
   end Num_ContrMatrix;

   procedure Num_EigenVec(A: in R_Matrix; Vr,Vi: in out R_Matrix) is
     use RF_NPeriodic.SSVN;
     Dim: constant Integer := A'Last(1);
     WNr,WNi: R_NVector(1..Dim);
     AN,VNr,VNi: R_NMatrix(1..Dim,1..Dim);
   begin
     Relax_Checks;
     S2N(A,AN);
     Eigen(AN,WNr,WNi,VNr,VNi);
     N2S(VNr,Vr);
     N2S(VNi,Vi);
     Restore_Checks;
   end Num_EigenVec;
   
   procedure CheckFix(P: in RT1_Param; Dim: in Positive; F: in out RT1_Fourier1; Ev: in RT1_Fourier1; KMax: in Radius) is
     use RT1F_Periodic;
     A: RT1_Matrix(1..Dim,1..Dim);
   begin
     Num_ContrMatrix(P,F,Ev,A);
     CheckFix(P,A,F,Ev,KMax);
   end CheckFix;
   
-------- spectrum
   
   function CountEigen(DVel,Length: in Flt; FT: in RT1_Fourier1) return Natural is
      use Gen_Scalars,R_Matrices,C_Matrices,C_ScalVectors,RF_Periodic;
      Tr: constant Tracer := Starting("CountEigen");
      R: constant Flt := Half**6;
      P: constant R_Param := SetParam(CApprox,DVel,Length);
      Dim: Positive := Deg_Fourier1*4;
      Rad,RP: R_Scalar;
      L0,F0,A,Vr,Vi: R_Matrix(1..Dim,1..Dim);
      D,M,Minv,AC: C_Matrix(1..Dim,1..Dim);
      F: R_Fourier1;
   begin
      T2R(FT,F);
      MatrixL0(P,L0);
      MatrixF(F,F0);
      Sum(L0,F0,A);
      Num_EigenVec(A,Vr,Vi);
      Set_Complex(Vr,Vi,M);
      Set_Real(A,AC);
      Copy(M,Minv);
      Invert(Minv);
      Prod(AC,M,D);
      Mult(Minv,D);
      for J in D'Range(1) loop
        SetZero(Rad);
        for K in D'Range(2) loop
          if J /= K then Rad := Rad+Norm(D(J,K)); end if;
        end loop;
        Real_Part(D(J,J),RP);
        if Sup(RP+Rad)<Zero and Inf(Rad+Norm(D(J,J)))>R then
          Dim := Dim-1;
        end if;
      end loop;
      return Dim;
   end CountEigen;

   procedure CheckLinear(DVel,Length: in Flt; FT: in RT1_Fourier1; Y1,Y2: in R_Scalar) is
      use Gen_Scalars,R_Matrices,C_Matrices,C_ScalVectors;
      use C_Fouriers1,RT1_Fouriers1,RF_Periodic,CF_Periodic;
      Tr: constant Tracer := Starting("CheckLinear");
      Dim: constant Positive := 4*Deg_Fourier1;
      MaxProd: constant Radius := One-Half**10;
      RF: constant Radius := 0.05;
      EpsF: constant Radius := 0.015;
      Eps: constant R_Scalar := Scal(EpsF);
      R: constant R_Scalar := Scal(Half**6);
      ANorm,LNorm,Temp: Radius;
      DY,S,Y: R_Scalar;
      YC: C_Scalar;
      P: R_Param;
      PC: C_Param;
      L0,F0,A: R_Matrix(1..Dim,1..Dim);
      AC: C_Matrix(1..Dim,1..Dim);
      U: R_Diag(1..Dim);
      UC: C_Diag(1..Dim);
      F: R_Fourier1;
      FC: C_Fourier1;
   begin
      P(1) := Two*Pi*(CApprox+BallAt0(DVel))/Length;
      P(2) := Sqr(Two*Pi/Length);
      Set_Real(P,PC);
      T2R(FT,F);
      MatrixL0(P,L0);
      MatrixF(F,F0);
      Sum(L0,F0,A);
      for K in U'Range loop
         Assign(1,U(K));
      end loop;
      for K in 1..Deg_Fourier1 loop
         S := Eps*Exp(Scal(RF)*Scal(Deg_Fourier1-K));
         if Approx(S)<One then
            U(K) := S;
            U(K+Deg_Fourier1) := S;
         end if;
      end loop;
      ANorm := NormA(F,U,Sup(P(2)));
      Show1("||UAU|| ",ANorm);
      Set_Real(U,UC);
      SetZero(Rho(FT),FC);
      Y := Y1;
      while Y <= Y2 loop
         if Verbosity>1 then Show1("y ",Approx(Y)); end if;
         LNorm := Zero;
         if abs(Y) > R then
           Set_Imag(Y,YC);
         elsif Sup(Y)>Zero then
           Set_Complex(Y-R,Y,YC);
         else
           Set_Complex(-Y-R,Y,YC);
         end if;
         Temp := Sup(Inv(Eps*Eps))*NormL0H(PC,Deg_Fourier1+1,YC);
         LNorm := RMax(LNorm,Temp);
         Set_Real(A,AC);
         AddConst(-YC,AC);
         Mult(UC,AC);
         Mult(AC,UC);
         Invert(AC);
         Temp := OpNorm(Rho(FC),AC);
         LNorm := RMax(LNorm,Temp);
         if Verbosity>1 then Show1("||A||||(L-iy)^(-1)|| ",ANorm*LNorm); end if;
         if ANorm*LNorm>=One then
            Show0("FHN_Periodic.CheckLinear error");
            Show1("y=",Approx(Y));
            Show1("||A||||(L-iy)^(-1)||=",ANorm*LNorm);
            raise Domain_Error;
         end if;
         DY := Scal((MaxProd/ANorm-LNorm)/(MaxProd*LNorm/ANorm));
         if Y > R or Y+DY < -R then
            Y := Y+DY;
         else
            Y := Y+DY*Sqrt(Scal(Half));
         end if;
      end loop;
   end CheckLinear;
   
----------- numerical
   
   procedure Find_Fix is
      use RT1_Home,RT1_Vectors,RT1_Fouriers1,RT1F_Periodic;
      Dim: constant Positive := 60;  --must be even
      Length: constant Radius := Flt(128);
      Eps: constant Radius := 1.0e-50;
      DVel: constant Radius := Half**60;
      Vel: constant R_Scalar := CApprox;
      P: constant RT1_Param := SetParam(Vel,DVel,Length);
      ST: RT1_Scalar;
      A: RT1_Matrix(1..Dim,1..Dim);
      X1,X2,X3,Y1,Y2,Y3,S: R_Scalar;
      V1,V2,V3: RT1_Fourier1;
   begin
      if not Trunc then
        raise Sorry with "FHN_Periodic.Find: run me in numeric mode";
      end if;
      Read(PData_Dir & "fixt.data",V1);
      Read(PData_Dir & "evt.data",V3);
      Map(P,V1,V3,V2);
      Show1("norm: ",MaxNorm(V1));
      Show1("diff: ",MaxNorm(V2-V1));
      MapA(P,V1,V2);
      Compo1(V2-V1,V3,ST);
      Val(ST,Scal(One),S);
      Show2("V,Q = ",S,Vel+Scal(DVel));
      Val(ST,Scal(-1),S);
      Show2("V,Q = ",S,Vel-Scal(DVel));
      Assign(-1,X1);
      Assign( 1,X2);
      Val(ST,X1,Y1);
      Val(ST,X2,Y2);
      if Approx(Y1*Y2)>Zero then Show("wrong interval"); end if;
      Copy(Y2,Y3);
      while MaxNorm(Y3)>Eps loop
         X3 := Half*(X1+X2);
         Val(ST,X3,Y3);
         if Approx(Y3*Y1)>Zero then
            Copy(X3,X1);
         else
           Copy(X3,X2);
         end if;
         Show2("V,Q = ",Y3,Vel+DVel*X3);
      end loop;
      --  Write("CApprox",Vel+DVel*X3);
      DerMatrix(P,V1,A);
      ShowEigen(A);
   end Find_Fix;

end FHN_Periodic;
