with Ada.Text_IO, Logicals, Globals, Param, Ints, Rationals, Flts, Strings, Reps, Num_Rep_Scalars;
use Ada.Text_IO, Logicals, Globals, Param, Ints, Rationals, Flts, Strings;
with Scalars, Matrices, Vectors, ScalVectors.RNum;
with Taylors1, Fouriers1, Fouriers1.Periodic;

pragma Elaborate_All (Ada.Text_IO,Logicals,Globals,Param,Ints,Rationals,Flts,Strings,Reps,Num_Rep_Scalars);
pragma Elaborate_All (Scalars,Matrices,Vectors,ScalVectors.RNum,Taylors1,Fouriers1.Periodic);

generic

  Deg_Taylor1:  in Positive := 2;
  Deg_Fourier1: in Positive;
  SPDeg1:       in Positive := 2;

-------------------- begin_include scalars Real

  type Real is private;

--- basic
  with function Info(Dummy: Real) return Scalar_Info is <>;
  with function IsSharp(S: Real) return Boolean is <>;
  with function IsZero(S: Real) return Boolean is <>;
  with function "="(S1,S2: Real) return Boolean is <>;
  with procedure SetZero(S: in out Real) is <>;
  with procedure Copy(S1: in Real; S2: in out Real) is <>;
  with procedure Swap(S1,S2: in out Real) is <>;

--- sets
  with function Center0(S: Real) return Boolean is <>;
  with function Contains0(S: Real) return Logical is <>;
  with function Contains(S1,S2: Real) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Real) is <>;
  with function BallAt0(R: Flt) return Real is <>;
  with procedure ToErr(S: in out Real) is <>;
  with procedure ToErr(S1: in Real; S2: in out Real) is <>;
  with function ToErr(S: Real) return Real is <>;
  with procedure Center(S: in out Real) is <>;
  with procedure Center(S1: in Real; S2: in out Real) is <>;
  with function Center(S: Real) return Real is <>;
  with procedure ModCenter(S: in out Real) is <>;
  with procedure ModCenter(S1: in Real; S2: in out Real) is <>;
  with function ModCenter(S: Real) return Real is <>;
  with procedure ErrMult(R: in Radius; S: in out Real) is <>;
  with procedure Union(S1: in Real; S2: in out Real) is <>;
  with function Union(S1,S2: Real) return Real is <>;
  with procedure Intersection(S1: in Real; S2: in out Real; Empty: out Logical) is <>;

--- order
  with function Sign(S: Real) return Integer is <>;
  with function Compare(S1,S2: Real) return Integer is <>;
  with function "<"(S1,S2: Real) return Boolean is <>;
  with function "<="(S1,S2: Real) return Boolean is <>;
  with function ">"(S1,S2: Real) return Boolean is <>;
  with function ">="(S1,S2: Real) return Boolean is <>;
  with procedure Min(S1: in Real; S2: in out Real) is <>;
  with procedure Min(S1,S2: in Real; S3: in out Real) is <>;
  with function Min(S1,S2: Real) return Real is <>;
  with procedure Max(S1: in Real; S2: in out Real) is <>;
  with procedure Max(S1,S2: in Real; S3: in out Real) is <>;
  with function Max(S1,S2: Real) return Real is <>;
  with function Inf(S: Real) return Flt is <>; -- inf of real part
  with function Sup(S: Real) return Flt is <>; -- sup of real part

--- arithmetic
  with procedure Neg(S: in out Real) is <>;
  with procedure Neg(S1: in Real; S2: in out Real) is <>;
  with function "-"(S: Real) return Real is <>;
  with procedure Add(I: in Integer; S: in out Real) is <>;
  with procedure Add(S1: in Real; S2: in out Real) is <>;
  with procedure Sum(S1,S2: in Real; S3: in out Real) is <>;
  with function "+"(S1,S2: Real) return Real is <>;
  with procedure Sub(S1: in Real; S2: in out Real) is <>;
  with procedure Diff(S1,S2: in Real; S3: in out Real) is <>;
  with function "-"(S1,S2: Real) return Real is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Real) is <>;
  with procedure Mult(R: in Flt; S: in out Real) is <>;
  with procedure Prod(R: in Flt; S1: in Real; S2: in out Real) is <>;
  with function "*"(R: Flt; S: Real) return Real is <>;
  with procedure AddProd(R: in Flt; S1: in Real; S2,Tmp: in out Real) is <>;
  with procedure AddProd(R: in Flt; S1: in Real; S2: in out Real) is <>;
  with procedure Mult(Q: in Rational; S: in out Real) is <>;
  with procedure Mult(S1: in Real; S2: in out Real) is <>;
  with procedure Prod(S1,S2: in Real; S3: in out Real) is <>;
  with function "*"(S1,S2: Real) return Real is <>;
  with procedure AddProd(S1,S2: in Real; S3,Tmp: in out Real) is <>;
  with procedure AddProd(S1,S2: in Real; S3: in out Real) is <>;
  with procedure SubProd(S1,S2: in Real; S3,Tmp: in out Real) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Real) is <>;
  with procedure Div(R: in Flt; S: in out Real) is <>;
  with procedure Quot(S1: in Real; R: in Flt; S2: in out Real) is <>;
  with function "/"(S: Real; R: Flt) return Real is <>;
  with procedure Div(S1: in Real; S2: in out Real) is <>;
  with procedure Quot(S1,S2: in Real; S3: in out Real) is <>;
  with function "/"(S1,S2: Real) return Real is <>;
  with procedure Inv(S1: in Real; S2: in out Real) is <>;
  with function Inv(S: Real) return Real is <>;
  with function "**"(S: Real; I: Integer) return Real is <>;

--- fun
  with function IsReal(S: Real) return Boolean is <>;
  with procedure Adjoint(S: in out Real) is <>;
  with procedure Adjoint(S1: in Real; S2: in out Real) is <>;
  with function Adjoint(S: Real) return Real is <>;
  with procedure Real_Part(S: in out Real) is <>;
  with procedure Imag_Part(S: in out Real) is <>;
  with procedure Eval0(S: in out Real) is <>;
  with function "abs"(S: Real) return Real is <>;
  with procedure Norm(S1: in Real; S2: in out Real) is <>;
  with function Norm(S: Real) return Real is <>;
  with function MaxNorm(S: Real) return Radius is <>;
  with procedure Sqr(S: in out Real) is <>;
  with function Sqr(S: Real) return Real is <>;
  with procedure Sqrt(S1: in Real; S2: in out Real) is <>;
  with function Sqrt(S: Real) return Real is <>;
  with procedure Root(K: in Positive; S1: in Real; S2: in out Real) is <>;
  with function Root(K: Positive; S: Real) return Real is <>;
  with procedure Exp(S1: in Real; S2: in out Real) is <>;
  with function Exp(S: Real) return Real is <>;
  with procedure ArcCos(S1: in Real; S2: in out Real) is <>;
  with function ArcCos(S: Real) return Real is <>;
  with procedure ArcSin(S1: in Real; S2: in out Real) is <>;
  with function ArcSin(S: Real) return Real is <>;
  with procedure Cos(S1: in Real; S2: in out Real) is <>;
  with function Cos(S: Real) return Real is <>;
  with procedure Sin(S1: in Real; S2: in out Real) is <>;
  with function Sin(S: Real) return Real is <>;
  with procedure Log(S1: in Real; S2: in out Real) is <>;
  with function Log(S: Real) return Real is <>;
  with procedure Simple_Random(S: in out Real) is <>;
  with procedure QPower(Q: in Rational; S1: in Real; S2: in out Real) is <>;
  with procedure Roots2(B,C: in Real; U1,U2,V: in out Real) is <>;
  with procedure Roots3(B,C,D: in Real; U0,U1,U2,V: in out Real) is <>;

--- conversion and i/o
  with function Approx(S: Real) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Real) is <>;
  with procedure Assign(Q: in Rational; S: in out Real) is <>;
  with procedure Assign(R: in Flt; S: in out Real) is <>;
  with function Scal(I: Integer) return Real is <>;
  with function Scal(Q: Rational) return Real is <>;
  with function Scal(R: Flt) return Real is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Real) is <>;
  with function Enclose(R1,R2: Flt) return Real is <>;
  with procedure Show1(N: in String; S: in Real; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Real; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Real; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Real; Decimal: in Boolean := False) is <>;

--- misc
  with function Get_Precision(R: Real) return Positive is <>;
  with procedure Set_Active_Mode(Dummy: in Real) is <>;
  with procedure Set_Wait_Mode(Dummy: in Real) is <>;
  with procedure Free_Cache(S: in Real) is <>;
-------------------- end_include scalars
  
  with function Pi return Real is <>;
  with procedure Write(FileName: in String; R: in Real; Decimal: in Boolean := False) is <>;
  with procedure Read(FileName: in String; R: in out Real; Decimal: in Boolean := False) is <>;
  with procedure PutNum(S: in Real; R: in out Reps.Rep) is <>;
  with procedure GetNum(R: in Reps.Rep; S: in out Real) is <>;

package FHN_Periodic is

  pragma Elaborate_Body;

  package Gen_Scalars is new Scalars (SPDeg1 => SPDeg1, Real => Real);

------------------------------------------------------------------
--- scalar types -------------------------------------------------
------------------------------------------------------------------

-------- Real Numbers
  subtype R_Scalar is Gen_Scalars.Real_Number;
  package R_Matrices renames Gen_Scalars.Real_Matrices;
  subtype R_Matrix is Gen_Scalars.Real_Matrix;
  subtype R_Diag is Gen_Scalars.Real_Diag;
  subtype R_Matrix2 is Gen_Scalars.Real_Matrix2;
  subtype R_Matrix3 is Gen_Scalars.Real_Matrix3;
  package R_Vectors renames Gen_Scalars.Real_Vectors;
  subtype R_Vector is Gen_Scalars.Real_Vector;
  subtype R_Vector2 is Gen_Scalars.Real_Vector2;
  subtype R_Vector3 is Gen_Scalars.Real_Vector3;
  subtype R_Param is R_Vector2;
  package R_ScalVectors renames Gen_Scalars.Real_ScalVectors;

-------- Real Taylor1
  package RT1_Home renames Gen_Scalars.Real_Taylors1;
  subtype RT1_Scalar is Gen_Scalars.Real_Taylor1;
  package RT1_Matrices renames Gen_Scalars.Real_Taylor1_Matrices;
  subtype RT1_Matrix is Gen_Scalars.Real_Taylor1_Matrix;
  subtype RT1_Matrix2 is Gen_Scalars.Real_Taylor1_Matrix2;
  subtype RT1_Matrix3 is Gen_Scalars.Real_Taylor1_Matrix3;
  package RT1_Vectors renames Gen_Scalars.Real_Taylor1_Vectors;
  subtype RT1_Vector is Gen_Scalars.Real_Taylor1_Vector;
  subtype RT1_Vector2 is Gen_Scalars.Real_Taylor1_Vector2;
  subtype RT1_Vector3 is Gen_Scalars.Real_Taylor1_Vector3;
  subtype RT1_Param is RT1_Vector2;
  package RT1_ScalVectors renames Gen_Scalars.Real_Taylor1_ScalVectors;

-------- Complex Numbers
  package C_Home renames Gen_Scalars.Complex_Numbers;
  subtype C_Scalar is Gen_Scalars.Complex_Number;
  package C_Matrices renames Gen_Scalars.Complex_Matrices;
  subtype C_Matrix is Gen_Scalars.Complex_Matrix;
  subtype C_Diag is Gen_Scalars.Complex_Diag;
  subtype C_Matrix2 is Gen_Scalars.Complex_Matrix2;
  subtype C_Matrix3 is Gen_Scalars.Complex_Matrix3;
  package C_Vectors renames Gen_Scalars.Complex_Vectors;
  subtype C_Vector is Gen_Scalars.Complex_Vector;
  subtype C_Vector2 is Gen_Scalars.Complex_Vector2;
  subtype C_Vector3 is Gen_Scalars.Complex_Vector3;
  subtype C_Param is C_Vector2;
  package C_ScalVectors renames Gen_Scalars.Complex_ScalVectors;

-------- Complex Taylor1
  package CT1_Home renames Gen_Scalars.Complex_Taylors1;
  subtype CT1_Scalar is Gen_Scalars.Complex_Taylor1;
  package CT1_Matrices renames Gen_Scalars.Complex_Taylor1_Matrices;
  subtype CT1_Matrix is Gen_Scalars.Complex_Taylor1_Matrix;
  subtype CT1_Matrix2 is Gen_Scalars.Complex_Taylor1_Matrix2;
  subtype CT1_Matrix3 is Gen_Scalars.Complex_Taylor1_Matrix3;
  package CT1_Vectors renames Gen_Scalars.Complex_Taylor1_Vectors;
  subtype CT1_Vector is Gen_Scalars.Complex_Taylor1_Vector;
  subtype CT1_Vector2 is Gen_Scalars.Complex_Taylor1_Vector2;
  subtype CT1_Vector3 is Gen_Scalars.Complex_Taylor1_Vector3;
  subtype CT1_Param is CT1_Vector2;
  package CT1_ScalVectors renames Gen_Scalars.Complex_Taylor1_ScalVectors;

------------------------------------------------------------------
--- precision etc ------------------------------------------------
------------------------------------------------------------------

  procedure Set_Scalar_Precision(Binary_Digs: in Natural := 0);
  procedure Set_Numeric_Precision(Binary_Digs: in Natural := 0);
  procedure Set_Precisions(Binary_Digs: in Positive);

  procedure Set_SRad1(R: in Radius); --- set default radius for Taylor1-type scalars
  procedure Set_Fourier_Parallel(P: in Boolean := True);
  procedure Show_Param;
  
  R_Zero: constant R_Scalar := Scal(0);
  Trunc:  constant Boolean  := Info(R_Zero).IsNumeric;
  
------------------------------------------------------------------
--- FHN types ----------------------------------------------------
------------------------------------------------------------------
  
  Package R_Fouriers1 is new Fouriers1 (PDeg => Deg_Fourier1,
                                        Scalar => R_Scalar,
                                        SM => R_Matrices,
                                        SV => R_Vectors);
  package RF_Periodic is new R_Fouriers1.Periodic(SSV => R_ScalVectors);
  subtype R_Fourier1 is R_Fouriers1.Fourier1;
  
  use C_Home;
  package C_Fouriers1 is new Fouriers1 (PDeg => Deg_Fourier1,
                                        Scalar => C_Scalar,
                                        SM => C_Matrices,
                                        SV => C_Vectors);
  package CF_Periodic is new C_Fouriers1.Periodic(SSV => C_ScalVectors);
  subtype C_Fourier1 is C_Fouriers1.Fourier1;
  
  use RT1_Home;
  Package RT1_Fouriers1 is new Fouriers1 (PDeg => Deg_Fourier1,
                                          Scalar => RT1_Scalar,
                                          SM => RT1_Matrices,
                                          SV => RT1_Vectors);
  package RT1F_Periodic is new RT1_Fouriers1.Periodic(SSV => RT1_ScalVectors);
  subtype RT1_Fourier1 is RT1_Fouriers1.Fourier1;

------------------------------------------------------------------
--- the real stuff -----------------------------------------------
------------------------------------------------------------------
  
  function CApprox return R_Scalar;
  function SetParam(C: in R_Scalar; DC,Length: in Radius) return R_Param;
  function SetParam(C: in R_Scalar; DC,Length: in Radius) return RT1_Param;
  
  procedure CheckFix(P: in RT1_Param; Dim: in Positive; F: in out RT1_Fourier1; Ev: in RT1_Fourier1; KMax: in Radius);
  procedure CheckLinear(DVel,Length: in Flt; FT: in RT1_Fourier1; Y1,Y2: in R_Scalar);
  function CountEigen(DVel,Length: in Flt; FT: in RT1_Fourier1) return Natural;
  
--- purely numerical  
  procedure Find_Fix;
  
private
  
------------------------------------------------------------------
--- numeric scalar types -----------------------------------------
------------------------------------------------------------------
  
  package NRS is new Num_Rep_Scalars (SPDeg1 => SPDeg1);
  package Num_Scalars renames NRS.Num_Scalars;
  
--------- Real Numbers
  package R_NHome renames Reps;
  subtype R_Numeric is Num_Scalars.Real_Number;
  package R_NMatrices renames Num_Scalars.Real_Matrices;
  subtype R_NMatrix is Num_Scalars.Real_Matrix;
  package R_NVectors renames Num_Scalars.Real_Vectors;
  subtype R_NVector is Num_Scalars.Real_Vector;
  subtype R_NParam is Num_Scalars.Real_Vector2;
  package R_NScalVectors renames Num_Scalars.Real_ScalVectors;
  
--------- Real Taylor1
  package RT1_NHome renames Num_Scalars.Real_Taylors1;
  subtype RT1_Numeric is Num_Scalars.Real_Taylor1;
  package RT1_NMatrices renames Num_Scalars.Real_Taylor1_Matrices;
  subtype RT1_NMatrix is Num_Scalars.Real_Taylor1_Matrix;
  package RT1_NVectors renames Num_Scalars.Real_Taylor1_Vectors;
  subtype RT1_NVector is Num_Scalars.Real_Taylor1_Vector;
  package RT1_NScalVectors renames Num_Scalars.Real_Taylor1_ScalVectors;
  subtype RT1_NParam is Num_Scalars.Real_Taylor1_Vector2;

--------- Complex Numbers
  package C_NHome renames Num_Scalars.Complex_Numbers;
  subtype C_Numeric is Num_Scalars.Complex_Number;
  package C_NMatrices renames Num_Scalars.Complex_Matrices;
  subtype C_NMatrix is Num_Scalars.Complex_Matrix;
  package C_NVectors renames Num_Scalars.Complex_Vectors;
  subtype C_NVector is Num_Scalars.Complex_Vector;
  package C_NScalVectors renames Num_Scalars.Complex_ScalVectors;
  subtype C_NParam is Num_Scalars.Complex_Vector2;

--------- Complex Taylor1
  package CT1_NHome renames Num_Scalars.Complex_Taylors1;
  subtype CT1_Numeric is Num_Scalars.Complex_Taylor1;
  package CT1_NMatrices renames Num_Scalars.Complex_Taylor1_Matrices;
  subtype CT1_NMatrix is Num_Scalars.Complex_Taylor1_Matrix;
  package CT1_NVectors renames Num_Scalars.Complex_Taylor1_Vectors;
  subtype CT1_NVector is Num_Scalars.Complex_Taylor1_Vector;
  package CT1_NScalVectors renames Num_Scalars.Complex_Taylor1_ScalVectors;
  subtype CT1_NParam is Num_Scalars.Complex_Taylor1_Vector2;

  R_NZero: constant R_Numeric := R_NHome.Scal(0);

------------------------------------------------------------------
--- conversion of scalar types -----------------------------------
------------------------------------------------------------------

-- Real
  procedure S2N(S: in R_Scalar; N: in out R_Numeric) renames PutNum;
  procedure N2S(N: in R_Numeric; S: in out R_Scalar) renames GetNum;
  procedure S2N(S: in R_Matrix; N: in out R_NMatrix);
  procedure N2S(N: in R_NMatrix; S: in out R_Matrix);
  procedure S2N(S: in R_Vector; N: in out R_NVector);
  procedure N2S(N: in R_NVector; S: in out R_Vector);
-- Real Taylor1
  procedure S2N(S: in RT1_Scalar; N: in out RT1_Numeric);
  procedure N2S(N: in RT1_Numeric; S: in out RT1_Scalar);
  procedure S2N(S: in RT1_Matrix; N: in out RT1_NMatrix);
  procedure N2S(N: in RT1_NMatrix; S: in out RT1_Matrix);
  procedure S2N(S: in RT1_Vector; N: in out RT1_NVector);
  procedure N2S(N: in RT1_NVector; S: in out RT1_Vector);
-- Complex
  procedure S2N(S: in C_Scalar; N: in out C_Numeric);
  procedure N2S(N: in C_Numeric; S: in out C_Scalar);
  procedure S2N(S: in C_Matrix; N: in out C_NMatrix);
  procedure N2S(N: in C_NMatrix; S: in out C_Matrix);
  procedure S2N(S: in C_Vector; N: in out C_NVector);
  procedure N2S(N: in C_NVector; S: in out C_Vector);
-- Complex Taylor1
  procedure S2N(S: in CT1_Scalar; N: in out CT1_Numeric);
  procedure N2S(N: in CT1_Numeric; S: in out CT1_Scalar);
  procedure S2N(S: in CT1_Matrix; N: in out CT1_NMatrix);
  procedure N2S(N: in CT1_NMatrix; S: in out CT1_Matrix);
  procedure S2N(S: in CT1_Vector; N: in out CT1_NVector);
  procedure N2S(N: in CT1_NVector; S: in out CT1_Vector);

------------------------------------------------------------------
--- other numeric types ------------------------------------------
------------------------------------------------------------------

  use R_NHome;
  package R_NFouriers1 is new Fouriers1 (PDeg => Deg_Fourier1,
					 Scalar => R_Numeric,
					 SM => R_NMatrices,
					 SV => R_NVectors);
  package RF_NPeriodic is new R_NFouriers1.Periodic(SSV => R_NScalVectors);
  subtype R_NFourier1 is R_NFouriers1.Fourier1;
  
  use RT1_NHome;
  Package RT1_NFouriers1 is new Fouriers1 (PDeg => Deg_Fourier1,
                                          Scalar => RT1_Numeric,
                                          SM => RT1_NMatrices,
                                          SV => RT1_NVectors);
  package RT1F_NPeriodic is new RT1_NFouriers1.Periodic(SSV => RT1_NScalVectors);
  subtype RT1_NFourier1 is RT1_NFouriers1.Fourier1;

------------------------------------------------------------------
--- conversion of other types ------------------------------------
------------------------------------------------------------------

  procedure S2N(S: in RT1_Fourier1; N: in out RT1_NFourier1);
  procedure N2S(N: in RT1_NFourier1; S: in out RT1_Fourier1);
  procedure T2R(T: in RT1_NFourier1; R: in out R_NFourier1);
  procedure R2T(R: in R_NFourier1; T: in out RT1_NFourier1);
  procedure T2R(T: in RT1_Fourier1; R: in out R_Fourier1);
  procedure R2T(R: in R_Fourier1; T: in out RT1_Fourier1);
  procedure S2N(S: in R_Fourier1; N: in out R_NFourier1);
  procedure N2S(N: in R_NFourier1; S: in out R_Fourier1);

------------------------------------------------------------------
--- precision etc ------------------------------------------------
------------------------------------------------------------------

  function Effective_Precision(Binary_Digs: Positive) return Positive;
  procedure Relax_Checks;
  procedure Restore_Checks;

  Scalar_Precision: Positive  := Initial_Precision;
  Numeric_Precision: Positive := Initial_Precision;
  Fourier_Parallel:  Boolean  := False;
  
------------------------------------------------------------------
--- the real stuff -----------------------------------------------
------------------------------------------------------------------
  
  procedure Num_ContrMatrix(P: in RT1_Param; F,Ev: in RT1_Fourier1; A: in out RT1_Matrix);
  procedure Num_EigenVec(A: in R_Matrix; Vr,Vi: in out R_Matrix);
  
  Pi2: constant R_Scalar := Flt(4)*ArcCos(Scal(0));  -- 2*Pi
  
end FHN_Periodic;
