with Globals;
use Globals;

pragma Elaborate_All(Globals);

package body Flts is

  package Txt_IO renames Ada.Text_IO;

  function RMin(R1,R2,R3: Flt) return Flt is
  begin
    return RMin(Rmin(R1,R2),R3);
  end RMin;

  function RMax(R1,R2,R3: Flt) return Flt is
  begin
    return RMax(Rmin(R1,R2),R3);
  end RMax;

  function Convert(K: Integer) return Flt is
  begin
    return Flt(K);
  end Convert;

  procedure Swap(P1,P2: in out RadPair) is
    P: constant RadPair := P1;
  begin
    P1 := P2;
    P2 := P;
  end Swap;

  function Between(A,B: Flt; P: Natural; Q: Positive) return Flt is
  begin
    if P=0 then return A; end if;
    if P=Q then return B; end if;
    if P>Q then raise Domain_Error; end if;
    return (Flt(Q-P)*A+Flt(P)*B)/Flt(Q);
  end Between;

  function Scaled(Eps: in Flt; Binary_Digits: in Positive) return Flt is
    use Flt_EF;
  begin
    if Binary_Digits=64 then
      return Eps;
    else
      return Exp(Log(Eps)*Flt(Binary_Digits)/Flt(64));
    end if;
  end Scaled;
  
  function Epsilon(Eps: in Flt; Binary_Digits: in Positive) return Flt is
    Log2:   constant Flt := 0.65625; -- less
    use Flt_EF;
  begin
    if Binary_Digits <= 64 then
      return Eps;
    else
      return Eps*Exp(-Log2*Flt(Binary_Digits-64));
    end if;
  end Epsilon;

------------------------------------------------------
  
  procedure Plot(F: in File_Type; R: in FltVec) is
  begin
    Flt_IO.Put(F,R(R'First),3,5,3);
    for K in R'First+1 .. R'Last loop
      Txt_IO.Put(F," ");
      Flt_IO.Put(F,R(K),3,5,3);
    end loop;
    New_Line(F);
  end Plot;
  
  procedure Plot(FileName: in String; R: in FltVec) is
    F: File_Type;
  begin
    if FileName="" then
      Plot(Standard_Output,R);
      return;
    end if;
    begin
      Open(F,Append_File,FileName);
    exception
      when Name_Error => Create(F,Out_File,FileName);
    end;
    Plot(F,R);
    Close(F);
  end Plot;
  
  procedure Plot2(F: in File_Type; R1,R2: in Flt) is
    R: constant FltVec(1..2) := (R1,R2);
  begin
    Plot(F,R);
  end Plot2;
  
  procedure Plot2(FileName: in String; R1,R2: in Flt) is
    R: constant FltVec(1..2) := (R1,R2);
  begin
    Plot(FileName,R);
  end Plot2;
  
  procedure Plot3(F: in File_Type; R1,R2,R3: in Flt) is
    R: constant FltVec(1..3) := (R1,R2,R3);
  begin
    Plot(F,R);
  end Plot3;
  
  procedure Plot3(FileName: in String; R1,R2,R3: in Flt) is
    R: constant FltVec(1..3) := (R1,R2,R3);
  begin
    Plot(FileName,R);
  end Plot3;
  
  function Max(P: RadVec) return Radius is
    R: Radius := P(P'First);
  begin
    for I in P'First+1 .. P'Last loop
      R := RMax(P(I),R);
    end loop;
    return R;
  end Max;
  
  function Min(P1,P2: RadPair) return RadPair is
  begin
    return (RMin(P1(1),P2(1)),RMin(P1(2),P2(2)));
  end Min;

  function Max(P1,P2: RadPair) return RadPair is
  begin
    return (RMax(P1(1),P2(1)),RMax(P1(2),P2(2)));
  end Max;

  function Between(P1,P2: RadPair; R: Flt := Half) return RadPair is
--- no claim of accuracy
    P: RadPair;
  begin
    P(1) := (One-R)*P1(1)+R*P2(1);
    P(2) := (One-R)*P1(2)+R*P2(2);
    return P;
  end Between;

  function NumScale(P: RadPair; R: Flt) return RadPair is
--- no claim of accuracy
  begin
    return (R*P(1),R*P(2));
  end NumScale;
  
  function Simple_Random return Flt is
    use Ada.Numerics.Float_Random;
  begin
    return Two*Flt(Random(Gen))-One;
  end Simple_Random;
  
  procedure Reset_Random(I: Integer) is
    use Ada.Numerics.Float_Random;
  begin
    Reset(Gen,I);
  end Reset_Random;
  
  function Interpolate(Points,Values: FltVec; X: Flt) return Flt is
    Fac: Flt;
    Y: Flt := Zero;
  begin
    for I in Points'Range loop
      Fac := One;
      for J in Points'Range loop
        if I /= J then
          Fac := Fac*(X-Points(J))/(Points(I)-Points(J));
        end if;
      end loop;
      Y := Y+Values(I)*Fac;
    end loop;
    return Y;
  end Interpolate;
  
end Flts;
