with Ada.Text_IO, Ints, Ada.Numerics.Float_Random;
use Ada.Text_IO, Ints, Ada.Numerics.Float_Random;
with Ada.Numerics.Generic_Elementary_Functions, FltStr, AddETC, Matrices, Vectors;

pragma Elaborate_All(Ada.Text_IO,Ints, Ada.Numerics.Float_Random);
pragma Elaborate_All(Ada.Numerics.Generic_Elementary_Functions,FltStr,AddETC,Matrices,Vectors);

package Flts is

  type Flt is new Long_Long_Float;

  function RMin(R1,R2: Flt) return Flt renames Flt'Min;
  function RMax(R1,R2: Flt) return Flt renames Flt'Max;
  function RMin(R1,R2,R3: Flt) return Flt;
  function RMax(R1,R2,R3: Flt) return Flt;
  function Convert(K: Integer) return Flt;
  function Between(A,B: Flt; P: Natural; Q: Positive) return Flt;
  function Scaled(Eps: in Flt; Binary_Digits: in Positive) return Flt;
  function Epsilon(Eps: in Flt; Binary_Digits: in Positive) return Flt;
  
  package Flt_FS is new FltStr(Flt);
  package Flt_IO renames Flt_FS.Flt_IO;
  package Flt_EF is new Ada.Numerics.Generic_Elementary_Functions(Flt);
  
  procedure Put(P: in File_Access; R: in Flt; F: in Field := Flt_IO.Default_Fore;
                A: in Field := Flt_IO.Default_Aft; E: in Field := Flt_IO.Default_Exp) renames Flt_FS.Put;
  procedure Put(F: in File_Type; R: in Flt; Decimal: in Boolean := False) renames Flt_FS.Put;
  procedure Get(F: in File_Type; R: in out Flt; Decimal: in Boolean := False) renames Flt_FS.Get;
  procedure Show1(N: in String; R: in Flt; NewLine: in Boolean := True) renames Flt_FS.Show1;
  procedure Show2(N: in String; R1,R2: in Flt; NewLine: in Boolean := True) renames Flt_FS.Show2;
  procedure Show3(N: in String; R1,R2,R3: in Flt; NewLine: in Boolean := True) renames Flt_FS.Show3;
  
  package FAE is new AddETC(Number => Flt);
  use FAE;
  package FM is new Matrices (Scalar => Flt);
  package FV is new Vectors (Scalar => Flt, Component => Flt, SM => FM);

  subtype FltVec is FV.Vector;
  subtype FltMat is FM.Matrix;
  
  procedure Plot(F: in File_Type; R: in FltVec);
  procedure Plot(FileName: in String; R: in FltVec);
  procedure Plot2(F: in File_Type; R1,R2: in Flt);
  procedure Plot2(FileName: in String; R1,R2: in Flt);
  procedure Plot3(F: in File_Type; R1,R2,R3: in Flt);
  procedure Plot3(FileName: in String; R1,R2,R3: in Flt);
  function Interpolate(Points,Values: FltVec; X: Flt) return Flt;

  Zero:    constant Flt := Flt(0);
  One:     constant Flt := Flt(1);
  Two:     constant Flt := Flt(2);
  Three:   constant Flt := Flt(3);
  Four:    constant Flt := Flt(4);
  Five:    constant Flt := Flt(5);
  NegOne:  constant Flt := Flt(-1);
  Half:    constant Flt := One/Two;
  NegHalf: constant Flt := NegOne/Two;
  Half3:   constant Flt := Three/Two;
  Quarter: constant Flt := One/Four;

  subtype Radius is Flt range Zero .. Flt'Last;
  type RadVec is array(Integer range <>) of Radius;
  subtype RadPair is RadVec(1..2);
  
  function Max(P: RadVec) return Radius;
  function Min(P1,P2: RadPair) return RadPair;
  function Max(P1,P2: RadPair) return RadPair;
  procedure Swap(P1,P2: in out RadPair);
  function Between(P1,P2: RadPair; R: Flt := Half) return RadPair; -- purely numerical
  function NumScale(P: RadPair; R: Flt) return RadPair;            -- purely numerical
  function Simple_Random return Flt;                               -- in [-1,1]
  procedure Reset_Random(I: Integer);
  
  Zero_RadPair: constant RadPair := (Zero,Zero);
  Gen: Generator;
  
  pragma Inline_Always (Min,Max,Swap,Simple_Random);
  
end Flts;
