with Param, ScalVectors, ScalVectors.RNum;
use Param;

pragma Elaborate_All (Param,ScalVectors,ScalVectors.RNum);

generic

   with package SSV is new Scalvectors (Scalar => Scalar, SM => SM, SV => SV);

package Fouriers1.Periodic is

   package SSVN is new SSV.RNum;
   use SM,SV,SSV;
   subtype Param is Vector2;
   
--- existence
   procedure FixAve(F: in out Fourier1);
   procedure DFixAve(F: in Fourier1; H: in out Fourier1);
   function Kappa(P: Param) return Scalar;
   procedure MapA(P: in Param; F: in Fourier1; G: in out Fourier1);
   procedure DMapA(P: in Param; F: in Fourier1; H: in out Fourier1);
   procedure Compo1(F0,F: in Fourier1; S: in out Scalar);
   function Compo1(F0,F: Fourier1) return Scalar;
   procedure Proj1(F0: in Fourier1; F: in out Fourier1);
   procedure Proj2(F0: in Fourier1; F: in out Fourier1);
   procedure Map(P: in Param; F,G: in Fourier1; H: in out Fourier1);
   procedure DMap(P: in Param; F,G: in Fourier1; H: in out Fourier1);
   
   procedure Pack(A: in Fourier1; V: in out Vector);
   procedure UnPack(V: in Vector; R: in Radius; A: in out Fourier1);
   procedure Pack(A1,A2: in Fourier1; V: in out Vector);
   procedure UnPack(V: in Vector; R: in Radius; A1,A2: in out Fourier1);
   procedure Prod(A: in Matrix; F: in Fourier1; G: in out Fourier1);
   function OpNorm(R: Radius; M: Matrix) return Radius;
   procedure Contract(P: in Param; F,Ev: in Fourier1; A: in Matrix; H: in out Fourier1; Err: out Radius);
   function DContractNorm(P: Param; F,Ev: Fourier1; A: Matrix) return Radius;
   procedure CheckFix(P: in Param; A: in Matrix; F: in out Fourier1; Ev: in Fourier1; KMax: in Radius);
   
--- spectrum
   procedure MatrixL0(P: in Param; A: in out Matrix);
   procedure MatrixF(F: in Fourier1; A: in out Matrix);
   function NormL0H(P: Param; K: Positive; Y: Scalar) return Radius;
   function NormA(F: Fourier1; U: Diag; C2: Radius) return Radius;
   
--- numerical
   procedure Der(F: in Fourier1; G: in out Fourier1);
   procedure Der2(F: in Fourier1; G: in out Fourier1);
   procedure DerMatrix(P: in Param; F,Ev: in Fourier1; A: in out Matrix);
   procedure DerMatrix(P: in Param; F: in Fourier1; A: in out Matrix);
   procedure ContrMatrix(P: in Param; F,Ev: in Fourier1; A: in out Matrix);
   procedure ShowEigen(A: in Matrix);
   procedure Contract0(P: in Param; A: in Matrix; F: in out Fourier1; Ev: in Fourier1; Err: out Radius);
   procedure ApproxFix(P: in Param; F: in out Fourier1; Ev: in Fourier1; A: in out Matrix; Eps: in Radius);
   procedure Plot(F: in Fourier1; L: Positive := 50);
   procedure Plot(FileName: in String; F: in Fourier1; L: Positive := 50);

private

   procedure D0Inv(F: in Fourier1; G: in out Fourier1);
   procedure D2mKInv(Kappa: in Scalar; F: in Fourier1; G: in out Fourier1);
   procedure Cubic(F: in Fourier1; G: in out Fourier1);
   procedure DCubic(F: in Fourier1; G: in out Fourier1);
   procedure NumCompo1(F0,F: in Fourier1; D: in Natural; S: in out Scalar);

   S_Epsilon: constant Scalar := Scal(Q_Epsilon);
   S_Gamma:   constant Scalar := Scal(Q_Gamma);
   S_Alpha:   constant Scalar := Scal(Q_Alpha);
   S_Alpha1:  constant Scalar:= Scal(Q_Alpha+1);
   
end Fouriers1.Periodic;
