package body MPFR.Floats.Rounding is
  
  procedure Set_Rounding_Mode(Dummy: in MPFloat; M: in Rounding_Mode) is
    pragma Unreferenced (Dummy);
  begin
    Clear_Flags;
    Check_Flags := (M.Mask=Strict);
    case M.Direction is
      when Up        => RoundingDirection := GMP_RNDU;
      when Nearest   => RoundingDirection := GMP_RNDN;
      when Down      => RoundingDirection := GMP_RNDD;
      when Truncate  => RoundingDirection := GMP_RNDZ;
    end case;
  end Set_Rounding_Mode;
  
  function ULP_Weight(V: MPFR_T) return LLFloat is
--- weight of the ulp (unit in the last place)
--- both IEEE and MPFR represent 1 as binary(0.1E+1)
    V1: MPFR_Struct renames V(1);
  begin
    if MPFR_Zero_P(V) /= 0 then
      raise MPFR_Exception with "underflow - should not happen";
    end if;
    return LLFloat'Compose(LLFloat(1),1-Int(V1.MP_Prec)+Int(V1.MPFR_Exp));
  end ULP_Weight;
  
  procedure Enclose(S: in MPFloat; C,R: out LLFloat) is
--- assuming LLFloat ops are rounded up
    use Std_Floats;
  begin
    C := LLFloat(MPFR_Get_Ld(S.Val,GMP_RNDU));
    R := C-LLFloat(MPFR_Get_Ld(S.Val,GMP_RNDD));
    if Problem then Report("Enclose"); end if;
  end Enclose;
  
  procedure AddErr(I: in Integer; S: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Adjust_Precision(S.Val);
    MPFR_Add_Si(Rounded,S.Val,S.Val,Long(I),GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S.Val); end if;
    if Problem then Report("AddErr 1"); end if;
  end AddErr;

  procedure AddErr(S1: in MPFloat; S2: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Adjust_Precision(S2.Val);
    MPFR_Add(Rounded,S2.Val,S2.Val,S1.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S2.Val); end if;
    if Problem then
      Show2("",S1,S2);
      Report("AddErr 2");
    end if;
  end AddErr;

  procedure SumErr(S1,S2: in MPFloat; S3: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Reformat(S3.Val);
    MPFR_Add(Rounded,S3.Val,S1.Val,S2.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S3.Val); end if;
    if Problem then Report("SumErr"); end if;
  end SumErr;

  procedure SubErr(S1: in MPFloat; S2: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Adjust_Precision(S2.Val);
    MPFR_Sub(Rounded,S2.Val,S2.Val,S1.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S2.Val); end if;
    if Problem then Report("AddErr"); end if;
  end SubErr;

  procedure DiffErr(S1,S2: in MPFloat; S3: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Reformat(S3.Val);
    MPFR_Sub(Rounded,S3.Val,S1.Val,S2.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S3.Val); end if;
    if Problem then Report("SumErr"); end if;
  end DiffErr;

  procedure MultErr(R: in Flt; S,Tmp: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Reformat(Tmp.Val);
    MPFR_Set_Ld(Rounded,Tmp.Val,Long_Double(R),GMP_RNDN);
    Adjust_Precision(S.Val);
    MPFR_Mul(Rounded,S.Val,Tmp.Val,S.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S.Val); end if;
    if Problem then Report("MultErr"); end if;
  end MultErr;

  procedure ProdErr(R: in Flt; S1: in MPFloat; S2: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Reformat(S2.Val);
    MPFR_Set_Ld(Rounded,S2.Val,Long_Double(R),GMP_RNDN);
    MPFR_Mul(Rounded,S2.Val,S1.Val,S2.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S2.Val); end if;
    if Problem then Report("ProdErr"); end if;
  end ProdErr;

  procedure MultErr(S1: in MPFloat; S2: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Adjust_Precision(S2.Val);
    MPFR_Mul(Rounded,S2.Val,S2.Val,S1.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S2.Val); end if;
    if Problem then Report("MultErr"); end if;
  end MultErr;

  procedure ProdErr(S1,S2: in MPFloat; S3: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Reformat(S3.Val);
    MPFR_Mul(Rounded,S3.Val,S1.Val,S2.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S3.Val); end if;
    if Problem then Report("ProdErr"); end if;
  end ProdErr;

  procedure AddProdErr(R: in Flt; S1: in MPFloat; S2,Tmp: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Reformat(Tmp.Val);
    MPFR_Set_Ld(Rounded,Tmp.Val,Long_Double(R),GMP_RNDN);
    Adjust_Precision(S2.Val);
    MPFR_FMA(Rounded,S2.Val,Tmp.Val,S1.Val,S2.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S2.Val); end if;
    if Problem then Report("AddProdErr"); end if;
  end AddProdErr;

  procedure AddProdErr(S1,S2: in MPFloat; S3,Tmp: in out MPFloat; Err: out LLFloat) is
--- not using slow MPFR_FMA
    use Std_Floats;
    Rounded: Int;
  begin
    Reformat(Tmp.Val);
    MPFR_Mul(Rounded,Tmp.Val,S1.Val,S2.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(Tmp.Val); end if;
    Adjust_Precision(S3.Val);
    MPFR_Add(Rounded,S3.Val,S3.Val,Tmp.Val,GMP_RNDN);
    if Rounded /= 0 then Err := Err+ULP_Weight(S3.Val); end if;
    if Problem then Report("AddProdErr"); end if;
  end AddProdErr;

  procedure AddProdErr(S1,S2: in MPFloat; S3: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Adjust_Precision(S3.Val);
    MPFR_FMA(Rounded,S3.Val,S1.Val,S2.Val,S3.Val,GMP_RNDN);  --- slow!
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S3.Val); end if;
    if Problem then Report("AddProdErr"); end if;
  end AddProdErr;

  procedure SumProdErr(S1,S2,S3: in MPFloat; S4: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Reformat(S4.Val);
    MPFR_FMA(Rounded,S4.Val,S1.Val,S2.Val,S3.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S4.Val); end if;
    if Problem then Report("SumProdErr"); end if;
  end SumProdErr;

  procedure SubProdErr(S1,S2: in MPFloat; S3,Tmp: in out MPFloat; Err: out LLFloat) is
--- not using slow MPFR_FMA
    use Std_Floats;
    Rounded: Int;
  begin
    Reformat(Tmp.Val);
    MPFR_Mul(Rounded,Tmp.Val,S1.Val,S2.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(Tmp.Val); end if;
    Adjust_Precision(S3.Val);
    MPFR_Sub(Rounded,S3.Val,S3.Val,Tmp.Val,GMP_RNDN);
    if Rounded /= 0 then Err := Err+ULP_Weight(S3.Val); end if;
    if Problem then Report("SubProdErr"); end if;
  end SubProdErr;

  procedure DivErr(R: in Flt; S,Tmp: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Reformat(Tmp.Val);
    MPFR_Set_Ld(Rounded,Tmp.Val,Long_Double(R),GMP_RNDN);
    Adjust_Precision(S.Val);
    MPFR_Div(Rounded,S.Val,S.Val,Tmp.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S.Val); end if;
    if Problem then Report("DivErr"); end if;
  end DivErr;

  procedure QuotErr(S1: in MPFloat; R: in Flt; S2: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Reformat(S2.Val);
    MPFR_Set_Ld(Rounded,S2.Val,Long_Double(R),GMP_RNDN);
    MPFR_Div(Rounded,S2.Val,S1.Val,S2.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S2.Val); end if;
    if Problem then Report("QuotErr"); end if;
  end QuotErr;

  procedure InvErr(S1: in MPFloat; S2: in out MPFloat; Err: out LLFloat) is
    Rounded: Int;
  begin
    Reformat(S2.Val);
    MPFR_Set_Si(Rounded,S2.Val,1,GMP_RNDN);
    MPFR_Div(Rounded,S2.Val,S2.Val,S1.Val,GMP_RNDN);
    if Rounded=0 then Err := LLFloat(0); else Err := ULP_Weight(S2.Val); end if;
    if Problem then Report("InvErr"); end if;
  end InvErr;
  
  procedure Root(K: in Positive; X,DX: in MPFloat; YL,YU: in out MPFloat) is
    A: MPFR_T renames YU.Val;
  begin
    Reformat(YL.Val);
    Reformat(YU.Val);
    MPFR_Abs(Ternary,A,DX.Val,GMP_RNDU);
    MPFR_Sub(Ternary,YL.Val,X.Val,A,GMP_RNDD);
    MPFR_Root(Ternary,YL.Val,YL.Val,Unsigned_Long(K),GMP_RNDD);
    MPFR_Add(Ternary,YU.Val,X.Val,A,GMP_RNDU);
    MPFR_Root(Ternary,YU.Val,YU.Val,Unsigned_Long(K),GMP_RNDU);
    if Problem then Report("Exp"); end if;
  end Root;
  
  procedure ExpUp(X: in MPFloat; Y: in out MPFloat) is
  begin
    Reformat(Y.Val);
    MPFR_Exp(Ternary,Y.Val,X.Val,GMP_RNDU);
    if Check_Flags and then Problem then Report("ExpUp"); end if;
  end ExpUp;
  
  procedure ExpDown(X: in MPFloat; Y: in out MPFloat) is
  begin
    Reformat(Y.Val);
    MPFR_Exp(Ternary,Y.Val,X.Val,GMP_RNDD);
    if Check_Flags and then Problem then Report("ExpDown"); end if;
  end ExpDown;
  
  procedure Exp(X,DX: in MPFloat; YL,YU: in out MPFloat) is
    A: MPFR_T renames YU.Val;
  begin
    Reformat(YL.Val);
    Reformat(YU.Val);
    MPFR_Abs(Ternary,A,DX.Val,GMP_RNDU);
    MPFR_Sub(Ternary,YL.Val,X.Val,A,GMP_RNDD);
    MPFR_Exp(Ternary,YL.Val,YL.Val,GMP_RNDD);
    MPFR_Add(Ternary,YU.Val,X.Val,A,GMP_RNDU);
    MPFR_Exp(Ternary,YU.Val,YU.Val,GMP_RNDU);
    if Problem then Report("Exp"); end if;
  end Exp;
  
  procedure Log(X,DX: in MPFloat; YL,YU: in out MPFloat) is
    A: MPFR_T renames YU.Val;
  begin
    Reformat(YL.Val);
    Reformat(YU.Val);
    MPFR_Abs(Ternary,A,DX.Val,GMP_RNDU);
    MPFR_Sub(Ternary,YL.Val,X.Val,A,GMP_RNDD);
    MPFR_Log(Ternary,YL.Val,YL.Val,GMP_RNDD);
    MPFR_Add(Ternary,YU.Val,X.Val,A,GMP_RNDU);
    MPFR_Log(Ternary,YU.Val,YU.Val,GMP_RNDU);
    if Problem then Report("Log"); end if;
  end Log;

  procedure ArcCos(X,DX: in MPFloat; YL,YU: in out MPFloat) is
    A: MPFR_T renames YU.Val;
  begin
    Reformat(YL.Val);
    Reformat(YU.Val);
    MPFR_Abs(Ternary,A,DX.Val,GMP_RNDU);
    MPFR_Add(Ternary,YL.Val,X.Val,A,GMP_RNDU);
    MPFR_ACos(Ternary,YL.Val,YL.Val,GMP_RNDD);
    MPFR_Sub(Ternary,YU.Val,X.Val,A,GMP_RNDD);
    MPFR_ACos(Ternary,YU.Val,YU.Val,GMP_RNDU);
    if Problem then Report("ArcCos"); end if;
  end ArcCos;

  procedure ArcSin(X,DX: in MPFloat; YL,YU: in out MPFloat) is
    A: MPFR_T renames YU.Val;
  begin
    Reformat(YL.Val);
    Reformat(YU.Val);
    MPFR_Abs(Ternary,A,DX.Val,GMP_RNDU);
    MPFR_Sub(Ternary,YL.Val,X.Val,A,GMP_RNDD);
    MPFR_ASin(Ternary,YL.Val,YL.Val,GMP_RNDD);
    MPFR_Add(Ternary,YU.Val,X.Val,A,GMP_RNDU);
    MPFR_ASin(Ternary,YU.Val,YU.Val,GMP_RNDU);
    if Problem then Report("ArcSin"); end if;
  end ArcSin;

  procedure Cos(X: in MPFloat; YL,YU: in out MPFloat) is
  begin
    Reformat(YL.Val);
    Reformat(YU.Val);
    MPFR_Cos(Ternary,YL.Val,X.Val,GMP_RNDD);
    MPFR_Cos(Ternary,YU.Val,X.Val,GMP_RNDU);
    if Problem then Report("Cos"); end if;
  end Cos;

  procedure Sin(X: in MPFloat; YL,YU: in out MPFloat) is
  begin
    Reformat(YL.Val);
    Reformat(YU.Val);
    MPFR_Sin(Ternary,YL.Val,X.Val,GMP_RNDD);
    MPFR_Sin(Ternary,YU.Val,X.Val,GMP_RNDU);
    if Problem then Report("Sin"); end if;
  end Sin;
  
end MPFR.Floats.Rounding;
