with Param, ScalVectors.Homoclinic, Ada.Unchecked_Deallocation;
use Param;

pragma Elaborate_All (Param,ScalVectors.Homoclinic,Ada.Unchecked_Deallocation);

generic
  
  with procedure HiErrs(S: in out Scalar) is <>;
  
package MultiTaylors1.Homoclinic is
  
  package SSVH is new SSV.Homoclinic;
  use ST1, SSVH;
  
  subtype Scalar_Cubic1 is SSVH.Cubic1;
  type Simple_Operator1 is array (0..PDeg) of Scalar_Matrix3; -- first two unused
  
  procedure Mult(T: in Simple_Operator1; M,N: in OneTwoThree; P: in out Taylor1_Vector3);
  procedure Mult(T: in Simple_Operator1; P: in out Taylor1_Vector3);
  procedure Local_Stable(D: in FHN_Data0; Y0: in Flt; W: in out Scalar_Vector3; Phi: in out Taylor1_Vector3);
  procedure Local_Stable(C: in Scalar; Y0: in Flt; W: in out Scalar_Vector3; Phi: in out Taylor1_Vector3);
  procedure Vector_Field(C: in Scalar; F: in Scalar_Cubic1; P: in Taylor1_Vector3; V: in out Taylor1_Vector3);
  
  generic  -- vector field is V0+V1 with V1 vanishing at the origin
    with procedure V1(P: in out Taylor1_Vector3);
  procedure Flow(V0: in Scalar_Vector3; P: in out Taylor1_Vector3);
  
  procedure FHN_Flow(C: in Scalar; DY: in Flt; W: in out Scalar_Vector3; Phi: in out Taylor1_Vector3);
  procedure Plot_Stable(C: in Scalar; Y0,YL: in Flt);
  function Times(N: in Natural) return FltVec;
  procedure Stable(C: in Scalar; Y: in FltVec; W: in out Scalar_Vector3);
  procedure Write_Ws(FileName: in String; C: in Scalar);
  procedure Write_Stable(FileName: in String; Y: in FltVec; C: in Scalar);
  
  generic
    with procedure Inspect(T: in Flt; P: in Taylor1_Vector3);
  procedure Follow_Stable(C: in Scalar; Y: in FltVec; W: in out Scalar_Vector3);
  
  type Taylor1_Vector3_Array is array(Integer range <>) of Taylor1_Vector3;
  
  type Curve3(L: Natural) is
    record
      C: Scalar;
      Y: FltVec(0..L);
      P: Taylor1_Vector3_Array(0..L);
    end record;
  
  type Curve3_Access is access Curve3;
  
  procedure Read(FileName: in String; C: in out Curve3);
  function Read(FileName: in String) return Curve3;
  procedure Read(FileName: in String; A: in out Curve3_Access);
  procedure Free is new Ada.Unchecked_Deallocation(Curve3,Curve3_Access);
  procedure Show_Curve3(FileName: in String);
  
--- now Scalar needs to be complex (complex Taylor1)
  
  procedure Spec_Local_Stable(D: in FHN_DataZ; Phi2: in Taylor1; U: in out Taylor1_Vector3);
  procedure Spec_Local_Stable(D: in FHN_DataZ; Phi2: in Taylor1; Y0: in Flt; UY: in out Scalar_Vector3);
  procedure FHN_Spec_Flow(D: in FHN_DataZ; Phi2: in Taylor1; DY: in Flt; UY: in out Scalar_Vector3);
  
  generic
    with procedure Phi_Segment(L: in Natural; DY: out Flt; Phi2: in out Taylor1);
  procedure Spec_Stable(C,Z: in Scalar; N: in Natural; UY: in out Scalar_Vector3);
  
private
  
  procedure InvDyMinusM(Mu0: in Scalar; M: in Scalar_Matrix3; T: in out Simple_Operator1);
  procedure FpDiff(Phi2: in Taylor1; P12,PTmp: in out Taylor1);
  
  IY: constant IntVec(0..26) := (10,5,1,-3,-7,-12,-17,-23,-30,-38, -46, -53,-58,-63,-67,
                                 -71,-75,-79,-84,-89,-95,-102,-111,-122,-135,-151,-172);
  
end MultiTaylors1.Homoclinic;
