
package body MultiTaylors1 is
  
  use SV_More;
  
  procedure SetRho(R: in Radius; P: in out Taylor1_Vector; Check: in Boolean := True) is
  begin
    for N in P'Range loop SetRho(R,P(N),Check); end loop;
  end SetRho;
  
  function Rho(P: in Taylor1_Vector) return Radius is
    R: Radius := Radius'Last;
  begin
    for N in P'Range loop R := RMin(Rho(P(N)),R); end loop;
    return R;
  end Rho;
  
  function GuessRho(P: in Taylor1_Vector) return Radius is
    R: Radius := Radius'Last;
  begin
    for N in P'Range loop R := RMin(GuessRho(P(N)),R); end loop;
    return R;
  end GuessRho;
  
  procedure Components(I: in Natural; P: in Taylor1_Vector; W: in out Scalar_Vector) is
  begin
    for N in W'Range loop Copy(P(N).C(I),W(N)); end loop;
  end Components;
  
  procedure ZeroComponents(I: in Natural; P: in out Taylor1_Vector) is
  begin
    for N in P'Range loop SetZero(P(N).C(I)); end loop;
  end ZeroComponents;
  
  procedure AddCoeffs(I: in Natural; W: in Scalar_Vector; P: in out Taylor1_Vector) is
  begin
    for N in P'Range loop AddCoeff(I,W(N),P(N)); end loop;
  end AddCoeffs;
  
  procedure ErrMult(I: in Natural; R: in Flt; P: in out Taylor1_Vector) is
  begin
    for N in P'Range loop ErrMult(R,P(N).C(I)); end loop;
  end ErrMult;
  
  procedure Val(P: in Taylor1_Vector; S: in Scalar; W: in out Scalar_Vector) is
  begin
    for N in W'Range loop Val(P(N),S,W(N)); end loop;
  end Val;
  
  procedure ValDer(P: in Taylor1_Vector; S: in Scalar; W: in out Scalar_Vector) is
  begin
    for N in W'Range loop ValDer(P(N),S,W(N)); end loop;
  end ValDer;
  
  procedure Plot_Arc(F: in File_Type; Y: in Flt; P: in Taylor1_Vector3; Steps: in Positive) is
    use SV;
    YN: Flt;
    W: Scalar_Vector3;
  begin
    for N in 0 .. Steps loop
      YN := Y*Flt(N)/Flt(Steps);
      Val(P,Scal(YN),W);
      Plot3(F,Approx(W(1)),Approx(W(2)),Approx(W(3)));
    end loop;
    if Verbosity>4 then
      Val(P,Scal(Y),W);
      Put(Standard_Output,W);
    end if;
  end Plot_Arc;
  
end MultiTaylors1;
