with Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts, VGeneric;
use Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts, VGeneric;
with Recycling, Matrices, Vectors, Vectors.More, ScalVectors, Taylors1;

pragma Elaborate_All (Ada.Text_IO,Logicals,Globals,Ints,Rationals,Flts,VGeneric);
pragma Elaborate_All (Recycling,Matrices,Vectors,Vectors.More,ScalVectors,Taylors1);

generic
  
  PDeg: in Positive;  --- max degree
  
-------------------- begin_include scalars Scalar

  type Scalar is private;

--- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

--- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

--- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

--- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

--- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

--- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

--- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  with procedure Set_Active_Mode(Dummy: in Scalar) is <>;
  with procedure Set_Wait_Mode(Dummy: in Scalar) is <>;
  with procedure Free_Cache(S: in Scalar) is <>;
-------------------- end_include scalars
  
  with package SM is new Matrices (Scalar => Scalar);
  with package SV is new Vectors (Scalar => Scalar, Component => Scalar, SM => SM);
  with package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  with package ST1 is new Taylors1 (PDeg => PDeg, Scalar => Scalar, SM => SM, SV => SV);
  
package MultiTaylors1 is
  
  subtype Poly1 is ST1.Poly1;
  subtype Taylor1 is ST1.Taylor1;
  
  use ST1;
  package STV is new Vectors (Scalar => Scalar, Component => Taylor1, SM => SM);
  package STV_More is new STV.More;
  
  subtype Scalar_Matrix is SM.Matrix;
  subtype Scalar_Matrix2 is SM.Matrix2;
  subtype Scalar_Matrix3 is SM.Matrix3;
  subtype Scalar_Vector is SV.Vector;
  subtype Scalar_Vector2 is SSV.Vector2;
  subtype Scalar_Vector3 is SSV.Vector3;
  subtype Scalar_Polynom1 is Scalar_Vector;
  subtype Taylor1_Vector is STV.Vector;
  subtype Taylor1_Vector2 is Taylor1_Vector(1..2);
  subtype Taylor1_Vector3 is Taylor1_Vector(1..3);
  
  procedure SetRho(R: in Radius; P: in out Taylor1_Vector; Check: in Boolean := True);
  function Rho(P: in Taylor1_Vector) return Radius;
  function GuessRho(P: in Taylor1_Vector) return Radius;
  procedure HiErrs is new VP_Modify1 (Taylor1,Taylor1_Vector,HiErrs);
  procedure Components(I: in Natural; P: in Taylor1_Vector; W: in out Scalar_Vector);
  procedure ZeroComponents(I: in Natural; P: in out Taylor1_Vector);
  procedure AddCoeffs(I: in Natural; W: in Scalar_Vector; P: in out Taylor1_Vector);
  procedure ErrMult(I: in Natural; R: in Flt; P: in out Taylor1_Vector);
  procedure AntiDer is new VP_Modify1 (Taylor1,Taylor1_Vector,AntiDer);
  procedure NonConst is new VP_Modify1 (Taylor1,Taylor1_Vector,NonConst);
  procedure Real_Part is new VP_Modify1 (Taylor1,Taylor1_Vector,Real_Part);
  procedure Val(P: in Taylor1_Vector; S: in Scalar; W: in out Scalar_Vector);
  procedure ValDer(P: in Taylor1_Vector; S: in Scalar; W: in out Scalar_Vector);
  procedure Plot_Arc(F: in File_Type; Y: in Flt; P: in Taylor1_Vector3; Steps: in Positive);
  
private
  
  SZero:   constant Scalar      := Scal(0);
  SInfo:   constant Scalar_Info := Info(SZero);
  STrunc:  constant Boolean     := SInfo.IsNumeric;
  
  package Alloc1 is new Recycling (Item => Taylor1);
  subtype LT_Pointer1 is Alloc1.LD_Pointer;
  Pool1: Alloc1.Protected_Data;
  
  package Alloc3 is new Recycling (Item => Taylor1_Vector3);
  subtype LT_Pointer3 is Alloc3.LD_Pointer;
  Pool3: Alloc3.Protected_Data;
  
end MultiTaylors1;
