
package body MultiTaylors2 is
  
  use SV_More;
  
  procedure SetRho(R: in Radius; P: in out Taylor2_Vector; Check: in Boolean := True) is
  begin
    for N in P'Range loop SetRho(R,P(N),Check); end loop;
  end SetRho;
  
  procedure Components(I,J: in Natural; P: in Taylor2_Vector; W: in out Scalar_Vector) is
  begin
    for N in W'Range loop Copy(P(N).C(I,J),W(N)); end loop;
  end Components;
  
  procedure ZeroComponents(I,J: in Natural; P: in out Taylor2_Vector) is
  begin
    for N in P'Range loop SetZero(P(N).C(I,J)); end loop;
  end ZeroComponents;
  
  procedure AddCoeffs(I,J: in Natural; W: in Scalar_Vector; P: in out Taylor2_Vector) is
  begin
    for N in P'Range loop AddCoeff(I,J,W(N),P(N)); end loop;
  end AddCoeffs;
  
  procedure ErrMult(I,J: in Natural; R: in Flt; P: in out Taylor2_Vector) is
  begin
    for N in P'Range loop ErrMult(R,P(N).C(I,J)); end loop;
  end ErrMult;
  
  procedure Val(P: in Taylor2_Vector; S1,S2: in Scalar; W: in out Scalar_Vector) is
  begin
    for N in W'Range loop Val(P(N),S1,S2,W(N)); end loop;
  end Val;
  
  procedure ValDer1(P: in Taylor2_Vector; S1,S2: in Scalar; W1: in out Scalar_Vector) is
  begin
    for N in W1'Range loop ValDer1(P(N),S1,S2,W1(N)); end loop;
  end ValDer1;
  
  procedure ValDer2(P: in Taylor2_Vector; S1,S2: in Scalar; W2: in out Scalar_Vector) is
  begin
    for N in W2'Range loop ValDer2(P(N),S1,S2,W2(N)); end loop;
  end ValDer2;
  
  procedure ValDer(P: in Taylor2_Vector; S1,S2: in Scalar; W1,W2: in out Scalar_Vector) is
  begin
    for N in W1'Range loop ValDer(P(N),S1,S2,W1(N),W2(N)); end loop;
  end ValDer;
  
  procedure LiftErrs(P: in out Taylor2_Vector; F: in Natural := 0) is
  begin
    for N in P'Range loop LiftErrs(P(N),F); end loop;
  end LiftErrs;
  
  procedure Plot(F: in File_Type; R: in Radius; P: in Taylor2_Vector3; Steps: in Positive) is
    Pi2: constant Flt := 6.283185307179586476925286766559;
    I: constant Scalar := Sqrt(Scal(-1));
    NMax: Integer;
    RM,Phi: Flt;
    S: Scalar;
    W: Scalar_Vector3;
    use Flt_EF;
  begin
    for M in 1 .. Steps loop
      RM := R*Flt(M)/Flt(Steps);
      NMax := 5+M;
      for N in 0 ..    NMax loop
        Phi := Pi2*Flt(N)/Flt(NMax);
        S := RM*Exp(Phi*I);
        Val(P,S,Adjoint(S),W);
        Plot3(F,Approx(W(1)),Approx(W(2)),Approx(W(3)));
      end loop;
    end loop;
  end Plot;
  
end MultiTaylors2;
