with Strings, Ints, Globals;
use Strings, Ints, Globals;

pragma Elaborate_All (Strings,Ints,Globals);

package body Paths is
  
  procedure Show_Segment(N: in String; NC,NR,Status: in Natural) is
    X,Y: Integer;
  begin
    NXY(NC,X,Y);
    Show0(N & Strng(NC,6) & Strng(NR,5) & Strng(X,5) & Strng(Y,5),False);
    if Status>1 then
      Show0("  checked");
    elsif Status=1 then
      Show0("     busy");
    elsif Status<0 then
      Show0("   failed");
    else
      Show0;
    end if;
  end Show_Segment;
      
  procedure Show(N: in String; S: in Segment) is
  begin
    Show0(N & "    NC   NR    X    Y   Status");
    Show_Segment(N,S.NC,S.NR,S.Status);
  end Show;
    
  function N2R(N: Integer) return Flt is
  begin
    return Flt(N)*Inv_R_Fac;
  end N2R;
  
  function R2N(R: Flt) return Integer is
    use Globals;
    N: constant Integer := Integer(R*R_Fac);
  begin
    if R /= Flt(N)*Inv_R_Fac then
      Show1("Paths.R2N: please choose a multiple of",Inv_R_Fac);
      raise Sorry;
    end if;
    return N;
  end R2N;
  
  procedure NXY(N: in Natural; X,Y: out Integer) is
    use Globals;
  begin
    if N <= N_TLeft then
      X := NX_Min;
      Y := N;
    elsif N <= N_TRight then
      X := NX_Min+(N-N_TLeft);
      Y := NY_Max;
    elsif N <= N_End then
      X := NX_Max;
      Y := NY_Max-(N-N_TRight);
    else
      raise Domain_Error with "N_XY error: index out of range";
    end if;
  end NXY;
  
  procedure RXY(N: in Natural; X,Y: out Flt) is
    NX,NY: Integer;
  begin
    NXY(N,NX,NY);
    X := Flt(NX)*Inv_R_Fac;
    Y := Flt(NY)*Inv_R_Fac;
  end RXY;
  
  procedure ShowCorners is
  begin
    Show0("              NC   NR    X    Y   Status");
    Show_Segment("low  left ",N_Start,0,0);
    Show_Segment("top  left ",N_TLeft,0,0);
    Show_Segment("top right ",N_TRight,0,0);
    Show_Segment("low right ",N_End,0,0);
  end ShowCorners;
  
  procedure Show(N: in String; P: in Path) is
  begin
    Show0(N & "    NC   NR    X    Y   Status");
    for I in P'Range loop
      Show_Segment(N,P(I).NC,P(I).NR,P(I).Status);
    end loop;
  end Show;
  
  procedure Write(FileName: in String; P: in Path; L: in Positive) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,L);
    for I in 1 .. L loop
      Put(F,P(I).NC);
      Put(F,P(I).NR);
      Put(F,P(I).Status);
    end loop;
    Close(F);
  end Write;
  
  procedure Write(FileName: in String; P: in Path) is
  begin
    Write(FileName,P,P'Last);
  end Write;
  
  procedure Read(FileName: in String; P: in out Path) is
    L: Positive := 1;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,L);
    if L /= P'Last then
      raise Sorry with "Paths.Read: dimension mismatch";
    end if;
    for I in 1 .. L loop
      Get(F,P(I).NC);
      Get(F,P(I).NR);
      Get(F,P(I).Status);
    end loop;
    Close(F);
  end Read;
  
  function Read(FileName: in String) return Path is
    pragma Warnings (Off); -- see below
    L: Positive := 1;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,L);
    declare
      P: Path(1..L);
    begin
      for I in 1 .. L loop
        Get(F,P(I).NC);   --- 2nd argument is effectively "out"
        Get(F,P(I).NR);   --- 2nd argument is effectively "out"
        Get(F,P(I).Status);
      end loop;
      Close(F);
      return P;
    end;
  end Read;
  
  function First_Status0(P: Path) return Natural is
  begin
    for I in P'Range loop
      if P(I).Status=0 then return I; end if;
    end loop;
    return 0;
  end First_Status0;
  
  function Min_Status(P: Path) return Integer is
    S: Integer := Integer'Last;
  begin
    for I in P'Range loop
      S := IMin(P(I).Status,S);
    end loop;
    return S;
  end Min_Status;
  
end Paths;
