with Ada.Text_IO, Flts;
use Ada.Text_IO, Flts;

package Paths is --- rectangular path ...
  
  type Segment is
    record
      NC: Natural;          --- center
      NR: Natural;          --- half diameter
      Status: Integer := 0;
    end record;
  
  procedure Show(N: in String; S: in Segment);
  
  RX_Min: constant Flt     := Zero;
  RX_Max: constant Flt     := Flt(2485)/Flt(8192);   --- 0.3033447265625
  RY_Min: constant Flt     := Zero;
  RY_Max: constant Flt     := Flt(5857)/Flt(16384);  --- 0.35748291015625
  N_Fac:  constant Natural := 2**14;
  
  R_Fac:  constant Flt     := Flt(N_Fac);
  NX_Min: constant Integer := Integer(RX_Min*R_Fac);
  NX_Max: constant Integer := Integer(RX_Max*R_Fac);
  NY_Min: constant Integer := Integer(RY_Min*R_Fac);
  NY_Max: constant Integer := Integer(RY_Max*R_Fac);
    
  N_Start:  constant Natural := 0;
  N_TLeft:  constant Natural := NY_Max-NY_Min;
  N_TRight: constant natural := N_TLeft+(NX_Max-NX_Min);
  N_End:    constant Natural := N_TRight+N_TLeft;
  
  function N2R(N: Integer) return Flt;
  function R2N(R: Flt) return Integer;
  procedure NXY(N: in Natural; X,Y: out Integer);
  procedure RXY(N: in Natural; X,Y: out Flt);
  procedure ShowCorners;
  
  type Path is array (Positive range <>) of Segment; --- starting with index 1
  
  procedure Show(N: in String; P: in Path);
  procedure Write(FileName: in String; P: in Path; L: in Positive);
  procedure Write(FileName: in String; P: in Path);
  procedure Read(FileName: in String; P: in out Path);
  function Read(FileName: in String) return Path;
  function First_Status0(P: Path) return Natural;
  function Min_Status(P: Path) return Integer;
  
private
  
  procedure Show_Segment(N: in String; NC,NR,Status: in Natural);
  
  Inv_R_Fac: constant Flt := One/R_Fac;
  
end Paths;
