with Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts, Strings, Messages, Param, Paths, Reps;
use Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts, Strings, Messages, Param, Paths;
with FHN_Homoclinic;

pragma Elaborate_All (FHN_Homoclinic);
pragma Elaborate_All (Ada.Text_IO,Logicals,Globals,Ints,Rationals,Flts,Strings,Messages,Param,Reps);

procedure Plot_Delta is -- plot Delta(z) along a given path

  Deg_Taylor1: constant Positive := 150;
  Deg_Taylor2: constant Positive := 150;
  SPDeg1:      constant Positive := 9;
  
  package R_Home renames Reps;
  use R_Home;
  package FHN is new FHN_Homoclinic (Deg_Taylor1  => Deg_Taylor1,
                                     Deg_Taylor2  => Deg_Taylor2,
                                     SPDeg1 => SPDeg1,
                                     Real => Reps.Rep);
  use FHN;
  use Gen_Scalars;
  use C_Home;
  
  procedure Plot_VLine(X,Y0,Y1: in Flt; Steps: in Positive) is
    R: FltVec(1..4);
  begin
    for K in 0 .. Steps loop
      R(1) := X;
      R(2) := (Flt(Steps-K)*Y0+Flt(K)*Y1)/Flt(Steps);
      C_Delta(R(1),R(2),R(3),R(4));
      Plot(Output_Dir & "delta_line",R);
      R(2) := -R(2);
      R(4) := -R(4);
      Plot(Output_Dir & "delta_rline",R);
    end loop;
  end Plot_VLine;
  
  procedure Plot_HLine(X0,X1,Y: in Flt; Steps: in Positive) is
    R: FltVec(1..4);
  begin
    for K in 0 .. Steps loop
      R(1) := (Flt(Steps-K)*X0+Flt(K)*X1)/Flt(Steps);
      R(2) := Y;
      C_Delta(R(1),R(2),R(3),R(4));
      Plot(Output_Dir & "delta_line",R);
      R(2) := -R(2);
      R(4) := -R(4);
      Plot(Output_Dir & "delta_rline",R);
    end loop;
  end Plot_HLine;
  
  procedure Plot_HalfCircle(Rad: in Radius; N: in Positive) is
    R: FltVec(1..4);
    X,Y: R_Scalar;
    Z,W: C_Scalar;
    D: CT1_Scalar;
  begin
    SetZero(Z);
    CT1_Delta(Z,Rad,D);
    for K in 0 .. N loop
      Val_HalfCircle(Rad,N,K,D,Z,W);
      Real_Part(Z,X);
      R(1) := Approx(X);
      Imag_Part(Z,Y);
      R(2) := Approx(Y);
      Real_Part(W,X);
      R(3) := Approx(X);
      Imag_Part(W,Y);
      R(4) := Approx(Y);
      Plot(Output_Dir & "delta_arc",R);
      R(2) := -R(2);
      R(4) := -R(4);
      Plot(Output_Dir & "delta_rarc",R);
    end loop;
  end Plot_HalfCircle;
  
  Steps: constant Positive := 64;
  Y0: constant Flt := 1.0/32.0;  
  
begin
  Verbosity := 5;
  Set_Precisions(256);
  Set_Taylor2_Tasks(5);
  Show_Param;
  
--  Plot_VLine(RX_Min,Y0,RY_Max,Steps);
  
--  Plot_HLine(RX_Min,RX_Max,RY_Max,Steps);
  
--  Plot_VLine(RX_Max,RY_Max,RY_Min,Steps);
  
  Plot_HalfCircle(Y0,32);
  
  
end Plot_Delta;
