with Strings;
use Strings;

pragma Elaborate_All (Strings);
  
package body Polynoms2 is
  
  procedure SetZero(P: in out Triangular; Deg: in Positive) is
  begin
    for I in 0 .. Deg loop
      for J in 0 .. Deg-I loop SetZero(P(I,J)); end loop;
    end loop;
  end SetZero;
  
  procedure ErrMult(R: in Radius; P: in out Polynom2) is
  begin
    for I in Power loop
      for J in 0 .. PDeg-I loop ErrMult(R,P(I,J)); end loop;
    end loop;
  end ErrMult;
  
  function QuasiDeg(P: Polynom2) return Natural is
  begin
    for D in reverse 1 .. PDeg loop
      for I in 0 .. D loop
        if not IsZero(P(I,D-I)) then return D; end if;
      end loop;
    end loop;
    return 0;
  end QuasiDeg;
  
  function QuasiDeg0(P: Polynom2) return Boolean is
  begin
    for D in 1 .. PDeg loop
      for I in 0 .. D loop
        if not IsZero(P(I,D-I)) then return False; end if;
      end loop;
    end loop;
    return True;
  end QuasiDeg0;
  
  procedure Intersection(P1: in Polynom2; P2: in out Polynom2; Empty: out Logical) is
    E: Logical;
  begin
    Empty := False;
    for I in Power loop
      for J in 0 .. PDeg-I loop Intersection(P1(I,J),P2(I,J),E); end loop;
      Empty := Empty or E;
    end loop;
  end Intersection;
  
  procedure Mult(R: in Flt; P: in out Polynom2; Tmp: in out Scalar) is
  begin
    if R=Zero then
      SetZero(P);
    elsif R=One then
      null;
    elsif R=NegOne then
      Neg(P);
    else
      for I in Power loop
        for J in 0 .. PDeg-I loop Mult(R,P(I,J),Tmp); end loop;
      end loop;
    end if;
  end Mult;
  
  procedure Prod(R: in Flt; P1: in Polynom2; P2: in out Polynom2) is
  begin
    if R=Zero then
      SetZero(P2);
    elsif R=One then
      Copy(P1,P2);
    elsif R=NegOne then
      Neg(P1,P2);
    else
      for I in Power loop
        for J in 0 .. PDeg-I loop
          Prod(R,P1(I,J),P2(I,J));
        end loop;
      end loop;
    end if;
  end Prod;
  
  procedure AddProd(R: in Flt; P1: in Polynom2; P2: in out Polynom2; Tmp: in out Scalar) is
  begin
    if R=Zero then
      null;
    elsif R=One then
      Add(P1,P2);
    elsif R=NegOne then
      Sub(P1,P2);
    else
      for I in Power loop
        for J in 0 .. PDeg-I loop
          AddProd(R,P1(I,J),P2(I,J),Tmp);
        end loop;
      end loop;
    end if;
  end AddProd;
  
  procedure Div(R: in Flt; P: in out Polynom2; Tmp: in out Scalar) is
  begin
    if R=One then
      null;
    elsif R=NegOne then
      Neg(P);
    else
      for I in Power loop
        for J in 0 .. PDeg-I loop Div(R,P(I,J),Tmp); end loop;
      end loop;
    end if;
  end Div;
  
  procedure Quot(P1: in Polynom2; R: in Flt; P2: in out Polynom2) is
  begin
    if R=One then
      Copy(P1,P2);
    elsif R=NegOne then
      Neg(P1,P2);
    else
      for I in Power loop
        for J in 0 .. PDeg-I loop Quot(P1(I,J),R,P2(I,J)); end loop;
      end loop;
    end if;
  end Quot;
  
  procedure Mult(S: in Scalar; P: in out Polynom2) is
  begin
    if IsZero(S) then
      SetZero(P);
    else
      for I in Power loop
        for J in 0 .. PDeg-I loop Mult(S,P(I,J)); end loop;
      end loop;
    end if;
  end Mult;
  
  procedure Prod(S: in Scalar; P1: in Polynom2; P2: in out Polynom2) is
  begin
    if IsZero(S) then
      SetZero(P2);
    else
      for I in Power loop
        for J in 0 .. PDeg-I loop Prod(S,P1(I,J),P2(I,J)); end loop;
      end loop;
    end if;
  end Prod;

  procedure AddProd(S: in Scalar; P1: in Polynom2; P2: in out Polynom2) is
  begin
    if not IsZero(S) then
      declare
        Tmp: Scalar;
      begin
        for I in Power loop
          for J in 0 .. PDeg-I loop AddProd(S,P1(I,J),P2(I,J),Tmp); end loop;
        end loop;
      end;
    end if;
  end AddProd;
  
  procedure Evaluate(P: in Polynom2; S1,S2: in Scalar; S3: in out Scalar; Deg: in Natural) is
  begin
    if Deg=0 then
      Copy(P(0,0),S3);
    else
      declare
        S4: Scalar;
      begin
        SetZero(S3);
        for I in reverse 0 .. Deg loop
          Copy(P(I,Deg-I),S4);
          for J in reverse 0 .. Deg-I-1 loop
            Mult(S2,S4);
            Add(P(I,J),S4);
          end loop;
          Mult(S1,S3);
          Add(S4,S3);
        end loop;
      end;
    end if;
  end Evaluate;
  
  procedure Evaluate(P: in Polynom2; S1,S2: in Scalar; S3: in out Scalar) is
  begin
    Evaluate(P,S1,S2,S3,QuasiDeg(P));
  end Evaluate;
  
  procedure Evaluate(P: in Polynom2; R1,R2: in Flt; S3: in out Scalar; Deg: in Natural) is
  begin
    if Deg=0 then
      Copy(P(0,0),S3);
    else
      declare
        S4: Scalar;
      begin
        SetZero(S3);
        for I in reverse 0 .. Deg loop
          Copy(P(I,Deg-I),S4);
          for J in reverse 0 .. Deg-I-1 loop
            Mult(R2,S4);
            Add(P(I,J),S4);
          end loop;
          Mult(R1,S3);
          Add(S4,S3);
        end loop;
      end;
    end if;
  end Evaluate;
  
  procedure Evaluate(P: in Polynom2; R1,R2: in Flt; S3: in out Scalar) is
  begin
    Evaluate(P,R1,R2,S3,QuasiDeg(P));
  end Evaluate;
  
  procedure Show1(N: in String; P: in Polynom2; Hide0: in Boolean := True) is
    W: constant Positive := 4;
    AllZero: Boolean := True;
    J: Integer;
  begin
    for D in 0 .. PDeg loop
      for I in 0 .. D loop
        J := D-I;
        if not IsZero(P(I,J)) then
          Show1(N & Strng(I,W) & Strng(J,W) & " ",P(I,J));
          AllZero := False;
        end if;
      end loop;
    end loop;
    if not Hide0 then
      if AllZero then Show0(N & "0"); end if;
      Show0;
    end if;
  end Show1;
    
  procedure Show2(N: in String; P1,P2: in Polynom2; Hide0: in Boolean := True) is
    W: constant Positive := 4;
    AllZero: Boolean := True;
    J: Integer;
  begin
    for D in 0 .. PDeg loop
      for I in 0 .. D loop
        J := D-I;
        if not (IsZero(P1(I,J)) and then IsZero(P2(I,J))) then
          Show2(N & Strng(I,W) & Strng(J,W) & " ",P1(I,J),P2(I,J));
          AllZero := False;
        end if;
      end loop;
    end loop;
    if not Hide0 then
      if AllZero then Show0(N & "0 0"); end if;
      Show0;
    end if;
  end Show2;
  
  procedure Num_Der1(P: in out Polynom2) is
  begin
    SetZero(P(0,PDeg));
    for I in 1 .. PDeg loop
      for J in 0 .. PDeg-I loop
        Prod(Flt(I),P(I,J),P(I-1,J));
      end loop;
      SetZero(P(I,PDeg-I));
    end loop;
  end Num_Der1;
  
  procedure Num_ValDer1(P: in Polynom2; S1,S2: in Scalar; D1: in out Scalar) is
    L: constant Integer := QuasiDeg(P);
  begin
    SetZero(D1);
    if L=0 then return; end if;
    declare
      S4,Tmp: Scalar;
    begin
      for I in reverse 1 .. L loop
        Prod(Flt(I),P(I,L-I),S4);
        for J in reverse 0 .. L-I-1 loop
          Mult(S2,S4);
          AddProd(Flt(I),P(I,J),S4,Tmp);
        end loop;
        Mult(S1,D1);
        Add(S4,D1);
      end loop;
    end;
  end Num_ValDer1;

  procedure Num_Der2(P: in out Polynom2) is
  begin
    SetZero(P(PDeg,0));
    for J in 1 .. PDeg loop
      for I in 0 .. PDeg-J loop
        Prod(Flt(J),P(I,J),P(I,J-1));
      end loop;
      SetZero(P(PDeg-J,J));
    end loop;
  end Num_Der2;
  
  procedure Num_ValDer2(P: in Polynom2; S1,S2: in Scalar; D2: in out Scalar) is
    L: constant Integer := QuasiDeg(P);
  begin
    SetZero(D2);
    if L=0 then return; end if;
    declare
      S4,Tmp: Scalar;
    begin
      for J in reverse 1 .. L loop
        Prod(Flt(J),P(L-J,J),S4);
        for I in reverse 0 .. L-J-1 loop
          Mult(S2,S4);
          AddProd(Flt(J),P(I,J),S4,Tmp);
        end loop;
        Mult(S1,D2);
        Add(S4,D2);
      end loop;
    end;
  end Num_ValDer2;

  procedure Num_MDer(P: in out Polynom2) is
    Tmp: Scalar renames P(0,0);
  begin
    for D in reverse 2 .. PDeg loop
      for I in 0 .. D loop
        Mult(Flt(D),P(I,D-I),Tmp);
      end loop;
    end loop;
    SetZero(P(0,0));
  end Num_MDer;
  
  procedure Inv_MDer(P: in out Polynom2; CheckConst: in Boolean := True) is
    Tmp: Scalar renames P(0,0);
  begin
    if CheckConst and then not IsZero(P(0,0)) then
      raise Undefined with "Matrices.Poly.Inv_MDer: nonzero constant";
    end if;
    for D in reverse 2 .. PDeg loop
      for I in 0 .. D loop
        Div(Flt(D),P(I,D-I),Tmp);
      end loop;
    end loop;
    SetZero(P(0,0));
  end Inv_MDer;
  
  procedure Get_High(C: in Polynom2; E: in out EVector) is
  begin
    for I in Power loop Copy(C(I,PDeg-I),E(I)); end loop;
  end Get_High;
  
  procedure Put_High(E: in EVector; C: in out Polynom2) is
  begin
    for I in Power loop Copy(E(I),C(I,PDeg-I)); end loop;
  end Put_High;
  
  procedure Add_High(E: in EVector; C: in out Polynom2) is
  begin
    for I in Power loop Add(E(I),C(I,PDeg-I)); end loop;
  end Add_High;
  
---------------------------------------------------------------------
---------------------------------------------------------------------
---------------------------------------------------------------------
  
  procedure AddProd_Deg(C1,C2: in Polynom2; C3: in out Polynom2; Deg: in Natural) is
--- just compte terms of degree Deg
    NT: constant Positive := IMin(Prod_Tasks,Deg+1);
    
    procedure Partial_Prod(I3: in out Integer) is
      Tmp: Scalar;
    begin
      loop
        declare
          J3: constant Integer := Deg-I3;
          S3: Scalar renames C3(I3,J3);
        begin
          for I1 in 0 .. I3 loop
            for J1 in 0 .. J3 loop
              AddProd(C1(I1,J1),C2(I3-I1,J3-J1),S3,Tmp);
            end loop;
          end loop;
        end;
        I3 := I3+NT;
        exit when I3>Deg;
      end loop;
    end Partial_Prod;
    
    task type Partial_Prod_Type is
      entry Start(IFirst: in Natural);
    end Partial_Prod_Type;
    
    task body Partial_Prod_Type is
      I3: Integer;
    begin
      accept Start(IFirst: in Natural) do I3 := IFirst; end Start;
      Partial_Prod(I3);
    end Partial_Prod_Type;
    
    I3: Integer := 0;
    
  begin
    if NT=1 then
      Partial_Prod(I3);
    else
      declare
        Run_Partial_Prod: array (1 .. NT-1) of Partial_Prod_Type;
      begin
        for N in 1 .. NT-1 loop
          Run_Partial_Prod(N).Start(N);
        end loop;
        Partial_Prod(I3);
      end;
    end if;
  end AddProd_Deg;
  
  procedure AddProd_Low(C1,C2: in Polynom2; D1,D2: in Natural; C3: in out Polynom2) is
    D3: constant Natural := IMin(D1+D2,PDeg);
    NT: constant Positive := IMin(Prod_Tasks,D3+1);
    
    procedure Partial_Prod(I3: in out Integer) is
      Tmp: Scalar;
    begin
      loop
        for J3 in 0 .. D3-I3 loop
          declare
            K: constant Integer := I3+J3-D2;
            S3: Scalar renames C3(I3,J3);
          begin
            for I1 in IMax(0,I3-D2) .. IMin(I3,D1) loop
              for J1 in IMax(0,K-I1) .. IMin(J3,D1-I1) loop
                AddProd(C1(I1,J1),C2(I3-I1,J3-J1),S3,Tmp);
              end loop;
            end loop;
          end;
        end loop;
        I3 := I3+NT;
        exit when I3>D3;
      end loop;
    end Partial_Prod;
    
    task type Partial_Prod_Type is
      entry Start(IFirst: in Natural);
    end Partial_Prod_Type;
    
    task body Partial_Prod_Type is
      I3: Integer;
    begin
      accept Start(IFirst: in Natural) do I3 := IFirst; end Start;
      Partial_Prod(I3);
    end Partial_Prod_Type;
    
    I3: Integer := 0;
    
  begin
    if NT=1 then
      Partial_Prod(I3);
    else
      declare
        Run_Partial_Prod: array (1 .. NT-1) of Partial_Prod_Type;
      begin
        for N in 1 .. NT-1 loop
          Run_Partial_Prod(N).Start(N);
        end loop;
        Partial_Prod(I3);
      end;
    end if;
  end AddProd_Low;
  
  procedure Prod_Err(C1,C2: in Polynom2; D1,D2: in Natural; R: in Radius; E: in out EVector) is
    D3: constant Integer := D1+D2;
    BP: LBP_Pointer;
  begin
    Pool.Allocate(BP);
    declare
      NT: constant Positive := IMin(Prod_Tasks,D3-PDeg);
      C3: Polynom2 renames BP.Data;
    
      procedure Partial_Prod(I3: in out Integer) is
        Tmp: Scalar;
      begin
        loop
          for J3 in 0 .. D3-I3 loop
            declare
              K: constant Integer := I3+J3-D2;
              S3: Scalar renames C3(I3,J3);
            begin
              for I1 in IMax(0,I3-D2) .. IMin(I3,D1) loop
                for J1 in IMax(0,K-I1) .. IMin(J3,D1-I1) loop
                  AddProd(C1(I1,J1),C2(I3-I1,J3-J1),S3,Tmp);
                end loop;
              end loop;
            end;
          end loop;
          I3 := I3+NT;
          exit when I3>D3;
        end loop;
      end Partial_Prod;
    
      task type Partial_Prod_Type is
        entry Start(IFirst: in Natural);
      end Partial_Prod_Type;
    
      task body Partial_Prod_Type is
        I3: Integer;
      begin
        accept Start(IFirst: in Natural) do I3 := IFirst; end Start;
        Partial_Prod(I3);
      end Partial_Prod_Type;
    
      I3: Integer := PDeg1;
      J: Integer;
      Fac: Flt;
      S,Tmp: Scalar;
      
    begin
      for D in PDeg1 .. D3 loop
        for I in 0 .. D loop SetZero(C3(I,D-I)); end loop;
      end loop;
      if NT=1 then
        Partial_Prod(I3);
      else
        declare
          Run_Partial_Prod: array (2 .. NT) of Partial_Prod_Type;
        begin
          for N in 2 .. NT loop
            Run_Partial_Prod(N).Start(PDeg+N);
          end loop;
          Partial_Prod(I3);
        end;
      end if;
      SV.SetZero(E);                       --- collect errors
      for D in reverse PDeg1 .. D3 loop
        Fac := R**(D-PDeg);
        for I in 0 .. D loop
          J := D-I;
          ToErr(C3(I,J),S);
          if I<J then
            AddProd(Fac,S,E(I),Tmp);
          elsif I>J then
            AddProd(Fac,S,E(PDeg-J),Tmp);
          else
            AddProd(Half*Fac,S,E(I),Tmp);
            AddProd(Half*Fac,S,E(PDeg-J),Tmp);
          end if;
        end loop;
      end loop;
    end;
    Pool.Recycle(BP);
  end Prod_Err;
  
end Polynoms2;
