with Ada.Text_IO, Globals, Ints, Rationals, Logicals, Flts;
use Ada.Text_IO, Globals, Ints, Rationals, Logicals, Flts;
with Recycling, Matrices, Vectors, Vectors.More, Triangulars.TGeneric;

pragma Elaborate_All (Ada.Text_IO,Globals,Ints,Rationals,Logicals,Flts);
pragma Elaborate_All (Recycling,Matrices,Vectors.More,Triangulars.TGeneric);

generic
  
  PDeg: in Positive;
  
-------------------- begin_include scalars Scalar

  type Scalar is private;

--- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

--- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

--- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>; --- sign of S1-S2
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

--- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

--- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

--- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

--- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  with procedure Set_Active_Mode(Dummy: in Scalar) is <>;
  with procedure Set_Wait_Mode(Dummy: in Scalar) is <>;
  with procedure Free_Cache(S: in Scalar) is <>;
-------------------- end_include scalars  
  
  with package SM is new Matrices (Scalar => Scalar);
  with package SV is new Vectors (Scalar => Scalar, Component => Scalar, SM => SM);
  with package ST is new Triangulars (Scalar => Scalar);
  
package Polynoms2 is
  
  subtype Power is Integer range 0 .. PDeg;
  subtype Triangular is ST.Triangular;
  subtype Polynom2 is Triangular(Power,Power); -- P(I,J) undefined if I+J>PDeg
  subtype EVector is SV.Vector(Power);
  
  package STG is new ST.TGeneric (PDeg => PDeg);
  use SM,STG;
  
  function "=" is new TF_BSimilar ("=");
  function IsZero is new TF_BProperty (IsZero);
  function IsReal is new TF_BProperty (IsReal);
  function IsSharp is new TF_BProperty (IsSharp);
  procedure SetZero(P: in out Triangular; Deg: in Positive);
  procedure SetZero is new TP_Modify1 (SetZero);
  procedure Copy is new TP_Map1 (Copy);
  procedure Swap is new TP_Modify2 (Swap);
  function Center0 is new TF_BProperty (Center0);
  function Contains0 is new TF_LProperty (Contains0);
  function Contains is new TF_LSimilar (Contains);
  procedure ToErr is new TP_Modify1 (ToErr);
  procedure ToErr is new TP_Map1 (ToErr);
  procedure Center is new TP_Modify1 (Center);
  procedure Center is new TP_Map1 (Center);
  procedure ModCenter is new TP_Modify1 (ModCenter);
  procedure ModCenter is new TP_Map1 (ModCenter);
  procedure ErrMult(R: in Radius; P: in out Polynom2);
  procedure Union is new TP_Map1 (Union);
  function QuasiDeg(P: Polynom2) return Natural;
  function QuasiDeg0(P: Polynom2) return Boolean;
  procedure Intersection(P1: in Polynom2; P2: in out Polynom2; Empty: out Logical);
  
  procedure Neg is new TP_Modify1 (Neg);
  procedure Neg is new TP_Map1 (Neg);
  procedure Add is new TP_Map1 (Add);
  procedure Sum is new TP_Map2 (Sum);
  procedure Sub is new TP_Map1 (Sub);
  procedure Diff is new TP_Map2 (Diff);
  procedure Mult(R: in Flt; P: in out Polynom2; Tmp: in out Scalar);
  procedure Prod(R: in Flt; P1: in Polynom2; P2: in out Polynom2);
  procedure Div(R: in Flt; P: in out Polynom2; Tmp: in out Scalar);
  procedure Quot(P1: in Polynom2; R: in Flt; P2: in out Polynom2);
  procedure Adjoint is new TP_Modify1 (Adjoint);
  procedure Adjoint is new TP_Map1 (Adjoint);
  procedure Real_Part is new TP_Modify1 (Real_Part);
  procedure Imag_Part is new TP_Modify1 (Imag_Part);
  procedure Mult(S: in Scalar; P: in out Polynom2);
  procedure Prod(S: in Scalar; P1: in Polynom2; P2: in out Polynom2);
  procedure AddProd(S: in Scalar; P1: in Polynom2; P2: in out Polynom2);
  procedure AddProd(R: in Flt; P1: in Polynom2; P2: in out Polynom2; Tmp: in out Scalar);
  procedure Evaluate(P: in Polynom2; S1,S2: in Scalar; S3: in out Scalar; Deg: in Natural);
  procedure Evaluate(P: in Polynom2; S1,S2: in Scalar; S3: in out Scalar);
  procedure Evaluate(P: in Polynom2; R1,R2: in Flt; S3: in out Scalar; Deg: in Natural);
  procedure Evaluate(P: in Polynom2; R1,R2: in Flt; S3: in out Scalar);
  procedure Show1(N: in String; P: in Polynom2; Hide0: in Boolean := True);
  procedure Show2(N: in String; P1,P2: in Polynom2; Hide0: in Boolean := True);
  
  pragma Inline (SetZero,Copy,Neg);
  
  procedure Num_Der1(P: in out Polynom2);
  procedure Num_ValDer1(P: in Polynom2; S1,S2: in Scalar; D1: in out Scalar);
  procedure Num_Der2(P: in out Polynom2);
  procedure Num_ValDer2(P: in Polynom2; S1,S2: in Scalar; D2: in out Scalar);
  procedure Num_MDer(P: in out Polynom2);                                     -- generator of dilation
  procedure Inv_MDer(P: in out Polynom2; CheckConst: in Boolean := True);     -- inverse of MDer
  
  package SVM is new SV.More;
  
  procedure ErrMult(R: in Radius; V: in out SV.Vector) renames SVM.ErrMult;
  function Contains(V1,V2: SV.Vector) return Logical renames SVM.Contains;
  procedure Get_High(C: in Polynom2; E: in out EVector);
  procedure Put_High(E: in EVector; C: in out Polynom2);
  procedure Add_High(E: in EVector; C: in out Polynom2);
  
--- stuff for Taylors2
  
  procedure AddProd_Deg(C1,C2: in Polynom2; C3: in out Polynom2; Deg: in Natural);
  procedure AddProd_Low(C1,C2: in Polynom2; D1,D2: in Natural; C3: in out Polynom2);
  procedure Prod_Err(C1,C2: in Polynom2; D1,D2: in Natural; R: in Radius; E: in out EVector);
  
  A_Prod_Err: access procedure(C1,C2: in Polynom2; D1,D2: in Natural; R: in Radius; E: in out EVector) := Prod_Err'Access;
  
  Prod_Tasks: Positive := 1;          -- number of tasks for Prod and AddProd
  
private
  
  PDeg1: constant Integer := 1+PDeg;
  PDeg2: constant Integer := 2*PDeg;
  
  subtype Power2 is Integer range 0 .. PDeg2;
  subtype BPolynom2 is Triangular(Power2,Power2);
  
  package Alloc is new Recycling (Item => BPolynom2);
  subtype LBP_Pointer is Alloc.LD_Pointer;
  Pool: Alloc.Protected_Data;
  
end Polynoms2;

