
package body Scalars is
  
------------------------------------------------------------------  
--- conversion real <--> complex ---------------------------------
------------------------------------------------------------------  

-- Basic Scalar
  
  procedure Set_Real(R: in Real_Number; C: in out Complex_Number) is
  begin
    Copy(R,C.Re);
    SetZero(C.Im);
  end Set_Real;
  
  function Set_Real(R: Real_Number) return Complex_Number is
    C: Complex_Number;
  begin
    Copy(R,C.Re);
    SetZero(C.Im);
    return C;
  end Set_Real;
  
  procedure Set_Imag(R: in Real_Number; C: in out Complex_Number) is
  begin
    Copy(R,C.Im);
    SetZero(C.Re);
  end Set_Imag;
  
  procedure Set_Complex(Rx,Ry: in Real_Number; C: in out Complex_Number) is
  begin
    Copy(Rx,C.Re);
    Copy(Ry,C.Im);
  end Set_Complex;
  
  function Set_Complex(Rx,Ry: Real_Number) return Complex_Number is
    C: Complex_Number;
  begin
    Copy(Rx,C.Re);
    Copy(Ry,C.Im);
    return C;
  end Set_Complex;
  
  procedure Real_Part(C: in Complex_Number; R: in out Real_Number) is
  begin
    Copy(C.Re,R);
  end Real_Part;
  
  procedure Imag_Part(C: in Complex_Number; R: in out Real_Number) is
  begin
    Copy(C.Im,R);
  end Imag_Part;
  
  procedure Set_Real(R: in Real_Matrix; C: in out Complex_Matrix) is
  begin
    for I in R'Range(1) loop
      for J in R'Range(2) loop Set_Real(R(I,J),C(I,J)); end loop;
    end loop;
  end Set_Real;
  
  procedure Set_Imag(R: in Real_Matrix; C: in out Complex_Matrix) is
  begin
    for I in R'Range(1) loop
      for J in R'Range(2) loop Set_Imag(R(I,J),C(I,J)); end loop;
    end loop;
  end Set_Imag;
  
  procedure Set_Complex(Rx,Ry: in Real_Matrix; C: in out Complex_Matrix) is
  begin
     for I in C'Range(1) loop
	for J in C'Range(2) loop Set_Complex(Rx(I,J),Ry(I,J),C(I,J)); end loop;
     end loop;
  end Set_Complex;
  
  procedure Real_Part(C: in Complex_Matrix; R: in out Real_Matrix) is
  begin
    for I in C'Range(1) loop
      for J in C'Range(2) loop Copy(C(I,J).Re,R(I,J)); end loop;
    end loop;
  end Real_Part;
  
  procedure Imag_Part(C: in Complex_Matrix; R: in out Real_Matrix) is
  begin
    for I in C'Range(1) loop
      for J in C'Range(2) loop Copy(C(I,J).Im,R(I,J)); end loop;
    end loop;
  end Imag_Part;
  
  procedure Set_Real(R: in Real_Triangular; C: in out Complex_Triangular) is
    D: constant Integer := R'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Set_Real(R(I,J),C(I,J)); end loop;
    end loop;
  end Set_Real;
  
  procedure Set_Imag(R: in Real_Triangular; C: in out Complex_Triangular) is
    D: constant Integer := R'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Set_Imag(R(I,J),C(I,J)); end loop;
    end loop;
  end Set_Imag;
  
  procedure Real_Part(C: in Complex_Triangular; R: in out Real_Triangular) is
    D: constant Integer := C'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Copy(C(I,J).Re,R(I,J)); end loop;
    end loop;
  end Real_Part;
  
  procedure Imag_Part(C: in Complex_Triangular; R: in out Real_Triangular) is
    D: constant Integer := C'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Copy(C(I,J).Im,R(I,J)); end loop;
    end loop;
  end Imag_Part;
  
  procedure Set_Real(R: in Real_Vector; C: in out Complex_Vector) is
  begin
    for I in R'Range loop Set_Real(R(I),C(I)); end loop;
  end Set_Real;
  
  procedure Set_Imag(R: in Real_Vector; C: in out Complex_Vector) is
  begin
    for I in R'Range loop Set_Imag(R(I),C(I)); end loop;
  end Set_Imag;
  
  procedure Set_Complex(Rx,Ry: in Real_Vector; C: in out Complex_Vector) is
  begin
    for I in C'Range loop Set_Complex(Rx(I),Ry(I),C(I)); end loop;     
  end Set_Complex;
  
  procedure Real_Part(C: in Complex_Vector; R: in out Real_Vector) is
  begin
    for I in C'Range loop Copy(C(I).Re,R(I)); end loop;
  end Real_Part;
  
  procedure Imag_Part(C: in Complex_Vector; R: in out Real_Vector) is
  begin
    for I in C'Range loop Copy(C(I).Im,R(I)); end loop;
  end Imag_Part;
  
  procedure Set_Real(R: in Real_Diag; C: in out Complex_Diag) is
  begin
    for I in R'Range loop Set_Real(R(I),C(I)); end loop;
  end Set_Real;
  
  procedure Set_Imag(R: in Real_Diag; C: in out Complex_Diag) is
  begin
    for I in R'Range loop Set_Imag(R(I),C(I)); end loop;
  end Set_Imag;
  
-- Taylor1
  
  procedure Set_Real(R: in Real_Taylor1; C: in out Complex_Taylor1) is
  begin
    Set_Real(R.C,C.C);
    C.F := R.F;
    C.R := R.R;
  end Set_Real;
  
  function Set_Real(R: Real_Taylor1) return Complex_Taylor1 is
    C: Complex_Taylor1;
  begin
    Set_Real(R,C);
    return C;
  end Set_Real;
  
  procedure Set_Imag(R: in Real_Taylor1; C: in out Complex_Taylor1) is
  begin
    Set_Imag(R.C,C.C);
    C.F := R.F;
    C.R := R.R;
  end Set_Imag;
  
  procedure Set_Complex(Rx,Ry: in Real_Taylor1; C: in out Complex_Taylor1) is
  begin
    Set_Real(Rx.C,C.C);
    Set_Imag(Ry.C,C.C);
    C.F := IMin(Rx.F,Ry.F);
    C.R := RMin(Rx.R,Ry.R);
  end Set_Complex;
  
  function Set_Complex(Rx,Ry: Real_Taylor1) return Complex_Taylor1 is
    C: Complex_Taylor1;
  begin
    Set_Complex(Rx,Ry,C);
    return C;
  end Set_Complex;
  
  procedure Real_Part(C: in Complex_Taylor1; R: in out Real_Taylor1) is
  begin
    Real_Part(C.C,R.C);
    R.F := C.F;
    R.R := C.R;
  end Real_Part;
  
  procedure Imag_Part(C: in Complex_Taylor1; R: in out Real_Taylor1) is
  begin
    Imag_Part(C.C,R.C);
    R.F := C.F;
    R.R := C.R;
  end Imag_Part;
  
  procedure Set_Real(R: in Real_Taylor1_Matrix; C: in out Complex_Taylor1_Matrix) is
  begin
    for I in R'Range(1) loop
      for J in R'Range(2) loop Set_Real(R(I,J),C(I,J)); end loop;
    end loop;
  end Set_Real;
  
  procedure Set_Imag(R: in Real_Taylor1_Matrix; C: in out Complex_Taylor1_Matrix) is
  begin
    for I in R'Range(1) loop
      for J in R'Range(2) loop Set_Imag(R(I,J),C(I,J)); end loop;
    end loop;
  end Set_Imag;
  
  procedure Real_Part(C: in Complex_Taylor1_Matrix; R: in out Real_Taylor1_Matrix) is
  begin
    for I in C'Range(1) loop
      for J in C'Range(2) loop Real_Part(C(I,J),R(I,J)); end loop;
    end loop;
  end Real_Part;
  
  procedure Imag_Part(C: in Complex_Taylor1_Matrix; R: in out Real_Taylor1_Matrix) is
  begin
    for I in C'Range(1) loop
      for J in C'Range(2) loop Imag_Part(C(I,J),R(I,J)); end loop;
    end loop;
  end Imag_Part;
  
  procedure Set_Real(R: in Real_Taylor1_Triangular; C: in out Complex_Taylor1_Triangular) is
    D: constant Integer := R'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Set_Real(R(I,J),C(I,J)); end loop;
    end loop;
  end Set_Real;
  
  procedure Set_Imag(R: in Real_Taylor1_Triangular; C: in out Complex_Taylor1_Triangular) is
    D: constant Integer := R'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Set_Imag(R(I,J),C(I,J)); end loop;
    end loop;
  end Set_Imag;
  
  procedure Real_Part(C: in Complex_Taylor1_Triangular; R: in out Real_Taylor1_Triangular) is
    D: constant Integer := C'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Real_Part(C(I,J),R(I,J)); end loop;
    end loop;
  end Real_Part;
  
  procedure Imag_Part(C: in Complex_Taylor1_Triangular; R: in out Real_Taylor1_Triangular) is
    D: constant Integer := C'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Imag_Part(C(I,J),R(I,J)); end loop;
    end loop;
  end Imag_Part;
  
  procedure Set_Real(R: in Real_Taylor1_Vector; C: in out Complex_Taylor1_Vector) is
  begin
    for I in R'Range loop Set_Real(R(I),C(I)); end loop;
  end Set_Real;
  
  procedure Set_Imag(R: in Real_Taylor1_Vector; C: in out Complex_Taylor1_Vector) is
  begin
    for I in R'Range loop Set_Imag(R(I),C(I)); end loop;
  end Set_Imag;
  
  procedure Real_Part(C: in Complex_Taylor1_Vector; R: in out Real_Taylor1_Vector) is
  begin
    for I in C'Range loop Real_Part(C(I),R(I)); end loop;
  end Real_Part;
  
  procedure Imag_Part(C: in Complex_Taylor1_Vector; R: in out Real_Taylor1_Vector) is
  begin
    for I in C'Range loop Imag_Part(C(I),R(I)); end loop;
  end Imag_Part;
  
------------------------------------------------------------------  
--- assign, evaluate etc -----------------------------------------
------------------------------------------------------------------  
  
  function Rho(V: Real_Taylor1_Vector) return Radius is
    use Real_Taylors1;
    R: Radius := Rho(V(V'First));
    RI: Radius;
  begin
    for I in V'First+1 .. V'Last loop
      RI := Rho(V(I));
      if RI /= R then
        if Verbosity>0 then
          Show2("Scalars.Rho warning: different radii in Real_Taylor1_Vector",RI,R);
        end if;
        R := RMin(RI,R);
      end if;
    end loop;
    return R;
  end Rho;
  
  function Rho(V: Complex_Taylor1_Vector) return Radius is
    use Complex_Taylors1;
    R: Radius := Rho(V(V'First));
    RI: Radius;
  begin
    for I in V'First+1 .. V'Last loop
      RI := Rho(V(I));
      if RI /= R then
        if Verbosity>0 then
          Show2("Scalars.Rho warning: different radii in Complex_Taylor1_Vector",RI,R);
        end if;
        R := RMin(RI,R);
      end if;
    end loop;
    return R;
  end Rho;
  
  function Rho(V: Real_Taylor1_Triangular) return Radius is
    use Real_Taylors1;
    R: Radius := Rho(V(0,0));
    RI: Radius;
  begin
    for D in 1 .. V'Last(1) loop
      for I in 0 .. D loop
        RI := Rho(V(I,D-I));
        if RI /= R then
          if Verbosity>0 then
            Show2("Scalars.Rho warning: different radii in Real_Taylor1_Triangular",RI,R);
          end if;
          R := RMin(RI,R);
        end if;
      end loop;
    end loop;
    return R;
  end Rho;
  
  function Rho(V: Complex_Taylor1_Triangular) return Radius is
    use Complex_Taylors1;
    R: Radius := Rho(V(0,0));
    RI: Radius;
  begin
    for D in 1 .. V'Last(1) loop
      for I in 0 .. D loop
        RI := Rho(V(I,D-I));
        if RI /= R then
          if Verbosity>0 then
            Show2("Scalars.Rho warning: different radii in Complex_Taylor1_Triangular",RI,R);
          end if;
          R := RMin(RI,R);
        end if;
      end loop;
    end loop;
    return R;
  end Rho;
  
  procedure SetRho(R: in Radius; V: in out Real_Taylor1_Vector; Check: in Boolean := True) is
    use Real_Taylors1;
  begin
    for I in V'Range loop SetRho(R,V(I),Check); end loop;
  end SetRho;
  
  procedure SetRho(R: in Radius; V: in out Complex_Taylor1_Vector; Check: in Boolean := True) is
    use Complex_Taylors1;
  begin
    for I in V'Range loop SetRho(R,V(I),Check); end loop;
  end SetRho;
  
  procedure SetRho(R: in Radius; V: in out Real_Taylor1_Triangular; Check: in Boolean := True) is
    use Real_Taylors1;
    D: constant Integer := V'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop SetRho(R,V(I,J),Check); end loop;
    end loop;
  end SetRho;
  
  procedure SetRho(R: in Radius; V: in out Complex_Taylor1_Triangular; Check: in Boolean := True) is
    use Complex_Taylors1;
    D: constant Integer := V'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop SetRho(R,V(I,J),Check); end loop;
    end loop;
  end SetRho;
  
  procedure Assign(V: in Real_Vector; W: in out Real_Taylor1_Vector) is
    use Real_Taylors1;
  begin
    for I in W'Range loop Assign(V(I),W(I)); end loop;
  end Assign;
  
  procedure Assign(V: in Real_Vector; W: in out Complex_Taylor1_Vector) is
    S: Complex_Number;
  begin
    for I in W'Range loop
      Set_Real(V(I),S);
      Assign(S,W(I));
    end loop;
  end Assign;
  
  procedure Assign(V: in Complex_Vector; W: in out Complex_Taylor1_Vector) is
    use Complex_Taylors1;
  begin
    for I in W'Range loop Assign(V(I),W(I)); end loop;
  end Assign;
  
  procedure Assign(V: in Real_Triangular; W: in out Real_Taylor1_Triangular) is
    use Real_Taylors1;
    D: constant Integer := W'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Assign(V(I,J),W(I,J)); end loop;
    end loop;
  end Assign;
  
  procedure Assign(V: in Real_Triangular; W: in out Complex_Taylor1_Triangular) is
    D: constant Integer := W'Last(1);
    S: Complex_Number;
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop
        Set_Real(V(I,J),S);
        Assign(S,W(I,J));
      end loop;
    end loop;
  end Assign;
  
  procedure Assign(V: in Complex_Triangular; W: in out Complex_Taylor1_Triangular) is
    use Complex_Taylors1;
    D: constant Integer := W'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Assign(V(I,J),W(I,J)); end loop;
    end loop;
  end Assign;
  
  procedure Val(V: in Real_Taylor1_Vector; R: in Real_Number; W: in out Real_Vector) is
    use Real_Taylors1;
  begin
    for I in W'Range loop Val(V(I),R,W(I)); end loop;
  end Val;
  
  procedure Val(V: in Complex_Taylor1_Vector; R: in Complex_Number; W: in out Complex_Vector) is
    use Complex_Taylors1;
  begin
    for I in W'Range loop Val(V(I),R,W(I)); end loop;
  end Val;
  
  procedure Val(V: in Real_Taylor1_Triangular; R: in Real_Number; W: in out Real_Triangular) is
    use Real_Taylors1;
    D: constant Integer := W'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Val(V(I,J),R,W(I,J)); end loop;
    end loop;
  end Val;
  
  procedure Val(V: in Complex_Taylor1_Triangular; R: in Complex_Number; W: in out Complex_Triangular) is
    use Complex_Taylors1;
    D: constant Integer := W'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop Val(V(I,J),R,W(I,J)); end loop;
    end loop;
  end Val;
  
  procedure ValDer(V: in Real_Taylor1_Vector; R: in Real_Number; W: in out Real_Vector) is
    use Real_Taylors1;
  begin
    for I in W'Range loop ValDer(V(I),R,W(I)); end loop;
  end ValDer;
  
  procedure ValDer(V: in Complex_Taylor1_Vector; R: in Complex_Number; W: in out Complex_Vector) is
    use Complex_Taylors1;
  begin
    for I in W'Range loop ValDer(V(I),R,W(I)); end loop;
  end ValDer;
  
  procedure ValDer(V: in Real_Taylor1_Triangular; R: in Real_Number; W: in out Real_Triangular) is
    use Real_Taylors1;
    D: constant Integer := W'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop ValDer(V(I,J),R,W(I,J)); end loop;
    end loop;
  end ValDer;
  
  procedure ValDer(V: in Complex_Taylor1_Triangular; R: in Complex_Number; W: in out Complex_Triangular) is
    use Complex_Taylors1;
    D: constant Integer := W'Last(1);
  begin
    for I in 0 .. D loop
      for J in 0 .. D-I loop ValDer(V(I,J),R,W(I,J)); end loop;
    end loop;
  end ValDer;
  
  
end Scalars;
