with Ada.Text_IO, Logicals, Globals, Param, Ints, Rationals, Flts;
use Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts;
with Matrices, Vectors, ScalVectors, Complexs, Taylors1, Triangulars;

pragma Elaborate_All (Ada.Text_IO,Logicals,Globals,Param,Ints,Rationals,Flts);
pragma Elaborate_All (Vectors,ScalVectors,Complexs,Taylors1,Triangulars);

generic
  
  SPDeg1: in Positive;
  
-------------------- begin_include scalars Real  

  type Real is private;

--- basic
  with function Info(Dummy: Real) return Scalar_Info is <>;
  with function IsSharp(S: Real) return Boolean is <>;
  with function IsZero(S: Real) return Boolean is <>;
  with function "="(S1,S2: Real) return Boolean is <>;
  with procedure SetZero(S: in out Real) is <>;
  with procedure Copy(S1: in Real; S2: in out Real) is <>;
  with procedure Swap(S1,S2: in out Real) is <>;

--- sets
  with function Center0(S: Real) return Boolean is <>;
  with function Contains0(S: Real) return Logical is <>;
  with function Contains(S1,S2: Real) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Real) is <>;
  with function BallAt0(R: Flt) return Real is <>;
  with procedure ToErr(S: in out Real) is <>;
  with procedure ToErr(S1: in Real; S2: in out Real) is <>;
  with function ToErr(S: Real) return Real is <>;
  with procedure Center(S: in out Real) is <>;
  with procedure Center(S1: in Real; S2: in out Real) is <>;
  with function Center(S: Real) return Real is <>;
  with procedure ModCenter(S: in out Real) is <>;
  with procedure ModCenter(S1: in Real; S2: in out Real) is <>;
  with function ModCenter(S: Real) return Real is <>;
  with procedure ErrMult(R: in Radius; S: in out Real) is <>;
  with procedure Union(S1: in Real; S2: in out Real) is <>;
  with function Union(S1,S2: Real) return Real is <>;
  with procedure Intersection(S1: in Real; S2: in out Real; Empty: out Logical) is <>;

--- order
  with function Sign(S: Real) return Integer is <>;
  with function Compare(S1,S2: Real) return Integer is <>;
  with function "<"(S1,S2: Real) return Boolean is <>;
  with function "<="(S1,S2: Real) return Boolean is <>;
  with function ">"(S1,S2: Real) return Boolean is <>;
  with function ">="(S1,S2: Real) return Boolean is <>;
  with procedure Min(S1: in Real; S2: in out Real) is <>;
  with procedure Min(S1,S2: in Real; S3: in out Real) is <>;
  with function Min(S1,S2: Real) return Real is <>;
  with procedure Max(S1: in Real; S2: in out Real) is <>;
  with procedure Max(S1,S2: in Real; S3: in out Real) is <>;
  with function Max(S1,S2: Real) return Real is <>;
  with function Inf(S: Real) return Flt is <>; -- inf of real part
  with function Sup(S: Real) return Flt is <>; -- sup of real part

--- arithmetic
  with procedure Neg(S: in out Real) is <>;
  with procedure Neg(S1: in Real; S2: in out Real) is <>;
  with function "-"(S: Real) return Real is <>;
  with procedure Add(I: in Integer; S: in out Real) is <>;
  with procedure Add(S1: in Real; S2: in out Real) is <>;
  with procedure Sum(S1,S2: in Real; S3: in out Real) is <>;
  with function "+"(S1,S2: Real) return Real is <>;
  with procedure Sub(S1: in Real; S2: in out Real) is <>;
  with procedure Diff(S1,S2: in Real; S3: in out Real) is <>;
  with function "-"(S1,S2: Real) return Real is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Real) is <>;
  with procedure Mult(R: in Flt; S: in out Real) is <>;
  with procedure Prod(R: in Flt; S1: in Real; S2: in out Real) is <>;
  with function "*"(R: Flt; S: Real) return Real is <>;
  with procedure AddProd(R: in Flt; S1: in Real; S2,Tmp: in out Real) is <>;
  with procedure AddProd(R: in Flt; S1: in Real; S2: in out Real) is <>;
  with procedure Mult(Q: in Rational; S: in out Real) is <>;
  with procedure Mult(S1: in Real; S2: in out Real) is <>;
  with procedure Prod(S1,S2: in Real; S3: in out Real) is <>;
  with function "*"(S1,S2: Real) return Real is <>;
  with procedure AddProd(S1,S2: in Real; S3,Tmp: in out Real) is <>;
  with procedure AddProd(S1,S2: in Real; S3: in out Real) is <>;
  with procedure SubProd(S1,S2: in Real; S3,Tmp: in out Real) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Real) is <>;
  with procedure Div(R: in Flt; S: in out Real) is <>;
  with procedure Quot(S1: in Real; R: in Flt; S2: in out Real) is <>;
  with function "/"(S: Real; R: Flt) return Real is <>;
  with procedure Div(S1: in Real; S2: in out Real) is <>;
  with procedure Quot(S1,S2: in Real; S3: in out Real) is <>;
  with function "/"(S1,S2: Real) return Real is <>;
  with procedure Inv(S1: in Real; S2: in out Real) is <>;
  with function Inv(S: Real) return Real is <>;
  with function "**"(S: Real; I: Integer) return Real is <>;

--- fun
  with function IsReal(S: Real) return Boolean is <>;
  with procedure Adjoint(S: in out Real) is <>;
  with procedure Adjoint(S1: in Real; S2: in out Real) is <>;
  with function Adjoint(S: Real) return Real is <>;
  with procedure Real_Part(S: in out Real) is <>;
  with procedure Imag_Part(S: in out Real) is <>;
  with procedure Eval0(S: in out Real) is <>;
  with function "abs"(S: Real) return Real is <>;
  with procedure Norm(S1: in Real; S2: in out Real) is <>;
  with function Norm(S: Real) return Real is <>;
  with function MaxNorm(S: Real) return Radius is <>;
  with procedure Sqr(S: in out Real) is <>;
  with function Sqr(S: Real) return Real is <>;
  with procedure Sqrt(S1: in Real; S2: in out Real) is <>;
  with function Sqrt(S: Real) return Real is <>;
  with procedure Root(K: in Positive; S1: in Real; S2: in out Real) is <>;
  with function Root(K: Positive; S: Real) return Real is <>;
  with procedure Exp(S1: in Real; S2: in out Real) is <>;
  with function Exp(S: Real) return Real is <>;
  with procedure ArcCos(S1: in Real; S2: in out Real) is <>;
  with function ArcCos(S: Real) return Real is <>;
  with procedure ArcSin(S1: in Real; S2: in out Real) is <>;
  with function ArcSin(S: Real) return Real is <>;
  with procedure Cos(S1: in Real; S2: in out Real) is <>;
  with function Cos(S: Real) return Real is <>;
  with procedure Sin(S1: in Real; S2: in out Real) is <>;
  with function Sin(S: Real) return Real is <>;
  with procedure Log(S1: in Real; S2: in out Real) is <>;
  with function Log(S: Real) return Real is <>;
  with procedure Simple_Random(S: in out Real) is <>;
  with procedure QPower(Q: in Rational; S1: in Real; S2: in out Real) is <>;
  with procedure Roots2(B,C: in Real; U1,U2,V: in out Real) is <>;
  with procedure Roots3(B,C,D: in Real; U0,U1,U2,V: in out Real) is <>;

--- conversion and i/o
  with function Approx(S: Real) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Real) is <>;
  with procedure Assign(Q: in Rational; S: in out Real) is <>;
  with procedure Assign(R: in Flt; S: in out Real) is <>;
  with function Scal(I: Integer) return Real is <>;
  with function Scal(Q: Rational) return Real is <>;
  with function Scal(R: Flt) return Real is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Real) is <>;
  with function Enclose(R1,R2: Flt) return Real is <>;
  with procedure Show1(N: in String; S: in Real; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Real; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Real; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Real; Decimal: in Boolean := False) is <>;

--- misc
  with function Get_Precision(R: Real) return Positive is <>;
  with procedure Set_Active_Mode(Dummy: in Real) is <>;
  with procedure Set_Wait_Mode(Dummy: in Real) is <>;
  with procedure Free_Cache(S: in Real) is <>;
-------------------- end_include scalars  
  
package Scalars is
  
------------------------------------------------------------------  
--- scalar types -------------------------------------------------  
------------------------------------------------------------------
  
-- Real Numbers ------------------------------------------------
  subtype Real_Number is Real;
  package Real_Matrices is new Matrices (Scalar => Real_Number);
  subtype Real_Matrix is Real_Matrices.Matrix;
  subtype Real_Matrix2 is Real_Matrices.Matrix2;
  subtype Real_Matrix3 is Real_Matrices.Matrix3;
  package Real_Triangulars is new Triangulars (Scalar => Real_Number);
  package Real_Vectors is new Vectors (Scalar => Real_Number,
                                       Component => Real_Number,
                                       SM => Real_Matrices);
  package Real_ScalVectors is new ScalVectors (Scalar => Real_Number,
                                               SM => Real_Matrices,
                                               SV => Real_Vectors);
  subtype Real_Vector is Real_Vectors.Vector;
  subtype Real_Diag is Real_ScalVectors.Diag;
  subtype Real_Vector2 is Real_ScalVectors.Vector2;
  subtype Real_Vector3 is Real_ScalVectors.Vector3;
  subtype Real_Triangular is Real_Triangulars.Triangular;
  
-- Real Taylor1 ------------------------------------------------
  package Real_Taylors1 is new Taylors1 (PDeg => SPDeg1,
                                         Scalar => Real_Number,
                                         SM => Real_Matrices,
                                         SV => Real_Vectors);
  subtype Real_Taylor1 is Real_Taylors1.Taylor1;
  use Real_Taylors1;
  package Real_Taylor1_Matrices is new Matrices (Scalar => Real_Taylor1);
  subtype Real_Taylor1_Matrix is Real_Taylor1_Matrices.Matrix;
  subtype Real_Taylor1_Matrix2 is Real_Taylor1_Matrices.Matrix2;
  subtype Real_Taylor1_Matrix3 is Real_Taylor1_Matrices.Matrix3;
  package Real_Taylor1_Triangulars is new Triangulars (Scalar => Real_Taylor1);
  package Real_Taylor1_Vectors is new Vectors (Scalar => Real_Taylor1,
                                               Component => Real_Taylor1,
                                               SM => Real_Taylor1_Matrices);
  package Real_Taylor1_ScalVectors is new ScalVectors (Scalar => Real_Taylor1,
                                                       SM => Real_Taylor1_Matrices,
                                                       SV => Real_Taylor1_Vectors);
  subtype Real_Taylor1_Vector is Real_Taylor1_Vectors.Vector;
  subtype Real_Taylor1_Vector2 is Real_Taylor1_ScalVectors.Vector2;
  subtype Real_Taylor1_Vector3 is Real_Taylor1_ScalVectors.Vector3;
  subtype Real_Taylor1_Triangular is Real_Taylor1_Triangulars.Triangular;
  
-- Complex Numbers ---------------------------------------------
  package Complex_Numbers is new Complexs (Real => Real_Number);
  subtype Complex_Number is Complex_Numbers.Complex;
  use Complex_Numbers;
  package Complex_Matrices is new Matrices (Scalar => Complex_Number);
  subtype Complex_Matrix is Complex_Matrices.Matrix;
  subtype Complex_Matrix2 is Complex_Matrices.Matrix2;
  subtype Complex_Matrix3 is Complex_Matrices.Matrix3;
  package Complex_Triangulars is new Triangulars (Scalar => Complex_Number);
  package Complex_Vectors is new Vectors (Scalar => Complex_Number,
                                          Component => Complex_Number,
                                          SM => Complex_Matrices);
  package Complex_ScalVectors is new ScalVectors (Scalar => Complex_Number,
                                                  SM => Complex_Matrices,
                                                  SV => Complex_Vectors);
  subtype Complex_Vector is Complex_Vectors.Vector;
  subtype Complex_Diag is Complex_ScalVectors.Diag;
  subtype Complex_Vector2 is Complex_ScalVectors.Vector2;
  subtype Complex_Vector3 is Complex_ScalVectors.Vector3;
  subtype Complex_Triangular is Complex_Triangulars.Triangular;

-- Complex Taylor1 ---------------------------------------------
  package Complex_Taylors1 is new Taylors1 (PDeg => SPDeg1,
                                            Scalar => Complex_Number,
                                            SM => Complex_Matrices,
                                            SV => Complex_Vectors);
  subtype Complex_Taylor1 is Complex_Taylors1.Taylor1;
  use Complex_Taylors1;
  package Complex_Taylor1_Matrices is new Matrices (Scalar => Complex_Taylor1);
  subtype Complex_Taylor1_Matrix is Complex_Taylor1_Matrices.Matrix;
  subtype Complex_Taylor1_Matrix2 is Complex_Taylor1_Matrices.Matrix2;
  subtype Complex_Taylor1_Matrix3 is Complex_Taylor1_Matrices.Matrix3;
  package Complex_Taylor1_Triangulars is new Triangulars (Scalar => Complex_Taylor1);
  package Complex_Taylor1_Vectors is new Vectors (Scalar => Complex_Taylor1,
                                                  Component => Complex_Taylor1,
                                                  SM => Complex_Taylor1_Matrices);
  package Complex_Taylor1_ScalVectors is new ScalVectors (Scalar => Complex_Taylor1,
                                                       SM => Complex_Taylor1_Matrices,
                                                       SV => Complex_Taylor1_Vectors);
  subtype Complex_Taylor1_Vector is Complex_Taylor1_Vectors.Vector;
  subtype Complex_Taylor1_Vector2 is Complex_Taylor1_ScalVectors.Vector2;
  subtype Complex_Taylor1_Vector3 is Complex_Taylor1_ScalVectors.Vector3;
  subtype Complex_Taylor1_Triangular is Complex_Taylor1_Triangulars.Triangular;
  
------------------------------------------------------------------
--- conversion real <--> complex ---------------------------------
------------------------------------------------------------------

----------- Numbers
  procedure Set_Real(R: in Real_Number; C: in out Complex_Number);
  function Set_Real(R: Real_Number) return Complex_Number;
  procedure Set_Imag(R: in Real_Number; C: in out Complex_Number);
  procedure Set_Complex(Rx,Ry: in Real_Number; C: in out Complex_Number);
  function Set_Complex(Rx,Ry: Real_Number) return Complex_Number;
  procedure Real_Part(C: in Complex_Number; R: in out Real_Number);
  procedure Imag_Part(C: in Complex_Number; R: in out Real_Number);
  procedure Set_Real(R: in Real_Matrix; C: in out Complex_Matrix);
  procedure Set_Imag(R: in Real_Matrix; C: in out Complex_Matrix);
  procedure Set_Complex(Rx,Ry: in Real_Matrix; C: in out Complex_Matrix);
  procedure Real_Part(C: in Complex_Matrix; R: in out Real_Matrix);
  procedure Imag_Part(C: in Complex_Matrix; R: in out Real_Matrix);
  procedure Set_Real(R: in Real_Triangular; C: in out Complex_Triangular);
  procedure Set_Imag(R: in Real_Triangular; C: in out Complex_Triangular);
  procedure Real_Part(C: in Complex_Triangular; R: in out Real_Triangular);
  procedure Imag_Part(C: in Complex_Triangular; R: in out Real_Triangular);
  procedure Set_Real(R: in Real_Vector; C: in out Complex_Vector);
  procedure Set_Imag(R: in Real_Vector; C: in out Complex_Vector);
  procedure Set_Complex(Rx,Ry: in Real_Vector; C: in out Complex_Vector);
  procedure Real_Part(C: in Complex_Vector; R: in out Real_Vector);
  procedure Imag_Part(C: in Complex_Vector; R: in out Real_Vector);
  procedure Set_Real(R: in Real_Diag; C: in out Complex_Diag);
  procedure Set_Imag(R: in Real_Diag; C: in out Complex_Diag);
----------- Taylor1  
  procedure Set_Real(R: in Real_Taylor1; C: in out Complex_Taylor1);
  function Set_Real(R: Real_Taylor1) return Complex_Taylor1;
  procedure Set_Imag(R: in Real_Taylor1; C: in out Complex_Taylor1);
  procedure Set_Complex(Rx,Ry: in Real_Taylor1; C: in out Complex_Taylor1);
  function Set_Complex(Rx,Ry: Real_Taylor1) return Complex_Taylor1;
  procedure Real_Part(C: in Complex_Taylor1; R: in out Real_Taylor1);
  procedure Imag_Part(C: in Complex_Taylor1; R: in out Real_Taylor1);
  procedure Set_Real(R: in Real_Taylor1_Matrix; C: in out Complex_Taylor1_Matrix);
  procedure Set_Imag(R: in Real_Taylor1_Matrix; C: in out Complex_Taylor1_Matrix);
  procedure Real_Part(C: in Complex_Taylor1_Matrix; R: in out Real_Taylor1_Matrix);
  procedure Imag_Part(C: in Complex_Taylor1_Matrix; R: in out Real_Taylor1_Matrix);
  procedure Set_Real(R: in Real_Taylor1_Triangular; C: in out Complex_Taylor1_Triangular);
  procedure Set_Imag(R: in Real_Taylor1_Triangular; C: in out Complex_Taylor1_Triangular);
  procedure Real_Part(C: in Complex_Taylor1_Triangular; R: in out Real_Taylor1_Triangular);
  procedure Imag_Part(C: in Complex_Taylor1_Triangular; R: in out Real_Taylor1_Triangular);
  procedure Set_Real(R: in Real_Taylor1_Vector; C: in out Complex_Taylor1_Vector);
  procedure Set_Imag(R: in Real_Taylor1_Vector; C: in out Complex_Taylor1_Vector);
  procedure Real_Part(C: in Complex_Taylor1_Vector; R: in out Real_Taylor1_Vector);
  procedure Imag_Part(C: in Complex_Taylor1_Vector; R: in out Real_Taylor1_Vector);
  
------------------------------------------------------------------  
--- assign, evaluate etc -----------------------------------------
------------------------------------------------------------------  
  
  function Rho(V: Real_Taylor1_Vector) return Radius;
  function Rho(V: Complex_Taylor1_Vector) return Radius;
  function Rho(V: Real_Taylor1_Triangular) return Radius;
  function Rho(V: Complex_Taylor1_Triangular) return Radius;
  
  procedure SetRho(R: in Radius; V: in out Real_Taylor1_Vector; Check: in Boolean := True);
  procedure SetRho(R: in Radius; V: in out Complex_Taylor1_Vector; Check: in Boolean := True);
  procedure SetRho(R: in Radius; V: in out Real_Taylor1_Triangular; Check: in Boolean := True);
  procedure SetRho(R: in Radius; V: in out Complex_Taylor1_Triangular; Check: in Boolean := True);
  
  procedure Assign(V: in Real_Vector; W: in out Real_Taylor1_Vector);
  procedure Assign(V: in Real_Vector; W: in out Complex_Taylor1_Vector);
  procedure Assign(V: in Complex_Vector; W: in out Complex_Taylor1_Vector);
  procedure Assign(V: in Real_Triangular; W: in out Real_Taylor1_Triangular);
  procedure Assign(V: in Real_Triangular; W: in out Complex_Taylor1_Triangular);
  procedure Assign(V: in Complex_Triangular; W: in out Complex_Taylor1_Triangular);
  
  procedure Val(V: in Real_Taylor1_Vector; R: in Real_Number; W: in out Real_Vector);
  procedure Val(V: in Complex_Taylor1_Vector; R: in Complex_Number; W: in out Complex_Vector);
  procedure Val(V: in Real_Taylor1_Triangular; R: in Real_Number; W: in out Real_Triangular);
  procedure Val(V: in Complex_Taylor1_Triangular; R: in Complex_Number; W: in out Complex_Triangular);
  
  procedure ValDer(V: in Real_Taylor1_Vector; R: in Real_Number; W: in out Real_Vector);
  procedure ValDer(V: in Complex_Taylor1_Vector; R: in Complex_Number; W: in out Complex_Vector);
  procedure ValDer(V: in Real_Taylor1_Triangular; R: in Real_Number; W: in out Real_Triangular);
  procedure ValDer(V: in Complex_Taylor1_Triangular; R: in Complex_Number; W: in out Complex_Triangular);
  
end Scalars;
