with Rationals;
use Rationals;

pragma Elaborate_All (Rationals);

generic
  
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure HiErrs(S: in out Scalar) is <>;
  
package ScalVectors.Homoclinic is
  
  subtype Cubic1 is Polynom1(0..3);
  
  procedure FHN_Poly(F: in out Cubic1);
  function FHN_Poly return Cubic1;
  procedure Make_A00(C: in Scalar; A: in out Matrix3);
  function Make_A00(C: Scalar) return Matrix3;
  procedure Make_A0Z(C,Z: in Scalar; A: in out Matrix3);
  function Make_A0Z(C,Z: Scalar) return Matrix3;
  procedure Make_AWZ(C,W2,Z: in Scalar; A: in out Matrix3);
  function Make_AWZ(C,W2,Z: Scalar) return Matrix3;
  procedure Make_M1(A: in Matrix3; Mu0,MuZ: in Scalar; M: in out Matrix3);
  procedure Make_M2(A: in Matrix3; Nu0,MuZ: in Scalar; M: in out Matrix3);
  procedure PPoly(C,Z,S: in Scalar; P: in out Scalar);
  procedure QPoly(C,Z,S: in Scalar; Q: in out Scalar);
  
  procedure Eigenvec_A0Z(C,Z,S: in Scalar; U: in out Vector3);
  function Eigenvec_A0Z(C,Z,S: Scalar) return Vector3;
  procedure Eigenvec_A0ZT(C,Z,S: in Scalar; V: in out Vector3);
  function Eigenvec_A0ZT(C,Z,S: Scalar) return Vector3;
  procedure Velocity(C: in Scalar; F: in Cubic1; P: in Vector3; V: in out Vector3);
  function Velocity(C: Scalar; F: Cubic1; P: Vector3) return Vector3;
  
  type FHN_Data0 is tagged
    record
      C: Scalar;     -- velocity
      F: Cubic1;     -- THE function F
      A00: Matrix3;  -- matrix A_0(0)
      Mu0: Scalar;   -- real negative eigenvalue of A00
      Nu0: Scalar;   -- eigenvalue of A00 with positive real part
    end record;
  
  procedure Make_FHN_Data(C,Dummy: in Scalar; D: in out FHN_Data0);
  function Make_FHN_Data(C,Z: Scalar) return FHN_Data0;
  
  type FHN_DataZ is new FHN_Data0 with
    record
      Z: Scalar;
      A0Z: Matrix3;
      MuZ: Scalar;
    end record;
  
  overriding procedure Make_FHN_Data(C,Z: in Scalar; D: in out FHN_DataZ);
  overriding function Make_FHN_Data(C,Z: Scalar) return FHN_DataZ;
  
private
  
  procedure Prepare_A00(C: in Scalar; A: in out Matrix3);
  procedure Sorted_Eigenvals3(A: in Matrix3; U0,U1,U2: in out Scalar);
  
end ScalVectors.Homoclinic;
