with Ada.Text_IO, Logicals, Globals, Ints, Flts, Strings, Messages, Reps, Rep_Balls;
use Ada.Text_IO, Logicals, Globals, Ints, Flts, Strings, Messages, Rep_Balls;
with FHN_Homoclinic;

pragma Elaborate_All (Ada.Text_IO,Logicals,Globals,Ints,Flts,Strings,Rep_Balls);
pragma Elaborate_All (FHN_Homoclinic);

procedure Show_Data is -- display content of data files

  Deg_Taylor1: constant Positive := 255;
  Deg_Taylor2: constant Positive := 255;
  SPDeg1:      constant Positive := 9;
  
  package R_Home renames Rep_Balls;
  use R_Home;
  package FHN is new FHN_Homoclinic (Deg_Taylor1  => Deg_Taylor1,
                                     Deg_Taylor2  => Deg_Taylor2,
                                     SPDeg1 => SPDeg1,
                                     Real => Ball);
  use FHN;
  use Gen_Scalars;  
  
  DataType: constant String := GetArg;
  FileName: constant String := GetArg;
  
  procedure Show_FltVec is
    use Flts.FV;
    V: constant FV.Vector := FV.Read(FileName,0);
  begin
    Show1("",V);
  end Show_FltVec;
  
  procedure Show_R_Scalar is
    use R_Home;
    R: R_Scalar;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_R_Scalar;
  
  procedure Show_C_Scalar is
    use C_Home;
    C: C_Scalar;
  begin
    Read(FileName,C);
    Show1("",C);
  end Show_C_Scalar;
  
  procedure Show_RT1_Scalar is
    use RT1_Home;
    R: RT1_Scalar;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_RT1_Scalar;
  
  procedure Show_RT1_Vector3 is
    use RT1_Vectors;
    R: RT1_Vector3;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_RT1_Vector3;
  
  procedure Show_CT1_Scalar is
    use CT1_Home;
    C: CT1_Scalar;
  begin
    Read(FileName,C);
    Show1("",C);
  end Show_CT1_Scalar;
  
  procedure Show_CT1_Vector3 is
    use CT1_Vectors;
    C: CT1_Vector3;
  begin
    Read(FileName,C);
    Show1("",C);
  end Show_CT1_Vector3;
  
  procedure Show_R_Taylor1 is
    use R_Taylors1;
    R: R_Taylor1;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_R_Taylor1;
  
  procedure Show_R_Taylor1_Vector3 is
    use R_Taylor1_Vectors;
    R: R_Taylor1_Vector3;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_R_Taylor1_Vector3;
  
  procedure Show_RT1_Taylor1 is
    use RT1_Taylors1;
    R: RT1_Taylor1;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_RT1_Taylor1;
  
  procedure Show_RT1_Taylor1_Vector3 is
    use RT1_Taylor1_Vectors;
    R: RT1_Taylor1_Vector3;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_RT1_Taylor1_Vector3;
  
  procedure Show_C_Taylor1 is
    use C_Taylors1;
    R: C_Taylor1;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_C_Taylor1;
  
  procedure Show_C_Taylor1_Vector3 is
    use C_Taylor1_Vectors;
    R: C_Taylor1_Vector3;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_C_Taylor1_Vector3;
  
  procedure Show_C_Taylor2 is
    use C_Taylors2;
    R: C_Taylor2;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_C_Taylor2;
  
  procedure Show_C_Taylor2_Vector3 is
    use C_Taylor2_Vectors;
    R: C_Taylor2_Vector3;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_C_Taylor2_Vector3;
  
  procedure Show_CT1_Taylor1 is
    use CT1_Taylors1;
    R: CT1_Taylor1;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_CT1_Taylor1;
  
  procedure Show_CT1_Taylor1_Vector3 is
    use CT1_Taylor1_Vectors;
    R: CT1_Taylor1_Vector3;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_CT1_Taylor1_Vector3;
  
  procedure Show_CT1_Taylor2 is
    use CT1_Taylors2;
    R: CT1_Taylor2;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_CT1_Taylor2;
  
  procedure Show_CT1_Taylor2_Vector3 is
    use CT1_Taylor2_Vectors;
    R: CT1_Taylor2_Vector3;
  begin
    Read(FileName,R);
    Show1("",R);
  end Show_CT1_Taylor2_Vector3;
  
  procedure Show_R_Curve3 is
    use R_Homoclinic1;
  begin
    Show_Curve3(FileName);
  end Show_R_Curve3;
  
  procedure Show_C_Curve3 is
    use C_Homoclinic1;
  begin
    Show_Curve3(FileName);
  end Show_C_Curve3;
  
begin --- make SPDeg1 sufficiently large
  
  if DataType="FltVec" then Show_FltVec;
  elsif DataType="R_Scalar" then Show_R_Scalar;
  elsif DataType="C_Scalar" then Show_C_Scalar;
  elsif DataType="RT1_Scalar" then Show_RT1_Scalar;
  elsif DataType="RT1_Vector3" then Show_RT1_Vector3;
  elsif DataType="Show_CT1_Scalar" then Show_CT1_Scalar;
  elsif DataType="Show_CT1_Vector3" then Show_CT1_Vector3;
  elsif DataType="R_Taylor1" then Show_R_Taylor1;
  elsif DataType="R_Taylor1_Vector3" then Show_R_Taylor1_Vector3;
  elsif DataType="RT1_Taylor1" then Show_RT1_Taylor1;
  elsif DataType="RT1_Taylor1_Vector3" then Show_RT1_Taylor1_Vector3;
  elsif DataType="C_Taylor1" then Show_C_Taylor1;
  elsif DataType="C_Taylor1_Vector3" then Show_C_Taylor1_Vector3;
  elsif DataType="C_Taylor2" then Show_C_Taylor2;
  elsif DataType="C_Taylor2_Vector3" then Show_C_Taylor2_Vector3;
  elsif DataType="CT1_Taylor1" then Show_CT1_Taylor1;
  elsif DataType="CT1_Taylor1_Vector3" then Show_CT1_Taylor1_Vector3;
  elsif DataType="CT1_Taylor2" then Show_CT1_Taylor2;
  elsif DataType="CT1_Taylor2_Vector3" then Show_CT1_Taylor2_Vector3;
  elsif DataType="R_Curve3" then Show_R_Curve3;
  elsif DataType="C_Curve3" then Show_C_Curve3;
  else Show0("type not yet on the list");
  end if;
  
end Show_Data;
