with Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts;
use Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts;
with FltStr, Ada.Numerics.Generic_Elementary_Functions;

pragma Elaborate_All (Ada.Text_IO,Logicals,Globals,Ints,Rationals,Flts,FltStr);
pragma Elaborate_All (FltStr,Ada.Numerics.Generic_Elementary_Functions);

package Std_Floats is

  pragma Elaborate_Body;

  type LLFloat is new Long_Long_Float;

  package Rep_FS is new FltStr(LLFloat);
  package Rep_IO renames Rep_FS.Flt_IO;
  package Rep_EF is new Ada.Numerics.Generic_Elementary_Functions(LLFloat);

--- auxiliary
  procedure NOP(Dummy: in out LLFloat);
  procedure NOP(Dummy1: in Flt; Dummy2: in out LLFloat);
  function ReturnTrue(Dummy: LLFloat) return Boolean;
  function IsZero(S: LLFloat) return Boolean;
  procedure SetZero(S: in out LLFloat);
  procedure SetZero(Dummy: in Flt; S: in out LLFloat);
  procedure SetZero(Dummy: in LLFloat; S: in out LLFloat);
  function ReturnZero(Dummy: LLFloat) return Flt;
  function ReturnZero(Dummy: Flt) return LLFloat;
  function ReturnZero(Dummy: LLFloat) return LLFloat;
  procedure Copy(S1: in LLFloat; S2: in out LLFloat);
  function Copy(S: LLFloat) return LLFloat;
  function Approx(S: LLFloat) return Flt;
  function FirstArg(Q: Rational; Dummy: Flt) return LLFloat;
  function FirstArg(S: LLFloat; Dummy: Flt) return LLFloat;
  procedure Between(S1,S2: in Flt; S3: in out LLFloat);
  function Between(S1,S2: Flt) return LLFloat;
  procedure Between(S1: in LLFloat; S2: in out LLFloat);
  function Between(S1,S2: LLFloat) return LLFloat;
  
  pragma Inline_Always (NOP,ReturnTrue,IsZero,SetZero,ReturnZero,Copy,Approx,FirstArg,Between);
  
--- basic
  function Info(Dummy: LLFloat) return Scalar_Info;
  function IsSharp(S: LLFloat) return Boolean renames ReturnTrue;
  procedure Swap(S1,S2: in out LLFloat);

--- sets
  function Center0(S: LLFloat) return Boolean renames IsZero;
  function Contains0(S: LLFloat) return Logical;
  function Contains(S1,S2: LLFloat) return Logical;
  procedure BallAt0(R: in Flt; S: in out LLFloat) renames SetZero;
  function BallAt0(R: Flt) return LLFloat renames ReturnZero;
  procedure ToErr(S: in out LLFloat) renames SetZero;
  procedure ToErr(S1: in LLFloat; S2: in out LLFloat) renames SetZero;
  function ToErr(S: LLFloat) return LLFloat renames ReturnZero;
  procedure Center(S: in out LLFloat) renames NOP;
  procedure Center(S1: in LLFloat; S2: in out LLFloat) renames Copy;
  function Center(S: LLFloat) return LLFloat renames Copy;
  procedure ModCenter(S: in out LLFloat) renames SetZero;
  procedure ModCenter(S1: in LLFloat; S2: in out LLFloat) renames SetZero;
  function ModCenter(S: LLFloat) return LLFloat renames ReturnZero;
  procedure ErrMult(R: in Radius; S: in out LLFloat) renames NOP;
  procedure Union(S1: in LLFloat; S2: in out LLFloat) renames Between;
  function Union(S1,S2: LLFloat) return LLFloat renames Between;
  procedure Intersection(S1: in LLFloat; S2: in out LLFloat; Empty: out Logical);

--- order
  function Sign(S: LLFloat) return Integer;
  function Compare(R: Flt; S: LLFloat) return Integer;
  function Compare(S1,S2: LLFloat) return Integer;
  procedure Min(S1: in LLFloat; S2: in out LLFloat);
  procedure Min(S1,S2: in LLFloat; S3: in out LLFloat);
  function Min(S1,S2: LLFloat) return LLFloat renames LLFloat'Min;
  procedure Max(S1: in LLFloat; S2: in out LLFloat);
  procedure Max(S1,S2: in LLFloat; S3: in out LLFloat);
  function Max(S1,S2: LLFloat) return LLFloat renames LLFloat'Max;
  function Sup(S: LLFloat) return Flt renames Approx;
  function Inf(S: LLFloat) return Flt renames Approx;

--- arithmetic
  procedure Neg(S: in out LLFloat);
  procedure Neg(S1: in LLFloat; S2: in out LLFloat);
  procedure Add(I: in Integer; S: in out LLFloat);
  procedure Add(S1: in LLFloat; S2: in out LLFloat);
  procedure Sum(S1,S2: in LLFloat; S3: in out LLFloat);
  procedure Sub(S1: in LLFloat; S2: in out LLFloat);
  procedure Diff(S1,S2: in LLFloat; S3: in out LLFloat);
  function "*"(R: Flt; S: LLFloat) return LLFloat;
  procedure Mult(R: in Flt; S,Dummy: in out LLFloat);
  procedure Mult(R: in Flt; S: in out LLFloat);
  procedure Prod(R: in Flt; S1: in LLFloat; S2: in out LLFloat);
  procedure AddProd(R: in Flt; S1: in LLFloat; S2,Dummy: in out LLFloat);
  procedure AddProd(R: in Flt; S1: in LLFloat; S2: in out LLFloat);
  procedure Mult(Q: in Rational; S: in out LLFloat);
  procedure Mult(S1: in LLFloat; S2: in out LLFloat);
  procedure Prod(S1,S2: in LLFloat; S3: in out LLFloat);
  procedure AddProd(S1,S2: in LLFloat; S3,Dummy: in out LLFloat);
  procedure AddProd(S1,S2: in LLFloat; S3: in out LLFloat);
  procedure SumProd(S1,S2,S3: in LLFloat; S4: in out LLFloat);
  procedure SubProd(S1,S2: in LLFloat; S3,Dummy: in out LLFloat);
  procedure Div(R: in Flt; S,Dummy: in out LLFloat);
  procedure Div(R: in Flt; S: in out LLFloat);
  procedure Quot(S1: in LLFloat; R: in Flt; S2: in out LLFloat);
  function "/"(S: LLFloat; R: Flt) return LLFloat;
  procedure Div(S1: in LLFloat; S2: in out LLFloat);
  procedure Quot(S1,S2: in LLFloat; S3: in out LLFloat);
  procedure Inv(S1: in LLFloat; S2: in out LLFloat);
  function Inv(S: LLFloat) return LLFloat;

--- fun
  function IsReal(S: LLFloat) return Boolean renames ReturnTrue;
  procedure Adjoint(S: in out LLFloat) renames NOP;
  procedure Adjoint(S1: in LLFloat; S2: in out LLFloat) renames Copy;
  function Adjoint(S: LLFloat) return LLFloat renames Copy;
  procedure Real_Part(R: in out LLFloat) renames NOP;
  function Real_Part(R: LLFloat) return LLFloat renames Copy;
  procedure Imag_Part(R: in out LLFloat) renames SetZero;
  function Imag_Part(R: LLFloat) return LLFloat renames ReturnZero;
  procedure Eval0(S: in out LLFloat) renames NOP;
  procedure Norm(S1: in LLFloat; S2: in out LLFloat);
  function Norm(S: LLFloat) return LLFloat renames "abs";
  function MaxNorm(S: LLFloat) return Radius;
  procedure Sqr(S: in out LLFloat);
  function Sqr(S: LLFloat) return LLFloat;
  procedure Sqrt(S1: in LLFloat; S2: in out LLFloat);
  function Sqrt(S: LLFloat) return LLFloat renames Rep_EF.Sqrt;
  procedure Root(K: Positive; S1: in LLFloat; S2: in out LLFloat);
  function Root(K: Positive; S: LLFloat) return LLFloat;
  procedure Exp(S1: in LLFloat; S2: in out LLFloat);
  function Exp(S: LLFloat) return LLFloat renames Rep_EF.Exp;
  procedure Log(S1: in LLFloat; S2: in out LLFloat);
  function Log(S: LLFloat) return LLFloat renames Rep_EF.Log;
  procedure ArcCos(S1: in LLFloat; S2: in out LLFloat);
  function ArcCos(S: LLFloat) return LLFloat renames Rep_EF.ArcCos;
  procedure ArcSin(S1: in LLFloat; S2: in out LLFloat);
  function ArcSin(S: LLFloat) return LLFloat renames Rep_EF.ArcSin;
  procedure Cos(S1: in LLFloat; S2: in out LLFloat);
  function Cos(S: LLFloat) return LLFloat renames Rep_EF.Cos;
  procedure Sin(S1: in LLFloat; S2: in out LLFloat);
  function Sin(S: LLFloat) return LLFloat renames Rep_EF.Sin;
  procedure Simple_Random(S: in out LLFloat);
  function Pi return LLFloat;
  procedure QPower(Q: in Rational; S1: in LLFloat; S2: in out LLFloat);
  procedure Roots2(B,C: in LLFloat; U1,U2,V: in out LLFloat);
  procedure Roots3(B,C,D: in LLFloat; U0,U1,U2,V: in out LLFloat);

--- conversion, io
  procedure Assign(I: in Integer; S: in out LLFloat);
  procedure Assign(Q: in Rational; S: in out LLFloat);
  procedure Assign(R: in Flt; S: in out LLFloat);
  function Scal(I: Integer) return LLFloat;
  function Scal(Q: Rational) return LLFloat;
  function Scal(R: Flt) return LLFloat;
  procedure Enclose(R1,R2: in Flt; S: in out LLFloat) renames Between;
  function Enclose(R1,R2: Flt) return LLFloat renames Between;
  function Val(S: LLFloat; R: Flt) return LLFloat renames FirstArg;
  function ValDec(N: String) return LLFloat renames LLFloat'Value;
  procedure ValHex(N: in String; R: in out LLFloat; Rounded: out Boolean) renames Rep_FS.ValHex;
  function ValHex(N: String) return LLFloat renames Rep_FS.ValHex;
  function DecStr(S: LLFloat; Digs: Natural := 0) return String renames Rep_FS.DecStr;
  function HexStr(S: LLFloat; Digs: Natural := 0) return String renames Rep_FS.HexStr;
  procedure Show1(N: in String; S: in LLFloat; NewLine: in Boolean := True) renames Rep_FS.Show1;
  procedure Show2(N: in String; S1,S2: in LLFloat; NewLine: in Boolean := True) renames Rep_FS.Show2;
  procedure Show3(N: in String; S1,S2,S3: in LLFloat; NewLine: in Boolean := True) renames Rep_FS.Show3;
  procedure Put(F: in File_Type; S: in LLFloat; Decimal: in Boolean := False) renames Rep_FS.Put;
  procedure Get(F: in File_Type; S: in out LLFloat; Decimal: in Boolean := False) renames Rep_FS.Get;
  procedure Write(FileName: in String; S: in LLFloat; Decimal: in Boolean := False) renames Rep_FS.Write;
  procedure Read(FileName: in String; S: in out LLFloat; Decimal: in Boolean := False) renames Rep_FS.Read;
  
  subtype Rep is LLFloat;
  
  procedure Val(S: in LLFloat; R: in out Rep) renames Copy;
  function Val(S: LLFloat) return Rep renames Copy;
  function Upper(S: Rep) return LLFloat renames Copy;
  function Lower(S: Rep) return LLFloat renames Copy;
  procedure PutStd(S: in LLFloat; C,R: out LLFloat);
  procedure GetStd(C,Dummy: in LLFloat; S: in out LLFloat);
  procedure PutNum(S: in LLFloat; R: in out Rep) renames Copy;
  procedure GetNum(R: in Rep; S: in out LLFloat) renames Copy;
  
--- misc
  procedure Set_Precision(Dummy: in Positive);
  function Get_Precision(Dummy: LLFloat) return Positive;
  function Scaled(Eps: Flt) return Flt;
  function Epsilon(Eps: Flt) return Flt;
  procedure Set_Active_Mode(Dummy: in LLFloat);
  procedure Set_Wait_Mode(Dummy: in LLFloat);
  procedure Free_Cache(Dummy: in LLFloat);
  
  Just_Rounded: Boolean renames Rep_FS.Rounded_Last;
  
  pragma Inline_Always (Val,Scal,Swap,Min,Max,Neg,Add,Sum,Sub,Diff,Mult,Prod,AddProd,Div,Quot,Inv,Norm);
  
private
  
  LL_Precision: constant Positive := LLFloat'Machine_Mantissa;
  RepZero:      constant LLFloat  := LLFloat(0);
  RepOne:       constant LLFloat  := LLFloat(1);
  RepHalf:      constant LLFloat  := LLFloat(1)/LLFloat(2);

end Std_Floats;
