with Strings, IPowers, Fun_Series, Newton, Roots;
use Strings;

pragma Elaborate_All (Strings,IPowers,Fun_Series,Newton, Roots);

package body Taylors1 is
  
  use SM,SV,SV_More;
  
  procedure AddProd(S: in Scalar; C1: in Poly1; C2: in out Poly1; Tmp: in out Scalar) is
  begin
    if not IsZero(S) then
      for I in Power loop AddProd(S,C1(I),C2(I),Tmp); end loop;
    end if;
  end AddProd;
  
  procedure SubProd(S: in Scalar; C1: in Poly1; C2: in out Poly1; Tmp: in out Scalar) is
  begin
    if not IsZero(S) then
      for I in Power loop SubProd(S,C1(I),C2(I),Tmp); end loop;
    end if;
  end SubProd;
  
--- basic

  function Info(Dummy: Taylor1) return Scalar_Info is
    pragma Unreferenced(Dummy);
  begin
    return TInfo;
  end Info;

  function IsSharp(P: Taylor1) return Boolean is
  begin
    return STrunc or else IsSharp(P.C);
  end IsSharp;
  
  function IsZero(P: Taylor1) return Boolean is
  begin
    if P.F=FConst then
      return IsZero(P.C(0));
    else
      return IsZero(P.C);
    end if;
  end IsZero;

  function "="(P1,P2: Taylor1) return Boolean is
    C1: Poly1 renames P1.C;
    C2: Poly1 renames P2.C;
  begin
    if (C1 /= C2) then return False; end if;
    if IsSharp(P1) then return True; end if;
    if (P1.R /= P2.R) then return False; end if;
    return TrueF(P1)=TrueF(P2);
  end "=";
  
  procedure SetZero(P: in out Taylor1) is
  begin
    SetZero(P.C);
    P.F := FConst;
    P.R := Rad;
  end SetZero;
  
  procedure Copy(P1: in Taylor1; P2: in out Taylor1) is
  begin
    Copy(P1.C,P2.C);
    P2.F := P1.F;
    P2.R := P1.R;
  end Copy;

  procedure Swap(P1,P2: in out Taylor1) is
    F1: constant Integer := P1.F;
    R1: constant Flt := P1.R;
  begin
    Swap(P1.C,P2.C);
    P1.F := P2.F; P2.F := F1;
    P1.R := P2.R; P2.R := R1;
  end Swap;

--- sets

  function Center0(P: Taylor1) return Boolean is
  begin
    return Center0(P.C);
  end Center0;
  
  function Contains0(P: Taylor1) return Logical is
--- very crude
    L: Logical := Contains0(P.C(0));
  begin
    if P.F=FConst then return L; end if;
    if L=False then return False; end if;
    L := Contains0(P.C);
    if L=True or else STrunc then return L; end if;
    return Uncertain;
  end Contains0;

  function Contains(P1,P2: Taylor1) return Logical is
--- very crude
    C1: Poly1 renames P1.C;
    C2: Poly1 renames P2.C;
    L: Logical := Contains(C1(0),C2(0));
  begin
    if L=False then return False; end if;
    L := Contains(C1,C2);
    if STrunc then return L; end if;
    if L=True then
      if P2.F>PDeg then return True; end if;
      if (P1.R>P2.R) then return False; end if;
      if P1.F=PDeg and then P2.F=PDeg then return True; end if;
      declare
        D1: constant Integer := TrueF(P1);
        D2: constant Integer := TrueF(P2);
      begin
        if D1=PDeg and then D2=PDeg then return True; end if;
        if D1>D2 then return False; end if;
      end;
    end if;
    return Uncertain;
  end Contains;
  
  procedure BallAt0(R: in Flt; P: in out Taylor1) is
  begin
    SetZero(P.C);
    if STrunc then
      P.F := FConst;
    else
      BallAt0(R,P.C(0));
      P.F := 0;
    end if;
    P.R := Rad;
  end BallAt0;
    
  function BallAt0(R: Flt) return Taylor1 is
    P: Taylor1;
  begin
    BallAt0(R,P);
    return P;
  end BallAt0;
  
  procedure ToErr(P: in out Taylor1) is
  begin
    P.F := TrueF(P);
    ToErr(P.C);
  end ToErr;
  
  procedure ToErr(P1: in Taylor1; P2: in out Taylor1) is
  begin
    ToErr(P1.C,P2.C);
    P2.F := TrueF(P1);
    P2.R := P1.R;
  end ToErr;
  
  function ToErr(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    ToErr(P,Q);
    return Q;
  end ToErr;
  
  procedure Center(P: in out Taylor1) is
  begin
    if not STrunc then
      if P.F=FConst then
        Center(P.C(0));
      else
        Center(P.C);
        if QuasiDeg0(P.C) then
          P.F := FConst;
        else
          P.F := PDeg1;
        end if;
      end if;
    end if;
  end Center;
  
  procedure Center(P1: in Taylor1; P2: in out Taylor1) is
  begin
    Center(P1.C,P2.C);
    if QuasiDeg0(P2.C) then
      P2.F := FConst;
    else
      P2.F := PDeg1;
    end if;
    P2.R := Rad;
  end Center;
  
  function Center(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Center(P,Q);
    return Q;
  end Center;
    
  procedure ModCenter(P: in out Taylor1) is
  begin
    if STrunc then
      SetZero(P);
    else
      ModCenter(P.C);
      P.F := TrueF(P);
    end if;
  end ModCenter;
  
  procedure ModCenter(P1: in Taylor1; P2: in out Taylor1) is
  begin
    if STrunc then
      SetZero(P2);
    else
      ModCenter(P1.C,P2.C);
      P2.F := TrueF(P2);
      P2.R := P1.R;
    end if;
  end ModCenter;
  
  function ModCenter(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    ModCenter(P,Q);
    return Q;
  end ModCenter;
  
  procedure ErrMult(R: in Radius; P: in out Taylor1) is
  begin
    if not STrunc then ErrMult(R,P.C); end if;
  end ErrMult;
  
  procedure Union(P1: in Taylor1; P2: in out Taylor1) is
  begin
    Union(P1.C,P2.C);
    if P2.F<P1.F then P2.F := P1.F; end if;
    if P2.R>P1.R then P2.R := P1.R; end if;
  end Union;
    
  function Union(P1,P2: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Copy(P2,Q);
    Union(P1,Q);
    return Q;
  end Union;
  
  procedure Intersection(P1: in Taylor1; P2: in out Taylor1; Empty: out Logical) is
--- far from optimal    
    C1: Poly1 renames P1.C;
    C2: Poly1 renames P2.C;
    F1: constant Integer := TrueF(P1);
    F2: Natural renames P2.F;
    E: Logical;
    Tmp: LT_Pointer;
  begin
    F2 := TrueF(P2);
    if F1>F2 then
      Pool.Allocate(Tmp);
      Copy(P2,Tmp.Data);
      Copy(P1,P2);
      Intersection(Tmp.Data,P2,Empty);
      Pool.Recycle(Tmp);
      return;
    end if;
    if P2.R<P1.R then P2.R := P1.R; end if;
    Empty := False;
    for I in 0 .. IMin(F1,PDeg) loop
      Intersection(C1(I),C2(I),E);
      Empty := Empty or E;
    end loop;
    if F1<PDeg then Empty := Empty or Uncertain; end if;
  end Intersection;
    
--- order
  
  function Sign(P: Taylor1) return Integer is
    Prec: constant Radius := 1.0E-2;
    C: Poly1 renames P.C;
  begin
    if not IsZero(P) then
      if not IsReal(P) then
        raise Undefined with Show0("Taylor1.Sign error: argument non-real");
      end if;
      if Inf(P,Prec)>Zero then return  1; end if;
      if Sup(P,Prec)<Zero then return -1; end if;
      raise Not_Certain;
    end if;
    return 0;
  end Sign;

  function Simple(P: Taylor1) return Boolean is
--- only P.C(0) nonzero
  begin
    return (P.F=FConst) or else QuasiDeg0(P.C);
  end Simple;
  
  function Compare(P1,P2: Taylor1) return Integer is
  begin
    if Simple(P1) and then Simple(P2) then
      return Compare(P1.C(0),P2.C(0));
    end if;
    declare
      S: Integer;
      Tmp: LT_Pointer;
    begin
      Pool.Allocate(Tmp);
      Diff(P1,P2,Tmp.Data);
      S := Sign(Tmp.Data);
      Pool.Recycle(Tmp);
      return S;
    end;
  end Compare;
  
  function "<"(P1,P2: Taylor1) return Boolean is
  begin
    return (Compare(P1,P2) < 0);
  end "<";

  function "<="(P1,P2: Taylor1) return Boolean is
  begin
    return (Compare(P1,P2) <= 0);
  end "<=";

  function ">="(P1,P2: Taylor1) return Boolean is
  begin
    return (Compare(P1,P2) >= 0);
  end ">=";

  function ">"(P1,P2: Taylor1) return Boolean is
  begin
    return (Compare(P1,P2) > 0);
  end ">";
  
  procedure Min(P1: in Taylor1; P2: in out Taylor1) is
  begin
    if P1<P2 then Copy(P1,P2); end if;
  end Min;

  procedure Min(P1,P2: in Taylor1; P3: in out Taylor1) is
  begin
    if P1<P2 then Copy(P1,P3); else Copy(P2,P3); end if;
  end Min;

  function Min(P1,P2: Taylor1) return Taylor1 is
  begin
    if P1<P2 then return P1; else return P2; end if;
  end Min;
  
  procedure Max(P1: in Taylor1; P2: in out Taylor1) is
  begin
    if P1>P2 then Copy(P1,P2); end if;
  end Max;

  procedure Max(P1,P2: in Taylor1; P3: in out Taylor1) is
  begin
    if P1>P2 then Copy(P1,P3); else Copy(P2,P3); end if;
  end Max;

  function Max(P1,P2: Taylor1) return Taylor1 is
  begin
    if P1>P2 then return P1; else return P2; end if;
  end Max;
  
  function QuasiDeg(P: Taylor1) return Natural is
  begin
    if P.F=FConst then
      return 0;
    else
      return QuasiDeg(P.C);
    end if;
  end QuasiDeg;
  
  function GuessMaxAbs(P: Taylor1; D: Natural) return Radius is
    KMax: constant Integer := 1+D/2;
    C: Poly1 renames P.C;
    R: Flt renames P.R;
    X,N: Flt;
    Y: Scalar;
  begin
    N := MaxNorm(C(0));
    for K in -KMax .. KMax loop
      if K /= 0 then
        X := R*Flt(K)/Flt(KMax);
        Copy(C(D),Y);                   -- inlined Evaluate
        for I in reverse 0 .. D-1 loop  -- inlined Evaluate 
          Mult(X,Y);                    -- inlined Evaluate 
          Add(C(I),Y);                  -- inlined Evaluate 
        end loop;                       -- inlined Evaluate Y := p(X)
        N := RMax(MaxNorm(Y),N);
      end if;
    end loop;
    return N;
  end GuessMaxAbs;
  
  function Sharp_Sup(P: Taylor1; Prec: Radius) return Flt is
    D: constant Natural := QuasiDeg(P);
    C: Poly1 renames P.C;
    R: Radius renames P.R;
  begin
    if D<2 then
      if D=0 then return Sup(C(0)); end if;
      return RMax(Sup(Evaluate(C,-R,1)),Sup(Evaluate(C,R,1)));
    end if;
    declare
      Eps: constant Radius := Prec*GuessMaxAbs(P,D);
      V: Flt := Sup(C(0));
      W: Flt;
      Y,Tmp: Scalar;
      
      procedure Val1(X: in Flt) is     -- Y := p(X)
      begin
        Copy(C(D),Y);
        for I in reverse 0 .. D-1 loop
          Mult(X,Y);
          Add(C(I),Y);
        end loop;
      end Val1;
      
      procedure Val2(X1,X2: in Flt) is  -- Y := p([X1,X2])
      begin
        Enclose(X1,X2,Tmp);
        Copy(C(D),Y);
        for I in reverse 0 .. D-1 loop
          Mult(Tmp,Y);
          Add(C(I),Y);
        end loop;
      end Val2;
      
      procedure FindSup(L: in Integer; X1,X2: in Flt) is
      begin
        if STrunc then
          Val1(X2);    -- Y := p(X2)
          W := Sup(Y);
          if W <= V then return; end if;
          Swap(Y,Tmp);
          Val1(X1);    -- Y := p(X1)
          Sub(Y,Tmp);
        else
          Val2(X1,X2); -- Y := p([X1,X2])
          W := Sup(Y);
          if W <= V then return; end if;
          ModCenter(Y,Tmp);
        end if;
        if MaxNorm(Tmp)<Eps then
          V := W;
        elsif L=0 then
          raise Sorry with "Taylors1.Sup: getting tired";
        else
          FindSup(L-1,X1,Half*(X1+X2)); -- half-interval by X1
          FindSup(L-1,Half*(X1+X2),X2); -- half-interval by X2
        end if;
      end FindSup;
      
      L: constant Integer := 128; --- maximal recursion level
      X0,X: Flt := Zero;
    begin
      for K in 1 .. D loop
        X0 := X;
        if K=D then X := R; else X := R*Flt(K)/Flt(D); end if;
        FindSup(L, X0, X);
        FindSup(L,-X0,-X);
      end loop;
      return V;
    end;
  end Sharp_Sup;
  
  function Sup(P: Taylor1; Prec: Radius) return Flt is
--- assumes Flt operations are rounded up
  begin
    if IsSharp(P) then
      return Sharp_Sup(P,Prec);
    else
      declare
        Tmp: LT_Pointer;
        V: Flt;
      begin
        Pool.Allocate(Tmp);
        Center(P,Tmp.Data);
        V := Sup(Tmp.Data,Prec);
        ModCenter(P,Tmp.Data);
        V := V+MaxNorm(Tmp.Data);
        Pool.Recycle(Tmp);
        return V;
      end;
    end if;
  end Sup;
  
  function Sup(P: Taylor1) return Flt is
  begin
    return Sup(P,Psi);
  end Sup;
  
  function Inf(P: Taylor1; Prec: Radius) return Flt is
    R: Flt;
    Tmp: LT_Pointer;
  begin
    Pool.Allocate(Tmp);
    Neg(P,Tmp.Data);
    R := Sup(Tmp.Data,Prec);
    Pool.Recycle(Tmp);
    return -R;
  end Inf;
  
  function Inf(P: Taylor1) return Flt is
  begin
    return Inf(P,Psi);
  end Inf;
  
--- addition and multiplication etc.

  procedure Neg(P: in out Taylor1) is
  begin
    Neg(P.C);
  end Neg;

  procedure Neg(P1: in Taylor1; P2: in out Taylor1) is
  begin
    Neg(P1.C,P2.C);
    P2.F := P1.F;
    P2.R := P1.R;
  end Neg;

  function "-"(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Neg(P.C,Q.C);
    Q.F := P.F;
    Q.R := P.R;
    return Q;
  end "-";
  
  procedure Add(I: in Integer; P: in out Taylor1) is
  begin
    Add(I,P.C(0));
  end Add;
  
  procedure Add(P1: in Taylor1; P2: in out Taylor1) is
  begin
    Add(P1.C,P2.C);
    if P2.F>P1.F then P2.F := P1.F; end if;
    if P2.R>P1.R then P2.R := P1.R; end if;
  end Add;

  procedure Sum(P1,P2: in Taylor1; P3: in out Taylor1) is
  begin
    Sum(P1.C,P2.C,P3.C);
    P3.F := IMin(P1.F,P2.F);
    P3.R := RMin(P1.R,P2.R);
  end Sum;

  function "+"(P1,P2: Taylor1) return Taylor1 is
    P3: Taylor1;
  begin
    Sum(P1,P2,P3);
    return P3;
  end "+";

  procedure Sub(P1: in Taylor1; P2: in out Taylor1) is
  begin
    Sub(P1.C,P2.C);
    if P2.F>P1.F then P2.F := P1.F; end if;
    if P2.R>P1.R then P2.R := P1.R; end if;
  end Sub;

  procedure Diff(P1,P2: in Taylor1; P3: in out Taylor1) is
  begin
    Diff(P1.C,P2.C,P3.C);
    P3.F := IMin(P1.F,P2.F);
    P3.R := RMin(P1.R,P2.R);
  end Diff;

  function "-"(P1,P2: Taylor1) return Taylor1 is
    P3: Taylor1;
  begin
    Diff(P1,P2,P3);
    return P3;
  end "-";
  
  procedure Mult(R: in Flt; P,Tmp: in out Taylor1) is
    STmp: Scalar renames Tmp.C(0);
  begin
    if R=Zero then
      SetZero(P);
    else
      Mult(R,P.C,STmp);
    end if;
  end Mult;
  
  procedure Mult(R: in Flt; P: in out Taylor1) is
  begin
    if R=Zero then
      SetZero(P);
    else
      declare
        STmp: Scalar;
      begin
        Mult(R,P.C,STmp);
      end;
    end if;
  end Mult;
  
  procedure Prod(R: in Flt; P1: in Taylor1; P2: in out Taylor1) is
  begin
    if R=Zero then
      SetZero(P2);
    else
      Prod(R,P1.C,P2.C);
      P2.F := P1.F;
      P2.R := P1.R;
    end if;
  end Prod;
  
  function "*"(R: Flt; P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Prod(R,P,Q);
    return Q;
  end "*";
  
  procedure AddProd(R: in Flt; P1: in Taylor1; P2,Tmp: in out Taylor1) is
    STmp: Scalar renames Tmp.C(0);
  begin
    if R /= Zero then
      AddProd(R,P1.C,P2.C,STmp);
      if P2.F>P1.F then P2.F := P1.F; end if;
      if P2.R>P1.R then P2.R := P1.R; end if;
    end if;
  end AddProd;

  procedure AddProd(R: in Flt; P1: in Taylor1; P2: in out Taylor1) is
  begin
    if R /= Zero then
      declare
        STmp: Scalar;
      begin
        AddProd(R,P1.C,P2.C,STmp);
      end;
      if P2.F>P1.F then P2.F := P1.F; end if;
      if P2.R>P1.R then P2.R := P1.R; end if;
    end if;
  end AddProd;
  
  procedure Mult(Q: in Rational; P: in out Taylor1) is
  begin
    if Num(Q)=0 then
      SetZero(P);
    else
      declare
        Tmp: Scalar;
      begin
        Mult(Flt(Num(Q)),P.C,Tmp);
        Div(Flt(Den(Q)),P.C,Tmp);
      end;
    end if;
  end Mult;
  
  procedure Mult(P1: in Taylor1; P2,Tmp: in out Taylor1) is
  begin
    if P1.F=FConst then
      Mult(P1.C(0),P2);
    elsif P2.F=FConst then
      declare
        STmp: Scalar renames Tmp.C(0);
      begin
        Copy(P2.C(0),STmp);
        Prod(STmp,P1,P2); 
      end;
    else
      Copy(P2,Tmp);
      SetZero(P2);
      AddProd(P1,Tmp,P2);
    end if;
  end Mult;
  
  procedure Mult(P1: in Taylor1; P2: in out Taylor1) is
    PTmp: LT_Pointer;
  begin
    if P1.F=FConst then
      Mult(P1.C(0),P2);
    elsif P2.F=FConst then
      declare
        STmp: Scalar;
      begin
        Copy(P2.C(0),STmp);
        Prod(STmp,P1,P2); 
      end;
    else
      Pool.Allocate(PTmp);
      Copy(P2,PTmp.Data);
      SetZero(P2);
      AddProd(P1,PTmp.Data,P2);
      Pool.Recycle(PTmp);
    end if;
  end Mult;
  
  procedure Prod(P1,P2: in Taylor1; P3: in out Taylor1) is
  begin
    if P1.F=FConst then
      Prod(P1.C(0),P2,P3);
    elsif P2.F=FConst then
      Prod(P2.C(0),P1,P3);
    else
      SetZero(P3);
      AddProd(P1,P2,P3);
    end if;
  end Prod;
  
  function "*"(P1,P2: Taylor1) return Taylor1 is
    P3: Taylor1;
  begin
    Prod(P1,P2,P3);
    return P3;
  end "*";
  
  procedure AddProd(P,Q: in Taylor1; H: in out Taylor1) is
    PC: Poly1 renames P.C;
    QC: Poly1 renames Q.C;
    HC: Poly1 renames H.C;
    P0: Scalar renames PC(0);
    Q0: Scalar renames QC(0);
  begin
    if P.F=FConst then AddProd(P0,Q,H); return; end if;
    if Q.F=FConst then AddProd(Q0,P,H); return; end if;
    if H.F>P.F then H.F := P.F; end if;
    if H.F>Q.F then H.F := Q.F; end if;
    if H.R>P.R then H.R := P.R; end if;
    if H.R>Q.R then H.R := Q.R; end if;
    if PDeg<3 then
      declare
        P1: Scalar renames PC(1);
        Q1: Scalar renames QC(1);
        H0: Scalar renames HC(0);
        H1: Scalar renames HC(1);
        Tmp: Scalar;
      begin
        AddProd(P0,Q0,H0,Tmp);
        AddProd(P0,Q1,H1,Tmp);
        AddProd(P1,Q0,H1,Tmp);
        if PDeg=1 then                      -- case PDeg=1
          if not STrunc then
            declare
              E: Scalar;
            begin
              Prod(P1,Q1,E);
              ToErr(E);
              ErrMult(Q.R,E);
              Add(E,H1);
            end;
          end if;
          return;
        end if;
        declare                              -- case PDeg=2
          P2: Scalar renames PC(PDeg);
          Q2: Scalar renames QC(PDeg);
          H2: Scalar renames HC(PDeg);
        begin
          AddProd(P0,Q2,H2,Tmp);
          AddProd(P1,Q1,H2,Tmp);
          AddProd(P2,Q0,H2,Tmp);
          if not STrunc then
            declare
              E: Scalar;
            begin
              Prod(P2,Q2,E);
              ToErr(E);
              ErrMult(Q.R,E);
              AddProd(P1,Q2,E,Tmp);
              AddProd(P2,Q1,E,Tmp);
              ToErr(E);
              ErrMult(Q.R,E);
              Add(E,H2);
            end;
          end if;
        end;
      end;
      return;
    end if;
    declare                        
      DP: constant Integer := QuasiDeg(PC);  -- case PDeg>2
      DQ: constant Integer := QuasiDeg(QC);
      DH: constant Integer := DP+DQ;
    begin
      if DP=0 then AddProd(P0,Q,H); return; end if;
      if DQ=0 then AddProd(Q0,P,H); return; end if;
      declare
        procedure Low_Prod is                -- coeffcicients
          Tmp: Scalar;  
        begin
          for D in Power loop
            declare
              HCD: Scalar renames HC(D);
            begin
              for I in 0 .. D loop
                AddProd(PC(I),QC(D-I),HCD,Tmp);
              end loop;
            end;
          end loop;
        end Low_Prod;
      begin
        if STrunc or else DH <= PDeg then Low_Prod; return; end if;
        declare
          E: Scalar;
          procedure Err_Prod is              -- error estimate
            S,Tmp: Scalar;
          begin
            for D in reverse PDeg+1 .. DH loop
              SetZero(S);
              for I in IMax(0,D-DQ) .. IMin(D,DP) loop
                AddProd(PC(I),QC(D-I),S,Tmp);
              end loop;
              ToErr(S);
              Add(S,E);
              Mult(Q.R,E,Tmp);
            end loop;
          end Err_Prod;
        begin
          SetZero(E);
          if Prod_Parallel then
            declare
              task Run_Err_Prod is end Run_Err_Prod;
              task body Run_Err_Prod is begin Err_Prod; end Run_Err_Prod;
            begin
              Low_Prod;
            end;
          else
            Low_Prod;
            Err_Prod;
          end if;
          Add(E,HC(PDeg));
        end;
      end;
    end;
  end AddProd;
  
  procedure AddProd(P1,P2: in Taylor1; P3,Dummy: in out Taylor1) is
    pragma Unreferenced (Dummy);
  begin
    AddProd(P1,P2,P3);
  end AddProd;
  
  procedure SubProd(P1,P2: in Taylor1; P3,Dummy: in out Taylor1) is
    pragma Unreferenced (Dummy);
  begin
    Neg(P3);
    AddProd(P1,P2,P3);
    Neg(P3);
  end SubProd;
  
  procedure Div(R: in Flt; P,Tmp: in out Taylor1) is
    STmp: Scalar renames Tmp.C(0);
  begin
    Div(R,P.C,STmp);
  end Div;
  
  procedure Div(R: in Flt; P: in out Taylor1) is
    STmp: Scalar;
  begin
    Div(R,P.C,STmp);
  end Div;
  
  procedure Quot(P1: in Taylor1; R: in Flt; P2: in out Taylor1) is
  begin
    Quot(P1.C,R,P2.C);
    P2.F := P1.F;
    P2.R := P1.R;
  end Quot;

  function "/"(P: Taylor1; R: Flt) return Taylor1 is
    Q: Taylor1;
  begin
    Quot(P,R,Q);
    return Q;
  end "/";

  procedure Div(P1: in Taylor1; P2: in out Taylor1) is
    Tmp: LT_Pointer;
  begin
    Pool.Allocate(Tmp);
    Inv(P1,Tmp.Data);
    Mult(Tmp.Data,P2);
    Pool.Recycle(Tmp);
  end Div;

  procedure Quot(P1,P2: in Taylor1; P3: in out Taylor1) is
  begin
    Inv(P2,P3);
    Mult(P1,P3);
  end Quot;

  function "/"(P1,P2: Taylor1) return Taylor1 is
    P: Taylor1;
  begin
    Inv(P2,P);
    Mult(P1,P);
    return P;
  end "/";
  
  procedure Inv(P: in out Taylor1) is
    Tmp: LT_Pointer;
  begin
    Pool.Allocate(Tmp);
    Inv(P,Tmp.Data);
    Copy(Tmp.Data,P);
    Pool.Recycle(Tmp);
  end Inv;

  function Inv(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Inv(P,Q);
    return Q;
  end Inv;
  
--- functions
  
  function IsReal(P: Taylor1) return Boolean is
  begin
    if TInfo.IsComplex then
      if P.F=FConst then
        return IsReal(P.C(0));
      else
        return IsReal(P.C);
      end if;
    else
      return True;
    end if;
  end IsReal;
  
  procedure Adjoint(P: in out Taylor1) is
  begin
    if TInfo.IsComplex then Adjoint(P.C); end if;
  end Adjoint;
  
  procedure Adjoint(P1: in Taylor1; P2: in out Taylor1) is
  begin
    Adjoint(P1.C,P2.C);
    P2.F := P1.F;
    P2.R := P1.R;
  end Adjoint;

  function Adjoint(P: Taylor1) return Taylor1 is
  begin
    if TInfo.IsComplex then
      declare
        Q: Taylor1;
      begin
        Adjoint(P,Q);
        return Q;
      end;
    end if;
    return P;
  end Adjoint;

  procedure Real_Part(P: in out Taylor1) is
  begin
    if TInfo.IsComplex then
      Real_Part(P.C);
      P.F := TrueF(P);
    end if;
  end Real_Part;

  procedure Imag_Part(P: in out Taylor1) is
  begin
    if TInfo.IsComplex then
      Imag_Part(P.C);
      P.F := TrueF(P);
    else
      SetZero(P);
    end if;
  end Imag_Part;
  
  procedure Eval0(P: in out Taylor1) is
    C: Poly1 renames P.C;
  begin
    Eval0(C(0));
    for I in 1 .. PDeg loop SetZero(C(I)); end loop;
    P.F := FConst;
  end Eval0;
  
  procedure Norm(P: in Taylor1; Q: in out Taylor1) is
  begin
    SetZero(Q);
    Norm1(P,Q.C(0));
  end Norm;

  function Norm(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    SetZero(Q);
    Norm1(P,Q.C(0));
    return Q;
  end Norm;

  function MaxNorm(P: Taylor1) return Radius is
--- assumes Flt operations are rounded up
    C: Poly1 renames P.C;
    R: Radius renames P.R;
    N: Flt := Zero;
  begin
    for I in reverse 0 .. QuasiDeg(C) loop
      N := R*N+MaxNorm(C(I));
    end loop;
    return N;
  end MaxNorm;
  
  procedure Sqr(P: in out Taylor1) is
    Tmp: LT_Pointer;
  begin
    if P.F=FConst then
      Sqr(P.C(0));
    else
      Pool.Allocate(Tmp);
      Copy(P,Tmp.Data);
      Mult(Tmp.Data,P);
      Pool.Recycle(Tmp);
    end if;
  end Sqr;
  
  function Sqr(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Copy(P,Q);
    Mult(P,Q);
    return Q;
  end Sqr;
  
  function Sqrt(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Sqrt(P,Q);
    return Q;
  end Sqrt;
  
  function Root(K: Positive; P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Root(K,P,Q);
    return Q;
  end Root;
  
  function Exp(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Exp(P,Q);
    return Q;
  end Exp;
  
  function ArcCos(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    ArcCos(P,Q);
    return Q;
  end ArcCos;
  
  function ArcSin(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    ArcSin(P,Q);
    return Q;
  end ArcSin;
  
  procedure Cos(P1: in Taylor1; P2: in out Taylor1) is
    Tmp: LT_Pointer;
  begin
    if P1.F=FConst then
      SetZero(P2);
      Cos(P1.C(0),P2.C(0));
    else
      Pool.Allocate(Tmp);
      CosSin(P1,P2,Tmp.Data);
      Pool.Recycle(Tmp);
    end if;
  end Cos;
  
  function Cos(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Cos(P,Q);
    return Q;
  end Cos;
  
  procedure Sin(P1: in Taylor1; P2: in out Taylor1) is
    Tmp: LT_Pointer;
  begin
    if P1.F=FConst then
      SetZero(P2);
      Sin(P1.C(0),P2.C(0));
    else
      Pool.Allocate(Tmp);
      CosSin(P1,Tmp.Data,P2);
      Pool.Recycle(Tmp);
    end if;
  end Sin;
  
  function Sin(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Sin(P,Q);
    return Q;
  end Sin;

  function Log(P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Log(P,Q);
    return Q;
  end Log;
  
  procedure Simple_Random(P: in out Taylor1) is
    R: constant Radius := P.R*Flt(9)/Flt(8); -- ad hoc
    C: Poly1 renames P.C;
    PowR: Flt := One;
  begin
    Simple_Random(C(0));
    for I in 1 .. PDeg loop
      PowR := PowR/R;
      Simple_Random(C(I));
      Mult(PowR,C(I));
    end loop;
    P.F := PDeg1;
  end Simple_Random;
  
--- conversion and i/o

  function Approx(P: Taylor1) return Flt is
  begin
    return Approx(P.C(0));
  end Approx;

  procedure Assign(I: in Integer; P: in out Taylor1) is
  begin
    SetZero(P);
    Assign(I,P.C(0));
  end Assign;
  
  procedure Assign(Q: in Rational; P: in out Taylor1) is
  begin
    SetZero(P);
    Assign(Q,P.C(0));
  end Assign;
  
  procedure Assign(R: in Flt; P: in out Taylor1) is
  begin
    SetZero(P);
    Assign(R,P.C(0));
  end Assign;
  
  function Scal(I: Integer) return Taylor1 is
    P: Taylor1;
  begin
    SetZero(P);
    Assign(I,P.C(0));
    return P;
  end Scal;
  
  function Scal(Q: Rational) return Taylor1 is
    P: Taylor1;
  begin
    SetZero(P);
    Assign(Q,P.C(0));
    return P;
  end Scal;

  function Scal(R: Flt) return Taylor1 is
    P: Taylor1;
  begin
    SetZero(P);
    Assign(R,P.C(0));
    return P;
  end Scal;
  
  procedure Enclose(R1,R2: in Flt; P: in out Taylor1) is
  begin
    SetZero(P);
    Enclose(R1,R2,P.C(0));
  end Enclose;
  
  function Enclose(R1,R2: Flt) return Taylor1 is
    P: Taylor1;
  begin
    SetZero(P);
    Enclose(R1,R2,P.C(0));
    return P;
  end Enclose;

  procedure Show1(N: in String; P: in Taylor1; Hide0: in Boolean := True) is
  begin
    if not IsZero(P) then
      Show1(N & "F ",P.F);
      Show1(N & "R ",P.R);
      Show1(N & "C ",P.C);
    elsif not Hide0 then
      Show0(N & "0");
    end if;
  end Show1;

  procedure Show2(N: in String; P1,P2: in Taylor1; Hide0: in Boolean := True) is
  begin
    if not (IsZero(P1) and IsZero(P2)) then
      Show2(N & "F ",P1.F,P2.F);
      Show2(N & "R ",P1.R,P2.R);
      Show2(N & "C ",P1.C,P2.C);
    elsif not Hide0 then
      Show0(N & "0 0");
    end if;
  end Show2;

  procedure Put(F: in File_Type; P: in Taylor1; Decimal: in Boolean := False) is
    C: Poly1 renames P.C;    
    Deg: constant Integer := QuasiDeg(C);
    PF:  constant Integer := IMin(TrueF(P),Deg+1);
  begin
    Put(F,PF);
    Put(F,P.R,Decimal);
    Put(F,Deg);
    for I in 0 .. Deg loop
      Put(F,C(I),Decimal);
    end loop;
  end Put;

  procedure Get(F: in File_Type; P: in out Taylor1; Decimal: in Boolean := False) is
    Deg: Integer := 0;
    S: Scalar;
  begin
    SetZero(P);
    Get(F,P.F);
    Get(F,P.R,Decimal);
    Get(F,Deg);
    for I in 0 .. Deg loop
      Get(F,S,Decimal);
      AddCoeff(I,S,P);
    end loop;
    P.F := TrueF(P);
  end Get;

  procedure Write(FileName: in String; P: in Taylor1; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,P,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; P: in out Taylor1; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,P,Decimal);
    Close(F);
  end Read;
  
  function Read(FileName: String; Decimal: Boolean := False) return Taylor1 is
    P: Taylor1;
  begin
    Read(FileName,P,Decimal);
    return P;
  end Read;
  
--- misc

  function Get_Precision(P: Taylor1) return Positive is
  begin
    return Get_Precision(P.C(0));
  end Get_Precision;

  procedure Set_Active_Mode(Dummy: in Taylor1) is
    pragma Unreferenced (Dummy);
  begin
    Set_Active_Mode(SZero);
  end Set_Active_Mode;

  procedure Set_Wait_Mode(Dummy: in Taylor1) is
    pragma Unreferenced (Dummy);
  begin
    Set_Wait_Mode(SZero);
  end Set_Wait_Mode;
  
  procedure Free_Cache(Dummy: in Taylor1) is
    pragma Unreferenced (Dummy);
  begin
    Pool.Free_All;
  end Free_Cache;
  
--------------- other standard procedures
  
  function Rho(P: Taylor1) return Radius is
  begin
    return P.R;
  end Rho;
  
  procedure SetRho(R: in Radius; P: in out Taylor1; Check: in Boolean := True) is
  begin
    P.F := TrueF(P);
    if (P.F <= PDeg) and then Check and then (R>P.R) then
      raise Domain_Error with Show0("Taylors1.SetRho: cannot increase domain");
    end if;
    P.R := R;
  end SetRho;

  function TrueF(P: Taylor1) return Natural is
--- the value FConst means that P is constant
--- the value PDeg1 means that P is a polynomial
    C: Poly1 renames P.C;
    Deg: constant Integer := QuasiDeg(C);
  begin
    if STrunc then
      if Deg=0 then return FConst; end if;
      return PDeg1;
    end if;
    if Deg=0 then
      if P.F>0 or else IsSharp(C(0)) then return FConst; end if;
      return 0;
    end if;
    for I in P.F .. Deg loop
      if not IsSharp(C(I)) then return I; end if;
    end loop;
    return PDeg1;
  end TrueF;

  procedure AdjustF(P: in out Taylor1) is
  begin
    P.F := TrueF(P);
  end AdjustF;
  
  procedure NonConst(P: in out Taylor1) is
  begin
    if P.F=FConst then P.F := PDeg1; end if;
  end NonConst;
  
  procedure LiftErrs(P: in out Taylor1; F: in Power := 0) is
  begin
    P.F := TrueF(P);
    if F>P.F then
      declare
        R: Radius renames P.R;
        C: Poly1 renames P.C;
        E: Scalar;
      begin
        for K in P.F .. F-1 loop
          ModCenter(C(K),E);
          Div(R,E);
          Add(E,C(K+1));
        end loop;
      end;
      P.F := F;
    end if;
  end LiftErrs;
  
  procedure HiErrs(P: in out Taylor1) is
  begin
    LiftErrs(P,PDeg);
  end HiErrs;
  
  procedure Assign(S: in Scalar; P: in out Taylor1) is
  begin
    SetZero(P);
    Copy(S,P.C(0));
  end Assign;
  
  function Scal(S: Scalar) return Taylor1 is
    P: Taylor1;
  begin
    SetZero(P);
    Copy(S,P.C(0));
    return P;
  end Scal;
  
  procedure SetZero(R: in Radius; P: in out Taylor1) is
  begin
    SetZero(P.C);
    P.F := FConst;
    P.R := R;
  end SetZero;
  
  procedure Affine(C0,C1: in Scalar; P: in out Taylor1) is
    C: Poly1 renames P.C;
  begin
    SetZero(C);
    Copy(C0,C(0));
    Copy(C1,C(1));
    P.F := PDeg1;
    P.R := Rad;
  end Affine;
  
  procedure Coeff(I: in Natural; P: in Taylor1; S: in out Scalar) is
    F: constant Integer := TrueF(P);
    C: Poly1 renames P.C;
  begin
    if I <= F then
      if I <= PDeg then Copy(C(I),S); else SetZero(S); end if;
    else
      SetZero(S);
      if not STrunc then
        declare
          R: Radius renames P.R;
          E: Scalar;
        begin
          for D in F .. IMin(I-1,PDeg) loop
            Div(R,S);
            ModCenter(C(D),E);
            Add(E,S);
          end loop;
          if I <= PDeg then
            Div(R,S);
            Add(C(I),S);
          else
            Mult(Scal(R)**(PDeg-I),S);
          end if;
        end;
      end if;
    end if;
  end Coeff;

  function Coeff(I: Natural; P: Taylor1) return Scalar is
    S: Scalar;
  begin
    Coeff(I,P,S);
    return S;
  end Coeff;
  
  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out Taylor1) is
--- assuming Flt ops are rounded up
  begin
    if not IsZero(S) then
      if I <= PDeg then
        Add(S,P.C(I));
        if P.F=FConst and then I>0 then P.F := PDeg1; end if;
      elsif not STrunc then
        AddProd(P.R**(I-PDeg),ToErr(S),P.C(PDeg));
        if P.F>PDeg then P.F := PDeg; end if;
      end if;
    end if;
  end AddCoeff;
  
  procedure Add(S: in Scalar; P: in out Taylor1) is
  begin
    Add(S,P.C(0));
  end Add;
  
  procedure Mult(S: in Scalar; P: in out Taylor1) is
  begin
    if IsZero(S) then SetZero(P); else Mult(S,P.C); end if;
  end Mult;

  procedure Prod(S: in Scalar; P1: in Taylor1; P2: in out Taylor1) is
  begin
    if IsZero(S) then
      SetZero(P2);
    else
      Prod(S,P1.C,P2.C);
      P2.F := P1.F;
      P2.R := P1.R;
    end if;
  end Prod;

  function "*"(S: Scalar; P: Taylor1) return Taylor1 is
    Q: Taylor1;
  begin
    Prod(S,P,Q);
    return Q;
  end "*";

  procedure AddProd(S: in Scalar; P1: in Taylor1; P2: in out Taylor1) is
  begin
    if not IsZero(S) then
      declare
        Tmp: Scalar;
      begin
        AddProd(S,P1.C,P2.C,Tmp);
      end;
      if P2.F>P1.F then P2.F := P1.F; end if;
      if P2.R>P1.R then P2.R := P1.R; end if;
    end if;
  end AddProd;
  
  procedure SubProd(S: in Scalar; P1: in Taylor1; P2: in out Taylor1) is
  begin
    if not IsZero(S) then
      declare
        Tmp: Scalar;
      begin
        SubProd(S,P1.C,P2.C,Tmp);
      end;
      if P2.F>P1.F then P2.F := P1.F; end if;
      if P2.R>P1.R then P2.R := P1.R; end if;
    end if;
  end SubProd;
  
  procedure Div(S: in Scalar; P: in out Taylor1) is
    Tmp: Scalar;
  begin
    Inv(S,Tmp);
    Mult(Tmp,P.C);
  end Div;
  
  procedure Val(P: in Taylor1; S1: in Scalar; S2: in out Scalar) is
  begin
    if (P.F <= PDeg) and then MaxNorm(S1)>P.R then
      Show2("",MaxNorm(S1),P.R);
      raise Undefined with "Taylor1s.Val error: argument too large";
    end if;
    Evaluate(P.C,S1,S2);
  end Val;
  
  function Val(P: Taylor1; S: Scalar) return Scalar is
    V: Scalar;
  begin
    Val(P,S,V);
    return V;
  end Val;
  
  procedure Val(P1: in Taylor1; S: in Scalar; P2: in out Taylor1) is
  begin
    SetZero(P2);
    Val(P1,S,P2.C(0));
  end Val;
  
  procedure ValBall(P: in Taylor1; R: in Radius; S: in out Scalar) is
--- assumes Flt operations are rounded up
    C: Poly1 renames P.C;
    N: Flt := Zero;
  begin
    if P.F=FConst then
      Copy(C(0),S);
    elsif (R <= P.R) or else (P.F>PDeg) then
      for I in reverse 1 .. PDeg loop
        N := R*(N+MaxNorm(C(I)));
      end loop;
      BallAt0(N,S);
      Add(P.C(0),S);
    else
      raise Undefined with "Taylor1s.ValBall error: argument too large";
    end if;
  end ValBall;
  
  procedure Ran(P: in Taylor1; S: in out Scalar) is
  begin
    ValBall(P,P.R,S);
  end Ran;
  
  function NotSharp(C: Poly1; F,L: Integer) return Boolean is
  begin
    if not STrunc then
      for I in F .. L loop
        if not IsSharp(C(I)) then return True; end if;
      end loop;
    end if;
    return False;
  end NotSharp;
  
  function MaxNorm(R: Radius; P: Taylor1) return Radius is
--- assumes Flt operations are rounded up
    C: Poly1 renames P.C;
    N: Flt := Zero;
  begin
    for I in reverse 0 .. QuasiDeg(C) loop
      N := R*N+MaxNorm(C(I));
    end loop;
    return N;
  end MaxNorm;
  
  procedure Norm1(R: in Radius; P: in Taylor1; S: in out Scalar) is
    C: Poly1 renames P.C;
    D: constant Integer := QuasiDeg(C);
    Tmp: Scalar;
  begin
    if R>P.R and then NotSharp(C,P.F,D) then
      raise Domain_Error with Show0("Taylors1.Norm1: R>P.R");
    end if;
    Norm(C(D),S);
    for I in reverse 0 .. D-1 loop
      Mult(R,S,Tmp);
      Norm(C(I),Tmp);
      Add(Tmp,S);
    end loop;
  end Norm1;
  
  procedure Norm1(P: in Taylor1; S: in out Scalar) is
  begin
    Norm1(P.R,P,S);
  end Norm1;
  
  function Norm1(R: Radius; P: Taylor1) return Scalar is
    S: Scalar;
  begin
    Norm1(R,P,S);
    return S;
  end Norm1;
  
  function Norm1(P: Taylor1) return Scalar is
    S: Scalar;
  begin
    Norm1(P.R,P,S);
    return S;
  end Norm1;
  
  procedure Monom(I: in Power; P: in out Taylor1) is
  begin
    SetZero(P.C);
    Assign(1,P.C(I));
    if I=0 then P.F := FConst; else P.F := PDeg1; end if;
    P.R := Rad;
  end Monom;
  
  procedure HiUnitBall(I: in Power; P: in out Taylor1; R: in Radius := Rad) is
  begin
    SetZero(P.C);
    P.C(I) :=  Scal(R)**(-I);
    ToErr(P.C(I));
    P.F := I;
    P.R := R;
  end HiUnitBall;
  
  procedure Compose(C: in SV.Polynom1; P1: in Taylor1; P2: in out Taylor1) is
    D: constant Integer := QuasiDeg(C);
  begin
    SetZero(P2);
    Add(C(D),P2.C(0));
    for I in reverse 0 .. D-1 loop
      Mult(P1,P2);
      Add(C(I),P2.C(0));
    end loop;
  end Compose;  
  
  procedure Translate(P1: in Poly1; T: in Flt; P2: in out Poly1) is
    Tmp: LT_Pointer;
  begin
    Pool.Allocate(Tmp);
    declare
      X: Poly1 renames Tmp.Data.C;
    begin
      Copy(P1,P2);
      Assign(1,X(0));
      for N in 1 .. QuasiDeg(P1) loop
        Assign(1,X(N));
        for K in reverse 1 .. N-1 loop  -- produce X(t)=(t+S)^N
          Mult(T,X(K));
          Add(X(K-1),X(K));
          AddProd(X(K),P1(N),P2(K));    -- and add P1(N)*X to P2
        end loop;
        Mult(T,X(0));
        AddProd(X(0),P1(N),P2(0));
      end loop;
    end;
    Pool.Recycle(Tmp);
  end Translate;
  
  procedure Translate(P1: in Taylor1; T: in Flt; P2: in out Taylor1; R: in Radius) is
--- assumes Flt operations are rounded up
  begin
    if T=Zero or else P1.F=FConst then
      Copy(P1,P2);
    else
      Translate(P1.C,T,P2.C); 
      if P1.F>PDeg then
        P2.F := PDeg1;
        P2.R := P1.R;
      else
        P2.F := 0;
        P2.R := -(abs(T)-P1.R);
        SetRho(R,P2);
        LiftErrs(P2,PDeg);
      end if;
    end if;
  end Translate;      
  
  procedure Der(R: in Radius; P: in out Taylor1) is
    C: Poly1 renames P.C;
  begin
    if P.F>PDeg then
      if P.F=FConst then
        SetZero(C);
      else
        Num_Der(C);
        P.F := TrueF(P);
      end if;
      P.R := R;
      return;
    end if;
    if P.R <= R then
      Show2("",P.R,R);
      raise Domain_Error with Show0("Taylors1.Der: need P.R>R");
    end if;
    LiftErrs(P,1);
    declare
      SM: constant Scalar := Inv(Log(Scal(P.R)/R));
      E1: constant Scalar := Exp(Scal(-1));
      RM: constant Flt := MaxNorm(SM);
      FM: constant Flt := MaxNorm(SM*E1/R);
      function DerFac(I: Positive) return Flt is
      begin
        if Flt(I)>RM then
          return ((R/P.R)**(I-1))/P.R;
        else
          return FM/Flt(I);
        end if;
      end DerFac;
    begin
      for I in P.F .. PDeg loop
        ErrMult(DerFac(I),C(I));
      end loop;
    end;
    Num_Der(C);
    P.F := P.F-1;
    P.R := R;
  end Der;
  
  procedure ValDer(P: in Taylor1; S: in Scalar; D: in out Scalar) is
    Tmp: LT_Pointer;
  begin
    if P.F>PDeg then
      if P.F=FConst then
        SetZero(D);
      else
        Num_ValDer(P.C,S,D);
      end if;
    else
      Pool.Allocate(Tmp);
      declare
        R: constant Radius := MaxNorm(S);
        Q: Taylor1 renames Tmp.Data;
      begin
        Copy(P,Q);
        if R=Zero then
          LiftErrs(Q,1);
          Copy(Q.C(1),D);
        else
          Der(R,Q);
          Val(Q,S,D);
        end if;
      end;
      Pool.Recycle(Tmp);
    end if;
  end ValDer;
  
  procedure DivArg(P: in out Taylor1) is
--- [P(x)-P(0)]/x
    C: Poly1 renames P.C;
  begin
    if P.F=FConst then
      SetZero(C(0));
    else
      LiftErrs(P,1);
      for I in 1 .. PDeg loop Copy(C(I),C(I-1)); end loop;
      SetZero(C(PDeg));
      if P.F <= PDeg then P.F := P.F-1; end if;
    end if;
  end DivArg;
  
  procedure AntiDer(P: in out Taylor1) is
    pragma Warnings (Off); -- see below
    C: Poly1 renames P.C;
    E: Scalar;
  begin
    ToErr(C(PDeg),E);
    for I in reverse 2 .. QuasiDeg(C) loop --- empty loop if PDeg=1
      Quot(C(I-1),Flt(I),C(I));
    end loop;
    Copy(C(0),C(1));
    if not IsZero(E) then
      Div(Flt(PDeg+1),E);
      AddProd(P.R,E,C(PDeg));
    end if;
    SetZero(C(0));
    if P.F<PDeg then P.F := P.F+1; end if;
  end AntiDer;
  
  procedure AntiDer(P1: in Taylor1; P2: in out Taylor1) is
    C1: Poly1 renames P1.C;
    C2: Poly1 renames P2.C;
    E: Scalar renames C2(0);
  begin
    for I in reverse 2 .. PDeg loop Quot(C1(I-1),Flt(I),C2(I)); end loop;
    Copy(C1(0),C2(1));
    if not STrunc then
      ToErr(C1(PDeg),E);
      Div(Flt(PDeg+1),E);
      Add(E,C2(PDeg));
    end if;
    SetZero(C2(0));
    if P1.F<PDeg then P2.F := P1.F+1; else P2.F := P1.F; end if;
    P2.R := P1.R;
  end AntiDer;
  
  function Integral(P: Taylor1; S1,S2: Scalar) return Scalar is
    Tmp: LT_Pointer;
    V1,V2: Scalar;
  begin
    Pool.Allocate(Tmp);
    AntiDer(P,Tmp.Data);
    Val(Tmp.Data,S1,V1);
    Val(Tmp.Data,S2,V2);
    Pool.Recycle(Tmp);
    Sub(V1,V2);
    return V2;
  end Integral;
  
  procedure Num_MDer(P: in out Taylor1) is
  begin
    Num_MDer(P.C);
  end Num_MDer;
  
  procedure Inv_MDer(P: in out Taylor1; CheckConst: in Boolean := True) is
  begin
    Inv_MDer(P.C,CheckConst);
  end Inv_MDer;
  
  procedure NewQuot(S0,S1: in Scalar; Q: in out Flt; N: in out Integer) is
  begin
    if not (IsZero(S0) or else IsZero(S1)) then
      Q := Q*MaxNorm(S0)/MaxNorm(S1);
      N := N+1;
    end if;
  end NewQuot;
  
  function GuessRho(P: Taylor1) return Radius is
    use Flt_EF;
    DMin: constant Integer := PDeg/5;
    DMax: constant Integer := (4*PDeg)/5;
    C: Poly1 renames P.C;
    N1: Integer := 0;
    Q1: Flt := One;
  begin
    for I in DMin .. DMax loop
      NewQuot(C(I),C(I+1),Q1,N1);
    end loop;
    return Exp(Log(Q1)/Flt(N1));
  end GuessRho;
  
  procedure Test_Some_Ops is
    R: constant Flt := 3.0/2.0;
    Fac: constant Flt := 1.0/4.0;
    X: constant Scalar := Scal(1.0/8.0);
    S0,S: Scalar;
    F0,F: Taylor1;
  begin
    SetZero(R,F0);
    Simple_Random(F0);
    Mult(Fac/MaxNorm(F0),F0);
    Add(1,F0);
    S0 := Val(F0,X);              -- S0 := F0(X);
    S := Inv(S0);
    F := Inv(F0);
    Show2("Inv(F(X)) ",Val(F,X),S);
    S := Sqrt(S0);
    F := Sqrt(F0);
    Show2("Sqrt(F(X)) ",Val(F,X),S);
    S := Root(5,S0);
    F := Root(5,F0);
    Show2("Root(5,F(X)) ",Val(F,X),S);
    S := Exp(S0);
    F := Exp(F0);
    Show2("Exp(F(X)) ",Val(F,X),S);
    S := Cos(S0);
    F := Cos(F0);
    Show2("Cos(F(X)) ",Val(F,X),S);
    S := Sin(S0);
    F := Sin(F0);
    Show2("Sin(F(X)) ",Val(F,X),S);
    S := Log(S0);
    F := Log(F0);
    Show2("Log(F(X)) ",Val(F,X),S);
    S := ArcCos(S0/Four);
    F := ArcCos(F0/Four);
    Show2("ArcCos((F(X))/4) ",Val(F,X),S);
  end Test_Some_Ops;
  
  function Num_Der(P: in Taylor1) return Taylor1 is
    Q: Taylor1 := P;
  begin
    Num_Der(Q.C);
    return Q;
  end Num_Der;
  
  function Num_MDer(P: in Taylor1) return Taylor1 is
    Q: Taylor1 := P;
  begin
    Num_MDer(Q.C);
    return Q;
  end Num_MDer;
  
-------------------------------------------------
-------------------------------------------------
  package TIP is new IPowers (Scalar => Taylor1);
-------------------------------------------------
-------------------------------------------------
  
  procedure IPower(I: in Integer; P1: in Taylor1; P2: in out Taylor1) renames TIP.IPower;
  
  function "**"(P: Taylor1; I: Integer) return Taylor1 is
    Q: Taylor1;
  begin
    IPower(I,P,Q);
    return Q;
  end "**";
  
-----------------------------------------------
--- stuff for Fun_Series ----------------------
-----------------------------------------------
  
  procedure Inv_Split(P1: in Taylor1; P2: in out Taylor1; S: in out Scalar) is
  begin
    Inv(P1.C(0),S);
    Prod(S,P1,P2);
    SetZero(P2.C(0));
  end Inv_Split;
  
  procedure QPower_Split(Q: in Rational; P1: in Taylor1; P2: in out Taylor1; S: in out Scalar) is
  begin
    Copy(P1,P2);
    LiftErrs(P2,1);
    Inv(P2.C(0),S);
    Mult(S,P2);
    SetZero(P2.C(0));
    QPower(Q,P1.C(0),S);
  end QPower_Split;
  
  procedure Exp_Split(P1: in Taylor1; P2: in out Taylor1; S: in out Scalar) is
  begin
    Copy(P1,P2);
    LiftErrs(P2,1);
    Exp(P2.C(0),S);
    SetZero(P2.C(0));
  end Exp_Split;
    
  procedure CosSin_Split(P1: in Taylor1; P2: in out Taylor1; C,S: in out Scalar) is
  begin
    Copy(P1,P2);
    LiftErrs(P2,1);
    Cos(P2.C(0),C);
    Sin(P2.C(0),S);
    SetZero(P2.C(0));
  end CosSin_Split;
  
  procedure Log_Split(P1: in Taylor1; P2: in out Taylor1; S: in out Scalar) is
  begin
    Copy(P1,P2);
    LiftErrs(P2,1);
    Inv(P2.C(0),S);
    Mult(S,P2);
    SetZero(P2.C(0));
    Log(P1.C(0),S);
  end Log_Split;
  
--------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------
  package TFS is new Fun_Series (Numeric => STrunc, Scalar => Scalar, Fun => Taylor1);
--------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------
  
  procedure Inv(P1: in Taylor1; P2: in out Taylor1) is
  begin
    if P1.F=FConst then
      SetZero(P2);
      Inv(P1.C(0),P2.C(0));
    else
      declare
        Iter: constant Integer := Choose(STrunc,PDeg,IMax(PDeg+8,32));
      begin
        TFS.Series_Inv(P1,P2,Iter);
      end;
      if not STrunc then P2.F := IMin(P1.F,PDeg); end if;
    end if;
  end Inv;
  
  procedure Sqrt(P1: in Taylor1; P2: in out Taylor1) is
  begin
    if P1.F=FConst then
      SetZero(P2);
      Sqrt(P1.C(0),P2.C(0));
    else
      declare
        Iter: constant Integer := Choose(STrunc,PDeg,IMax(PDeg+8,32));
      begin
        TFS.Series_QPower(1/2,P1,P2,Iter);
      end;
      if not STrunc then P2.F := IMin(P1.F,PDeg); end if;
    end if;
  end Sqrt;
  
  procedure Root(K: in Positive; P1: in Taylor1; P2: in out Taylor1) is
  begin
    if K=1 then
      Copy(P1,P2);
    elsif P1.F=FConst then
      SetZero(P2);
      Root(K,P1.C(0),P2.C(0));
    else
      declare
        Iter: constant Integer := Choose(STrunc,PDeg,IMax(PDeg+8,32));
      begin
        TFS.Series_QPower(1/K,P1,P2,Iter);
      end;
      if not STrunc then P2.F := IMin(P1.F,PDeg); end if;
    end if;
  end Root;
  
  procedure QPower(Q: in Rational; P1: in Taylor1; P2: in out Taylor1) is
  begin
    if P1.F=FConst then
      SetZero(P2);
      QPower(Q,P1.C(0),P2.C(0));
    elsif Den(Q)=1 then
      IPower(Num(Q),P1,P2);
    else
      declare
        Iter: constant Integer := Choose(STrunc,PDeg,IMax(PDeg+8,32));
      begin
        TFS.Series_QPower(Q,P1,P2,Iter);
      end;
      if not STrunc then P2.F := IMin(P1.F,PDeg); end if;
    end if;
  end QPower;
  
  procedure Exp(P1: in Taylor1; P2: in out Taylor1) is
  begin
    if P1.F=FConst then
      SetZero(P2);
      Exp(P1.C(0),P2.C(0));
    else
      declare
        Iter: constant Integer := Choose(STrunc,PDeg,IMax(PDeg+8,32));
      begin
        TFS.Series_Exp(P1,P2,Iter);
      end;
      if not STrunc then P2.F := IMin(P1.F,PDeg); end if;
    end if;
  end Exp;
  
  procedure CosSin(P: in Taylor1; PC,PS: in out Taylor1) is
  begin
    if P.F=FConst then
      SetZero(PC);
      Cos(P.C(0),PC.C(0));
      SetZero(PS);
      Sin(P.C(0),PS.C(0));
    else
      declare
        Iter: constant Integer := Choose(STrunc,PDeg/2,IMax(PDeg+8,32)/2); -- uses P**2
      begin
        TFS.Series_CosSin(P,PC,PS,Iter);
      end;
      if not STrunc then PC.F := IMin(P.F,PDeg); PS.F := PC.F; end if;
    end if;
  end CosSin;
 
  procedure Log(P1: in Taylor1; P2: in out Taylor1) is
  begin
    if P1.F=FConst then
      SetZero(P2);
      Log(P1.C(0),P2.C(0));
    else
      declare
        Iter: constant Integer := Choose(STrunc,PDeg,IMax(PDeg+8,32));
      begin
        TFS.Series_Log(P1,P2,Iter);
      end;
      if not STrunc then P2.F := IMin(P1.F,PDeg); end if;
    end if;
  end Log;
  
  procedure ArcSin_Part(P1: in Taylor1; P2: in out Taylor1) is
--- approx and modulo a constant
    Tmp: LT_Pointer;
  begin
    Pool.Allocate(Tmp);
    declare
      Q: Taylor1 renames Tmp.Data;
    begin
      Prod(P1,P1,Q);
      Center(Q);
      Neg(Q);
      Add(1,Q);
      QPower((-1)/2,Q,P2);  -- P2 := 1/Sqrt(1-P1*P1)
      Center(P2);
      Copy(P1,Q);
      Center(Q);
      Num_MDer(Q.C);
      Mult(Q,P2);           -- P2 := P1'/Sqrt(1-P1*P1)
    end;
    Pool.Recycle(Tmp);
    Center(P2);
    Inv_MDer(P2.C,False);   -- integrate P2
  end ArcSin_Part;
  
--------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------
  package TN is new Newton(Numeric => STrunc, Scalar => Taylor1, Operator => Taylor1);
--------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------
  
  procedure ArcCos(P1: in Taylor1; P2: in out Taylor1) is
  begin
    if P1.F=FConst then
      SetZero(P2);
      ArcCos(P1.C(0),P2.C(0));
      return;
    end if;
    ArcSin_Part(P1,P2);
    Neg(P2);
    ArcCos(P1.C(0),P2.C(0));
    Center(P2);
    if STrunc then return; end if;
    declare
      procedure F(X: in Taylor1; Y: in out Taylor1) is
      begin
        Cos(X,Y);
        Sub(P1,Y);
      end F;
      procedure DF(X: in Taylor1; Y: in out Taylor1) is
      begin
        Sin(X,Y);
        Neg(Y);
      end DF;
      procedure Find is new TN.FindZero(F => F, DF => DF);
      procedure Improve is new TN.ImproveZero(F => F, DF => DF);
    begin
      Find(P2,0,Newton_Parallel);
      Improve(P2,64,Newton_Parallel);
    end;
    LiftErrs(P2,1);
    ArcCos(P1.C(0),P2.C(0));
  end ArcCos;
  
  procedure ArcSin(P1: in Taylor1; P2: in out Taylor1) is
  begin
    if P1.F=FConst then
      SetZero(P2);
      ArcSin(P1.C(0),P2.C(0));
      return;
    end if;
    ArcSin_Part(P1,P2);
    ArcSin(P1.C(0),P2.C(0));
    Center(P2);
    if STrunc then return; end if;
    declare
      procedure F(X: in Taylor1; Y: in out Taylor1) is
      begin
        Sin(X,Y);
        Sub(P1,Y);
      end F;
      procedure DF(X: in Taylor1; Y: in out Taylor1) is
      begin
        Cos(X,Y);
      end DF;
      procedure Find is new TN.FindZero(F => F, DF => DF);
      procedure Improve is new TN.ImproveZero(F => F, DF => DF);
    begin
      Find(P2,0,Newton_Parallel);
      Improve(P2,64,Newton_Parallel);
    end;
    LiftErrs(P2,1);
    ArcSin(P1.C(0),P2.C(0));
  end ArcSin;
  
----------------------------------------------
----------------------------------------------
  package TR is new Roots (Scalar => Taylor1);
----------------------------------------------
----------------------------------------------
  
  procedure Roots2(B,C: in Taylor1; U1,U2,V: in out Taylor1) is
    Steps: constant Integer := 64;
  begin
    if Simple(B) and then Simple(C) then
      SetZero(U1);
      SetZero(U2);
      SetZero(V);
      Roots2(B.C(0),C.C(0),U1.C(0),U2.C(0),V.C(0));
    else
      TR.NewtonRoots2(B,C,U1,U2,V,Steps,Newton_Parallel);
    end if;
  end Roots2;
  
  procedure Roots3(B,C,D: in Taylor1; U0,U1,U2,V: in out Taylor1) is
    Steps: constant Integer := IMax(PDeg,64);
  begin
    if Simple(B) and then Simple(C) and then Simple(D) then
      SetZero(U0);
      SetZero(U1);
      SetZero(U2);
      SetZero(V);
      Roots3(B.C(0),C.C(0),D.C(0),U0.C(0),U1.C(0),U2.C(0),V.C(0));
    else
      TR.NewtonRoots3(B,C,D,U0,U1,U2,V,Steps,Newton_Parallel);
    end if;
  end Roots3;
  
begin
  
  if FConst <= PDeg1 then
    raise Sorry with "Taylors1: increase FConst";
  end if;
  
end Taylors1;
